#!/usr/bin/env python
###############################################################################
#  bva.py: BitVector Algorithm
#  Copyright (C) 2012 Brno University of Technology, ANT @ FIT
#  Author(s): Martin Foukal <xfouka00@stud.fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id: $

import sys

from netbench.classification.rule import *
from netbench.classification.ruleset import *
from netbench.classification.prefix import *
from netbench.classification.prefixset import *
from netbench.classification.common import *
from netbench.common.packetheader import *
from netbench.classification.algorithms.bclassification import * 

from netbench.classification.range import *
from netbench.classification.range_bv import *

"""
Module for experiments with Bitvector algorithm (BVA).
"""

class BVA(BClassification):
    '''
    Bitvector algorithm
    '''


    def __init__(self):
        '''
        Contructor
        '''
        
    def load_ruleset(self, ruleset):
        """
        Load rules and create bitvectors for non-overlapping 
        intervals (ranges) in all dimensions
        
        ruleset - Intance of Ruleset class
        """
        
        orig_count = ruleset.count_rules()
        
        if (orig_count == 0):
            print "ERROR: Ruleset is empty"
            return
        
        # Copy ruleset
        self._ruleset = RuleSet()
        for rule in ruleset.get_rules():
            self._ruleset.add_rule(rule)
        
        # Filter out duplicates and rules covered by other rules
        self._ruleset.remove_covered()
        
        print "Number of rules:", self._ruleset.count_rules(),
        print "(%d loaded, %d filtered out)"%(orig_count,orig_count-self._ruleset.count_rules())                        
        
        #add universal prefixes
        self._ruleset.add_universal_prefixes()
        
        # Dictionary (name:string, list of Range objects - for every dimension)                
        rules = {}                
        for field in self._ruleset.get_field_names():
            rules[field] = []
            
        #for all rules, separate its field, convert prefixset to range 
        #and add to list rules[field]
        for r in self._ruleset.get_rules():
            #r.display()
            for field in rules: 
                
                begin = 0
                end = 0                
                domain_size = 32
                
                if   (field is "srcipv4") or (field is "dstipv4"):
                    domain_size = 32
                elif (field is "src_port") or (field is "dst_port"):
                    domain_size = 16
                elif (field is "protocol"):
                    domain_size = 8
                
                prefix = r.get_condition(field)[0]
                
                if prefix.is_universal():
                    begin = 0
                    end = (1 << domain_size) - 1                
                else:
                    if isinstance(prefix, Prefix):
                        if len(r.get_condition(field).get_prefixes()) == 1:
                            l = prefix.get_range()
                            begin = l[0]
                            end = l[1]                            
                        else:
                            l = self._prefixes_to_range(r.get_condition(field).get_prefixes(), domain_size)                        
                            begin = l[0]
                            end = l[1]                            
                    elif isinstance(prefix, MaskedInt):
                        begin = prefix.get_value()
                        end = begin
                                                        
                rules[field].append(Range(begin, end, domain_size))                
                            
        self._ranges = {}
        self._rules = rules
        self._process_ranges()
        
        
    def _process_ranges(self):
        """
        Internal method for processing all ranges
        This method is overridden in BVA2 subclass.
        """ 
        for k,v in self._rules.iteritems():
            if   (k is "srcipv4") or (k is "dstipv4"):
                self._ranges[k] = self._create_bitvectors(rule_dimension_ranges=v, domain_size=32)
            elif (k is "src_port") or (k is "dst_port"):
                self._ranges[k] = self._create_bitvectors(rule_dimension_ranges=v,domain_size=16)
            elif (k is "protocol"):
                self._ranges[k] = self._create_bitvectors(rule_dimension_ranges=v,domain_size=8)
                
                    
    def _prefixes_to_range(self, prefixes, domain_size):
        """
        Convert list of prefixes to range
        I suppose that prefixes from prefixset are in one interval!
        """
        
        begin = (1 << domain_size) - 1
        end = 0
        
        #prefixes needn't be sorted, we find begin and end
        for p in prefixes:            
            if p.get_range()[0] < begin:
                begin = p.get_range()[0]                
            if p.get_range()[1] > end:
                end = p.get_range()[1]
        
        return [begin,end]
        
                                             
    def _create_bitvectors(self, rule_dimension_ranges, domain_size = 32):
        """
        Extract all non overlapping intervals (ranges) in one dimension
        
        rule_dimension_ranges: list of rule ranges in one dimension
        domain_size: 32/16/8
        
        Return list of non-overlapping ranges with bitvectors        
        Bits in bitvector indicate rules which cover given interval
        """
        
        #list of non-overlapping ranges with bitvector
        ranges = []
        #add first range which cover all dimension
        interval = RangeBV(0,(1 << domain_size) - 1,domain_size)
        interval.bitvector = [0 for i in range(len(rule_dimension_ranges))]        
        ranges.append(interval)
                
        #for all rules
        for i,interval in enumerate(rule_dimension_ranges):
                                                        
            #print "%d. interval, begin: %d, end: %d"%(i + 1, interval.begin, interval.end)
            #interval.display()            
            #print "\n"
            
            begin = interval.begin
            end = interval.end 
                                                
            begin_index = 0
            end_index = 0
            
            # find range where rule-range begins and ends    
            for j,r in enumerate(ranges):
                if (begin >= r.begin and begin <= r.end):
                    begin_index = j                    
                if (end >= r.begin and end <= r.end):
                    end_index = j                    
                    break
                        
            #rule_range is in same range
            if (begin_index == end_index):
                j = begin_index
                if ((begin == ranges[j].begin) and (end == ranges[j].end)):
                    #1.case: rule_range equals found range |begin...........end|
                    # set bit in rule position
                    ranges[j].bitvector[i] = 1
                elif ((begin == ranges[j].begin) and (end < ranges[j].end)):
                    #2.case:  |begin........end......|                                                    
                    # divide interval to two and set bit in first
                    ranges[j].begin = end + 1
                    interval = RangeBV(begin,end,domain_size)
                    interval.bitvector = ranges[j].bitvector[:]
                    interval.bitvector[i] = 1
                    ranges.insert(j, interval)
                elif ((begin > ranges[j].begin) and (end == ranges[j].end)):
                    #3.case:  |......begin.....end|
                    # divide interval to two and set bit in second
                    ranges[j].end = begin - 1
                    interval = RangeBV(begin,end,domain_size)
                    interval.bitvector = ranges[j].bitvector[:]
                    interval.bitvector[i] = 1
                    ranges.insert(j + 1, interval)
                else:                    
                    #4.case:  |....b.....e.....|
                    # divide interval to three and set bit in middle
                    tmp_end = ranges[j].end                    
                    ranges[j].end = begin - 1
                    
                    interval = RangeBV(begin,end,domain_size)
                    interval.bitvector = ranges[j].bitvector[:]
                    interval.bitvector[i] = 1
                    ranges.insert(j + 1, interval)
                    
                    interval = RangeBV(end + 1,tmp_end,domain_size)
                    interval.bitvector = ranges[j].bitvector[:]                    
                    ranges.insert(j + 2, interval)
            # rule_range covers more ranges
            else:
                j = begin_index
                # |begin........|....
                if (begin == ranges[j].begin):
                    ranges[j].bitvector[i] = 1
                    j += 1
                # |.....begin......|....
                else:
                    tmp_end = ranges[j].end
                    ranges[j].end = begin - 1
                    interval = RangeBV(begin,tmp_end,domain_size)
                    interval.bitvector = ranges[j].bitvector[:]
                    interval.bitvector[i] = 1
                    ranges.insert(j + 1, interval)
                    j += 2
                    end_index += 1
                                
                # set bits in all ranges which rule_range covers
                while (j < end_index):
                    ranges[j].bitvector[i] = 1
                    j += 1                                                
                        
                # ...|........end|
                if (end == ranges[j].end):
                    ranges[j].bitvector[i] = 1
                # ...|......end......|
                else:                                    
                    tmp_end = ranges[j].end
                    ranges[j].end = end                    
                    interval = RangeBV(end + 1,tmp_end,domain_size)
                    interval.bitvector = ranges[j].bitvector[:]                    
                    ranges.insert(j + 1, interval)
                    ranges[j].bitvector[i] = 1
                                                       
        return ranges   
    
    def classify(self, packetheader):
        """
        Classify packet by the algorithm.
        Return the list of all rules that matches given packet.
        
        packetheader: Instance of the PacketHeader class.
        """
            
        found_bitvectors = {}
                
        for key,intervals in self._ranges.iteritems(): # For every dimension (condition)
            try:
                # Get value from packet and convert it to integer
                field = extract_field(packetheader, key)
                
                #print key, ":", field
                
                if field < intervals[0].begin or field > intervals[len(intervals) - 1].end:
                    raise ValueError("incorrect packet field value") 
                left = 0
                right = len(intervals) - 1
                i = (left + right) / 2                            
                        
                #using bisection method to find range where packet field belongs
                while True:                    
                    if (intervals[i].begin <= field and intervals[i].end >= field):
                        break
                    if (right - left == 1):
                        i += 1
                        break
                    if (field < intervals[i].begin):
                        right = i
                    if (field > intervals[i].end):
                        left = i
                        
                    i = (left + right) / 2
                    
                #print "field is %d , index: %d\nin interval: "%(field, i)
                #intervals[i].display()
                
                if field < intervals[i].begin or field > intervals[i].end:
                    raise ValueError("field is not in found interval")
                                
                found_bitvectors[key] = intervals[i].bitvector
                
            except FieldNotFoundError:
                raise BadPacketError("Field '%s' not found."%key)

        
        result_bitvector = []
        # using binary AND to create result bitvector
        for field, bitvector in found_bitvectors.iteritems():
            #print bitvector
            if result_bitvector == []:
                result_bitvector = bitvector[:]
            else:
                for i,bit in enumerate(bitvector):
                    result_bitvector[i] &= bit
                    
        #print "result: \n", result_bitvector 
        
        rules = []
        
        # add all matching rules to list
        for i,bit in enumerate(result_bitvector):
            if bit == 1:
                rules.append(self._ruleset.get_rule(i))
                
        return rules
    
    def report_memory(self):
        """
        Print detailed info about algorithm memory requirements.
        """                    
        print "========== BVA memory report =========="
        print     "Number of all rules:                   %d\n"%(self._ruleset.count_rules())
        sum_bounds = 0
        space = ""
        for k,v in self._ranges.iteritems():
            if   (k is "srcipv4") or (k is "dstipv4"):
                domain_size = 32
                space = " "
            elif (k is "src_port") or (k is "dst_port"):
                domain_size = 16
                space = ""
            elif (k is "protocol"):
                domain_size = 8
                space = ""            
            print "Number of bitvectors in %s:%s      %d"%(k,space,len(v))
            print "Size   of bitvectors in %s:%s      %d bits"%(k,space,len(v)*self._ruleset.count_rules())
            print "Size   of bounds     in %s:%s      %d bits"%(k,space,len(v)*2*domain_size)        
            sum_bounds += len(v)*2*domain_size
            print "\n"
        print     "Number of all bitvectors:             ", sum((len(v) for k,v in self._ranges.iteritems()))
        bv_size_sum = sum((len(v) for k,v in self._ranges.iteritems()))*self._ruleset.count_rules()
        print     "Size   of all bitvectors:              %d bits"%(bv_size_sum)
        print     "Size   of all bounds:                  %d bits"%(sum_bounds)
        print     "\nMemory usage summary:                  %d bits"%(bv_size_sum+sum_bounds)
        
                
            
            
         
            
