#!/usr/bin/env python
###############################################################################
#  bva2.py: BitVector Algorithm - improved version
#  Copyright (C) 2012 Brno University of Technology, ANT @ FIT
#  Author(s): Martin Foukal <xfouka00@stud.fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id: $

import sys
import math

from netbench.classification.rule import *
from netbench.classification.ruleset import *
from netbench.classification.prefix import *
from netbench.classification.prefixset import *
from netbench.classification.common import *
from netbench.common.packetheader import *
from netbench.classification.algorithms.bclassification import * 

from netbench.classification.range import *
from netbench.classification.range_bv import *
from netbench.classification.algorithms.bva import *


"""
Module for experiments with second version of Bitvector algorithm (BVA2) 
based on incremental reads.
"""

class BVA2(BVA):
    '''
    Bitvector algorithm - second version
    It's subclass of BVA (main bv version)
    '''


    def __init__(self):
        '''
        Contructor        
        '''
        BVA.__init__(self)
    
    def load_ruleset(self, ruleset):
        BVA.load_ruleset(self, ruleset)
        
    
    def _process_ranges(self):
        """
        Internal method for processing all ranges        
        """                 
        for k,v in self._rules.iteritems():
            if   (k is "srcipv4") or (k is "dstipv4"):
                self._ranges[k] = self._create_intervals(rule_dimension_ranges=v, domain_size=32)                
            elif (k is "src_port") or (k is "dst_port"):
                self._ranges[k] = self._create_intervals(rule_dimension_ranges=v,domain_size=16)
            elif (k is "protocol"):
                self._ranges[k] = self._create_intervals(rule_dimension_ranges=v,domain_size=8)
            
            self._create_bitvectors_and_pointers(intervals=self._ranges[k], rule_dimension_ranges=v)
                                                
                                             
    def _create_intervals(self, rule_dimension_ranges, domain_size = 32):
        """
        Extract all non overlapping intervals (ranges) in one dimension
        
        rule_dimension_ranges: list of rule ranges in one dimension
        domain_size: 32/16/8
        
        Return list of non-overlapping ranges                
        """
        
        #list of non-overlapping ranges
        ranges = []
        #add first range which cover all dimension
        interval = RangeBV(0,(1 << domain_size) - 1,domain_size)        
        ranges.append(interval)
                
        #for all rules
        for interval in rule_dimension_ranges:                                                                    
            
            begin = interval.begin
            end = interval.end 
                                                
            begin_index = 0
            end_index = 0
            
            # find range where rule-range begins and ends    
            for j,r in enumerate(ranges):
                if (begin >= r.begin and begin <= r.end):
                    begin_index = j                    
                if (end >= r.begin and end <= r.end):
                    end_index = j                    
                    break
                        
            #rule_range is in same range
            if (begin_index == end_index):
                j = begin_index                
                if ((begin == ranges[j].begin) and (end < ranges[j].end)):
                    #2.case:  |begin........end......|                                                    
                    # divide interval
                    ranges[j].begin = end + 1
                    interval = RangeBV(begin,end,domain_size)                    
                    ranges.insert(j, interval)
                elif ((begin > ranges[j].begin) and (end == ranges[j].end)):
                    #3.case:  |......begin.....end|
                    # divide interval to two
                    ranges[j].end = begin - 1
                    interval = RangeBV(begin,end,domain_size)                    
                    ranges.insert(j + 1, interval)
                elif ((begin > ranges[j].begin) and (end < ranges[j].end)):                    
                    #4.case:  |....b.....e.....|
                    # divide interval to three
                    tmp_end = ranges[j].end                    
                    ranges[j].end = begin - 1
                    
                    interval = RangeBV(begin,end,domain_size)                    
                    ranges.insert(j + 1, interval)
                    
                    interval = RangeBV(end + 1,tmp_end,domain_size)                                        
                    ranges.insert(j + 2, interval)
            # rule_range covers more ranges
            else:
                j = begin_index
                # |.....begin......|....
                if (begin > ranges[j].begin):
                    tmp_end = ranges[j].end
                    ranges[j].end = begin - 1
                    interval = RangeBV(begin,tmp_end,domain_size)                    
                    ranges.insert(j + 1, interval)
                    #we have inserted new interval, so we must increment indices
                    j += 2
                    end_index += 1
                                                
                j = end_index
                # ...|......end......|
                if end < ranges[j].end:                                    
                    tmp_end = ranges[j].end
                    ranges[j].end = end                    
                    interval = RangeBV(end + 1,tmp_end,domain_size)                                        
                    ranges.insert(j + 1, interval)                    
                                                       
        return ranges   
    
    
    def _create_bitvectors_and_pointers(self, intervals, rule_dimension_ranges):
        """
        Method for storing bitvector or pointers to each range
        Bitvector is stored if l or more bits (indicating rules) changed
        Pointers indicating different bits are stored between bitvectors
        
        intervals: list of non-overlapping ranges in given dimension
        rule_dimension_ranges: list of rule ranges in given dimension        
        """
        
        n = len(rule_dimension_ranges)
        #calculate l value which indicate interval between stored bitvectors        
        l = int(math.ceil(0.5*(n/math.log10(n))))
        self._lvalue = l
        bits_diff = 0
        #calculate first bitvector
        intervals[0].bitvector = [0 for k in range(len(rule_dimension_ranges))]
        for i,r in enumerate(rule_dimension_ranges):
                if r.covers(intervals[0]):
                    intervals[0].bitvector[i] = 1
        
        prev_bitvector = intervals[0].bitvector        
        #index of lower bitvector
        lower_bv_i = 0                
        #store bitvector or pointers
        for i in range(1,len(intervals)):
            #calculate bitvector
            act_bitvector = [0 for k in range(len(rule_dimension_ranges))]
            for j,r in enumerate(rule_dimension_ranges):
                if r.covers(intervals[i]):                
                    act_bitvector[j] = 1
                                    
            #calculate pointers
            pointers = {}
            for k in range(len(rule_dimension_ranges)):
                if act_bitvector[k] != prev_bitvector[k]:
                    pointers[k] = act_bitvector[k]
            
            bits_diff += len(pointers)
            
            # different bits more than l value, so store bitvector
            if bits_diff >= l:
                intervals[i].bitvector = act_bitvector
                bits_diff = 0
                #assign index of actual bitvector to higher bitvector
                higher_bv_i = i                
                
                #if exist intervals between intervals with stored bitvectors (max l-1 such intervals)
                if higher_bv_i - lower_bv_i > 1:
                    # now we must calculate pointers at mid-point and upper half of l-1 intervals
                    # they are calculated from higher bitvector
                    prev2_bitvector = act_bitvector                
                    m = higher_bv_i - 1                
                    while m >= ((lower_bv_i + 1) + (higher_bv_i - 1 - lower_bv_i)/2):
                        act2_bitvector = [0 for k in range(len(rule_dimension_ranges))]
                        for j,r in enumerate(rule_dimension_ranges):
                            if r.covers(intervals[m]):                
                                act2_bitvector[j] = 1
                                
                        intervals[m].pointers = {}
                        for k in range(len(rule_dimension_ranges)):
                            if act2_bitvector[k] != prev2_bitvector[k]:
                                intervals[m].pointers[k] = act2_bitvector[k]
                        prev2_bitvector = act2_bitvector                
                        m -= 1      
                        
                
                lower_bv_i = higher_bv_i
                    
            #store pointers which are calculated from lower bitvector (can be later overwritten 
            #if interval will be at upper half of intervals)                
            else:
                intervals[i].pointers = pointers                                
            
            prev_bitvector = act_bitvector                                  
    
    def classify(self, packetheader):
        """
        Classify packet by the algorithm.
        Return the list of all rules that matches given packet.
        
        packetheader: Instance of the PacketHeader class.
        """
            
        found_bitvectors = {}
                
        for key,intervals in self._ranges.iteritems(): # For every dimension (condition)
            try:
                # Get value from packet and convert it to integer
                field = extract_field(packetheader, key)
                
                #print key, ":", field
                
                if field < intervals[0].begin or field > intervals[len(intervals) - 1].end:
                    raise ValueError("incorrect packet field value") 
                left = 0
                right = len(intervals) - 1
                i = (left + right) / 2                            
                        
                #using bisection method to find range where packet field belongs
                while True:                    
                    if (intervals[i].begin <= field and intervals[i].end >= field):
                        break
                    if (right - left == 1):
                        i += 1
                        break
                    if (field < intervals[i].begin):
                        right = i
                    if (field > intervals[i].end):
                        left = i
                        
                    i = (left + right) / 2
                    
                #print "field is %d , index: %d\nin interval: "%(field, i)
                #intervals[i].display("ip")
                
                if field < intervals[i].begin or field > intervals[i].end:
                    raise ValueError("field is not in found interval")
                                                
                if intervals[i].has_bitvector():                    
                    found_bitvectors[key] = intervals[i].bitvector
                # interval has pointers, so we must calculate bitvector from lower or higher bitvector
                # according position
                else:                    
                    #first find lower and higher indices of intervals with bitvector
                    j = i
                    while not intervals[j].has_bitvector():
                        j -= 1
                    lower_bv_i = j
                    
                    j = i
                    while not intervals[j].has_bitvector() and j < len(intervals) - 1:
                        j += 1
                    
                    higher_bv_i = j

                    #calculate middle index, higher index can be end of intervals and need't have
                    #bitvector, in such case middle_index is None
                    middle_index = 0
                    if not intervals[higher_bv_i].has_bitvector():                        
                        middle_index = None
                                                        
                    if middle_index is not None:
                        middle_index = ((lower_bv_i + 1) + (higher_bv_i - 1 - lower_bv_i)/2)

                    #calculare bitvector from lower bitvector
                    if middle_index is None or i < middle_index:
                        found_bitvectors[key] = intervals[lower_bv_i].bitvector[:]
                        j = lower_bv_i + 1
                        while j <= i:
                            for index,bitval in intervals[j].pointers.iteritems():
                                found_bitvectors[key][index] = bitval
                            j += 1
                            
                    #calculare bitvector from higher bitvector
                    else:                            
                        found_bitvectors[key] = intervals[higher_bv_i].bitvector[:]
                        j = higher_bv_i - 1
                        while j >= i:
                            for index,bitval in intervals[j].pointers.iteritems():
                                found_bitvectors[key][index] = bitval
                            j -= 1
                                     
            except FieldNotFoundError:
                raise BadPacketError("Field '%s' not found."%key)

        
        result_bitvector = []
        # using binary AND to create result bitvector
        for field, bitvector in found_bitvectors.iteritems():
            #print bitvector
            if result_bitvector == []:
                result_bitvector = bitvector[:]
            else:
                for i,bit in enumerate(bitvector):
                    result_bitvector[i] &= bit
                    
        #print "result:\n", result_bitvector 
        
        rules = []
        
        # add all matching rules to list
        for i,bit in enumerate(result_bitvector):
            if bit == 1:
                rules.append(self._ruleset.get_rule(i))
                
        return rules
    
    def report_memory(self):
        """
        Print detailed info about algorithm memory requirements.
        """                    
        print "========== BVA2 memory report =========="
        print     "Number of all rules:                   %d\n"%(self._ruleset.count_rules())
        pointers_sum = 0
        sum_bounds = 0
        space = ""
        for k,v in self._ranges.iteritems():
            if   (k is "srcipv4") or (k is "dstipv4"):
                domain_size = 32
                space = " "
            elif (k is "src_port") or (k is "dst_port"):
                domain_size = 16
                space = ""
            elif (k is "protocol"):
                domain_size = 8
                space = ""
                
            print "Number of bitvectors in %s:%s      %d"%(k,space,len([i for i in v if i.has_bitvector()]))
            print "Size   of bitvectors in %s:%s      %d bits"%(k,space,len([i for i in v if i.has_bitvector()])*self._ruleset.count_rules())
            
            print "Size   of bounds     in %s:%s      %d bits"%(k,space,len(v)*2*domain_size)        
            
            sum_bounds += len(v)*2*domain_size            
            pointers_num = sum([len(i.pointers) for i in v if i.has_pointers()])
            pointers_sum += pointers_num
            
            print "Number of pointers   in %s:%s      %d"%(k,space,pointers_num)    
            print "Size   of pointers   in %s:%s      %d bits"%(k,space,pointers_num*math.ceil(math.log(self._ruleset.count_rules(),2)))
            print "\n"
                              
        print     "Number of all bitvectors:             ", sum((len([i for i in v if i.has_bitvector()]) 
                                                                for k,v in self._ranges.iteritems()))
        bv_size_sum = sum((len([i for i in v if i.has_bitvector()]) 
                      for k,v in self._ranges.iteritems()))*self._ruleset.count_rules()
        print     "Size   of all bitvectors:              %d bits"%(bv_size_sum)
        print     "Size   of all bounds:                  %d bits"%(sum_bounds)        
        print     "Number of all pointers:                %d"%(pointers_sum)
        pointers_size_sum = (pointers_sum*math.ceil(math.log(self._ruleset.count_rules(),2)))
        print     "Size   of all pointers:                %d bits"%(pointers_size_sum)
        print     "Size   of l value:                    ",self._lvalue
        print     "\nMemory usage summary:                  %d bits"%(pointers_size_sum+bv_size_sum+sum_bounds)
            
        
                
            
            
         
            
