###############################################################################
#  hicuts.py: Module for experiments with Hierarchical Intelligent Cuttings Algorithm
#  Copyright (C) 2009 Brno University of Technology, ANT @ FIT
#  Author(s): Martin Spinler <xspinl00@stud.fit.vutbr.cz>
#             Viktor Pus <ipus@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

"""
Module for experiments with Hierarchical Intelligent Cuttings Algorithm
"""

import sys
import math
import random
from netbench.classification.rule import *
from netbench.classification.ruleset import *
from netbench.classification.prefix import *
from netbench.classification.prefixset import *
from netbench.classification.common import *

from netbench.common.packetheader import *
from struct import *
from sets import Set
import bclassification


class Node:
    def __init__(self):
        self.leaf = False
        self.dim = 0           # Which dimension is cutted
        self.ncuts = 0
        self.childs = []
        self.rules = []
        self.low = {}
        self.high = {}

class HiCuts(bclassification.BClassification):
    """
    Hierarchical Intelligent Cuttings.
    """
    def __init__(self, binth = 16, spfac = 2):
        self.spfac = spfac
        self.binth = binth
        self.cond_names = []
        self.prefixes = []
        self.subset = []
        self.maxcuts = 4000
        # Ruleset
        self._rules = None
        # Universal rule (rule that match all packets)
        self._universalrule = None

    def field_size(self, dim):
        if dim == "srcipv4" or dim == "dstipv4":
            return 32
        if dim == "src_port" or dim == "dst_port":
            return 16
        return 8

    def load_ruleset(self, ruleset):

        self._rules = RuleSet()
        self.cond_names = ['srcipv4','dstipv4','src_port','dst_port','protocol']

        # copy rules (this operation will remove duplicates and rules
        #             covered by other rules)
        for rule in ruleset.get_rules():
            self._rules.add_rule(rule, check=True)

        # If there is an universal rule, store it and delete it from rulesets
        self._universalrule = None
        if self._rules.get_rule(-1).is_universal():
            print "Universal rule detected:",
            self._rules.get_rule(-1).display()
            self._universalrule = self._rules.get_rule(-1)
            self._rules.del_rule(-1)

        self._rules.add_universal_prefixes(self.cond_names)

        # expand prefixsets for all rules in order to have rules only with
        # one prefix in each dimension
        self._rules.expand_prefixsets()

        # This is something we can work with.
        ruleset = self._rules

        #ruleset.display()

        node = Node()
        node.rules = ruleset.get_rules()

        self.root_node = node
        for dim in self.cond_names:
            node.low[dim] = 0
            node.high[dim] = int(math.pow(2, self.field_size(dim))) - 1

        node_queue = [self.root_node]
        iteration = 0
        while (len(node_queue) > 0):
            iteration = iteration+1
            #print "================================================"
            #print "iteration:", iteration
            node = node_queue[0]
            #self.print_node(node, 1)
            #print "len(node_queue): ", len(node_queue)
            node_queue = node_queue[1:]
            self.remove_redundancy(node)
            cut = self.select_cut_dimension(node, 3) # Only alg.3 works well
            #print "cut: ", cut
            node.dim = cut[0]
            node.ncuts = cut[1]
            if len(node.rules) <= self.binth or node.ncuts == 1:
                node.leaf = True
                #print "Node is leaf."
            else:
                #print "Node is NOT leaf."
                node.childs = []

                step = int((node.high[node.dim] - node.low[node.dim]) / node.ncuts)

                low = node.low[node.dim]
                high = low + step
                for i in range(node.ncuts):
                    child = Node()
                    node.childs.append(child)
                    node_queue.append(child)
                    for dim in self.cond_names:
                        if dim == node.dim:
                            child.low[dim] = low
                            child.high[dim] = high
                        else:
                            child.low[dim] = node.low[dim]
                            child.high[dim] = node.high[dim]

                    for rule in node.rules:
                        ruleB = self.get_range(rule, node.dim)
                        if (ruleB[0] >= low and ruleB[0] <= high) or \
                           (ruleB[1] >= low and ruleB[1] <= high) or \
                           (ruleB[0] <= low and ruleB[1] >= high):
                            child.rules.append(rule)

                    if len(child.rules) == 0:
                        node_queue.pop()
                        node.childs.pop()
                    else:
                        #print "appending:"
                        #self.print_node(child, 2)
                        pass

                    low = high + 1
                    high = low + step

        print "========= Tree ========="
        self.print_node(self.root_node, 1)
        self.print_XML()

    def print_node(self, node, tab=0, recur=1):
        print ("-" * tab), ("dim: %s,  cuts: %i,  rules: " % (node.dim, node.ncuts)),
        for r in node.rules:
            print repr(r._priority) + "," ,
        print " leaf: ", node.leaf,
        print
        if recur:
            for p in node.childs:
                self.print_node(p, tab + 1)


    def print_XML(self):
        file = open("hicuts.xml", "w")
        file.write('<?xml version="1.0" encoding="UTF-8"?>\n<nodes>\n')
        self.print_node_XML(file, self.root_node)
        file.write('</nodes>')

    def print_node_XML(self, file, node):
        file.write('<node ')
        for dim in self.cond_names:
            file.write('%s="0x%X - 0x%X" ' % (dim, node.low[dim], node.high[dim]))
        file.write('dim="%s" cuts="%i">\n' % (node.dim, node.ncuts))
        for p in node.childs:
            self.print_node_XML(file, p)
        for r in node.rules:
            file.write(repr(r._priority) + ",")
        file.write("</node>\n")
            
    def remove_redundancy(self, node):
        # If an rule locally covers fully another with lower priority, then the second one can be removed
        rule1 = len(node.rules) - 1
        rule2 =  rule1 - 1
        cover = True
        while (rule1 >= 0):
            for rule in range(rule2, -1, -1):
                cover = True
                for dim in self.cond_names:
                    if (max(self.get_range(node.rules[rule], dim)[0], node.low[dim])  > max(self.get_range(node.rules[rule1], dim)[0], node.low[dim])) \
                    or (min(self.get_range(node.rules[rule], dim)[1], node.high[dim]) < min(self.get_range(node.rules[rule1], dim)[1], node.high[dim])):
                        cover = False
                        break
                if cover == True:
                    #for r in node.rules:
                        #print r._priority,
                    #print "Rule %i with priority %i removed due to redundancy" % (rule1, node.rules[rule1]._priority)
                    node.rules.pop(rule1)
                    break
            rule1 = rule1 - 1
            rule2 = rule2 - 1

    def get_range(self, rule, dim):
        if dim == 'protocol':
            return (rule.get_condition(dim)[0].get_value(), rule.get_condition(dim)[0].get_value())
        else:
            return rule.get_condition(dim)[0].get_range()
        
    def select_cut_dimension(self, node, algorithm):
        """
        Return the [dimension, number of cuts] list.
        """
        rules_count = {}
        cuts = {}
        space = {}
        components = {}
        for dim in self.cond_names:
            if node.low[dim] == node.high[dim]:
                cuts[dim] = 1
            else:
                cuts[dim] = 2
            done = False
            while not done:
                space[dim] = 0
                rules_count[dim] = [0] * cuts[dim]
                for rule in node.rules:
                    step = (node.high[dim] - node.low[dim]) / cuts[dim]
                    low = node.low[dim]
                    high = node.low[dim] + step
                    child = self.get_range(rule, dim)
                    for i in range(cuts[dim]):
                        if (child[0] >= low and child[0] <= high) \
                        or (child[1] >= low and child[1] <= high) \
                        or (child[0] <= low and child[1] >= high):
                            space[dim] = space[dim] + 1
                            rules_count[dim][i] = rules_count[dim][i] + 1
                        low = high + 1
                        high = low + step
                space[dim] = space[dim] + cuts[dim]
                if (space[dim] < int(self.spfac * len(node.rules))) \
                and (cuts[dim] <= self.maxcuts/2) \
                and (cuts[dim] < (node.high[dim] - node.low[dim])):
                    cuts[dim] = cuts[dim] * 2
                else:
                    done = True

            max_range_ps = PrefixSet()
            max_range_ps.set_range(node.low[dim], node.high[dim], self.field_size(dim))
            components[dim] = []
            for rule in node.rules:
                cond = rule.get_condition(dim)
                if cond.covers(max_range_ps):
                    cond = max_range_ps
                found = False
                for comp in components[dim]:
                    if (comp.covers(cond) and cond.covers(comp)):
                        found = True
                if not found:
                    components[dim].append(cond)

            #print dim
            #for cond in components[dim]:
                #print cond

        #print "cuts:", cuts
        #print "rules_count:", rules_count
        #print "space:", space
        #print "components:", components
        if algorithm == 0:
            # Leads to infinite loop
            maximum = {}
            for dim in self.cond_names:
                maximum[dim] = max(rules_count[dim])
            minv = 4000000000 # TODO

            for dim in self.cond_names:
                if minv > maximum[dim] and cuts[dim] != 1:# and not maximum[dim] == 0:
                    minv = maximum[dim]
            for dim in self.cond_names:
                if minv == maximum[dim] and cuts[dim] != 1:
                    return (dim, cuts[dim])
        elif algorithm == 1:
            # Leads to infinite loop
            entropy = {}
            for dim in self.cond_names:
                entropy[dim] = 0
                for i in range(cuts[dim]):
                    tmp = float(rules_count[dim][i])/float(space[dim])
                    if tmp == 0:
                        continue
                    entropy[dim] = entropy[dim] - tmp * math.log10(tmp)
            maxfv = 0.0
            for dim in self.cond_names:
                if(maxfv < entropy[dim] and cuts[dim] != 1):
                    maxfv = entropy[dim]
            for dim in self.cond_names:
                if(maxfv == entropy[dim] and cuts[dim] != 1):
                    return (dim, cuts[dim])

            
        elif algorithm == 2:
            # Leads to infinite loop
            maxv = 0 # TODO
            for dim in self.cond_names:
                if (maxv < space[dim] and cuts[dim] != 1):
                    maxv = space[dim]
            for dim in self.cond_names:
                if(maxv == space[dim] and cuts[dim] != 1):
                    return (dim, cuts[dim])

        elif algorithm == 3:
            # Seems to work well
            maxv = 0
            for dim in self.cond_names:
                if (maxv < len(components[dim]) and cuts[dim] != 1):
                    maxv = len(components[dim])
                    maxd = dim
            return (maxd, cuts[maxd])
        
    def classify(self, packetheader):
        node = self.root_node
        bits = {}
        field = {}
        for dim in self.cond_names:
            bits[dim] = self.field_size(dim) # Numbers lowered during descent
            try:
                # Get value from packet and convert it to integer
                field[dim] = extract_field(packetheader, dim)
            except FieldNotFoundError:
                raise BadPacketError("Field '%s' not found."%dim)

        while not node.leaf:
            cuts = node.ncuts
            dim = node.dim
            #self.print_node(node,recur=0)
            #print "packet:{0:d}".format(field[dim])

            # Compute nbits = number of significant bits
            nbits = 0
            while cuts != 1:
                nbits = nbits + 1
                cuts = cuts / 2
            bits[dim] = bits[dim] - nbits

            # Find node sequentially
            newnode = None
            for child in node.childs:
                #print "child:({0}-{1})".format(child.low[dim], child.high[dim])
                if field[dim] >= child.low[dim] and \
                   field[dim] <= child.high[dim]:
                    #print "This one!"
                    newnode = child

            # Cant go deeper
            if newnode == None:
                return [self._universalrule]

            node = newnode

        result = []
        for rule in node.rules:
            if rule.match(packetheader):
                result.append(rule)

        result.append(self._universalrule)
                
        return result
       
    def report_memory(self):
        """
        Print detailed info about algorithm memory requirements.
        """
        print "========= HiCuts memory report ========="
        (maxdepth,numnodes,numnptr,numrptr) = self.count_nodes(self.root_node)
        print "Number of tree nodes: {0}".format(numnodes)
        print "Maximal tree depth: {0}".format(maxdepth)
        print "Number of node pointers: {0}".format(numnptr)
        print "Number of rule pointers: {0}".format(numrptr)
        print "Number of rules (after pref. expansion and without universal rule): {0}".format(self._rules.count_rules())

    def count_nodes(self, node):
        """
        Return tuple (depth, number of nodes, number of node pointers, 
        number of rule pointers)
        """
        if node.leaf:
            return (1, 1, 0, self.binth)
        else:
            maxdepth = 0
            numnodes = 1
            numnptr = node.ncuts
            numrptr = 0
            for child in node.childs:
                (d, n, np, rp) = self.count_nodes(child)
                if d > maxdepth:
                    maxdepth = d
                numnodes += n
                numnptr  += np
                numrptr  += rp
            return (maxdepth+1, numnodes, numnptr, numrptr)
