"""
Created on Dec 4, 2011

@author: Marek Visnovsky <xvisno00@stud.fit.vutbr.cz>
"""
import math
import copy

from netbench.classification.maskedint import MaskedInt
from netbench.classification.prefixset import PrefixSet
from netbench.classification.algorithms import bclassification
from netbench.classification.ruleset import RuleSet

class Node(object):
    """
    Class for node representation in HyperCuts trie
    """
    def __init__(self, region):
        """
        Constructor
        
        region: dictionary of rule fields ranges
        """        
        # Region described by node
        self._region = region
        # Is current node a leaf?
        self._is_leaf = False
        # Cuts in this node
        self._cuts = {}
        # Pointers to child nodes
        self._childs = []
        # Rules that may match
        self._rules = RuleSet()
        
    def add_rule(self, rule):
        """
        Add one rule to the nodes ruleset
        
        rule: rule to add
        """
        self._rules.add_rule(rule)
        
    def add_rules(self, ruleset):
        """
        Add rules stored in node
        
        ruleset: RuleSet containing rules for this node
        """
        self._rules = ruleset
        
    def keep_rules(self, ruleset, dim, max_rules):
        """
        Heuristic for keeping common rules in parent(current) nodes
        
        ruleset: rules to keep potentially
        dim: dimensions, where rules cover node
        max_rules: maximum number of rules to keep in node
        """
        index = 0
        for rule in ruleset.get_rules():
            if index > max_rules:
                break
            push = True
            for dimension in dim:
                try:
                    condition = rule.get_condition(dimension)
                    # PrefixSet
                    if isinstance(condition, PrefixSet):
                        if condition[0].covers(self._region[dimension]):
                            continue
                except KeyError:
                    continue
                push = False
                break
            if push:
                self._rules.add_rule(rule)
                ruleset.del_rule(index)
                index += 1
            else:
                break
            
        return index
    
    def get_rules(self):
        """
        Returns rules in node
        """
        return self._rules
        
    def match(self, packetheader):
        """
        Returns list of rules matching the packet header
        """
        return self._rules.classify(packetheader)
        
    def add_child(self, node):
        """
        Adds node into the list of childs
        
        node: child to add
        """
        self._childs.append(node)
    
    def get_child(self, index):
        """
        Returns a child on specified index
        """
        return self._childs[index]
    
    def get_childs(self):
        """
        Returns childs of the node
        """
        return self._childs
    
    def count_childs(self):
        """
        Returns number of childs
        """
        return len(self._childs)
    
    def get_region(self):
        """
        Returns nodes region
        """
        return self._region
    
    def set_number_of_cuts(self, nc):
        """
        Stores number of cuts in this node
        
        nc: dictionary of number of cuts
        """
        self._cuts = nc
        
    def get_cuts(self):
        """
        Returns cuts applied on node
        """
        return self._cuts
        
    def set_leaf(self):
        """
        Determine that node is leaf
        """
        self._is_leaf = True
    
    def is_leaf(self):
        """
        Return True if node is leaf, thus has no childs
        """
        return self._is_leaf
        
    def __str__(self):
        """
        String representation of node
        """
        repre = ""
        if self._is_leaf:
            repre += "Leaf : ["
        else:
            repre += "Node : ["
        
        for field in self._region.iterkeys():
            repre += field + ": "
            value, mask = self._region[field].get_data()
            repre += str(value) + "/" + str(mask) + ";"
        repre += "]"
        return repre

#######################################################################
# Trie class for HyperCuts algorithm
#######################################################################
class Trie(object):
    """
    Trie structure for HyperCuts algorithm
    """
    # Constants for statistics in bits
    NODE_SIZE = 32
    LEAF_SIZE = 32
    RULE_SIZE = 128
    RULEP_SIZE = 32
    CHILDP_SIZE = 32
    # helper list for keeping dimensions in order
    fields_in_order = ["srcipv4", "dstipv4", "protocol", "src_port", "dst_port"]
    
    def __init__(self, spfac = 2, bucket_size = 8):
        """
        Constructor
        """
        # Maximum number of rules in leaf nodes
        self._bucket_size = bucket_size
        # Space factor
        self._spfac = spfac
        # Initialization of root
        self._root = None
        # Offsets for region description
        self._offsets = None
        # Heuristics
        self._heuristics = {"push_rules": True, "node_merge": True}
        # Trie statistics
        self._statistics = None
        # Empty node
        self._empty_node = Node(None)
        self._empty_node.set_leaf()
        # Helper list for node merging, contains tuples (child, parent, index)
        self._leafs = []
        
    def _reset_stats(self):
        """
        Reset trie statistics
        """
        self._statistics = {"n_of_nodes": 0, "depth": 0, "size": 0, "leafs": 0, "pushed_rules": 0, "merged_nodes": 0}
        
    def _choose_dimensions(self, ruleset):
        """
        Choose dimensions to cut according to its number of
        unique elements. Returns list of chosen dimensions
        
        ruleset: rules
        """
        mean = 0
        unique = { "srcipv4": [], "dstipv4": [], "protocol": [],
                        "src_port": [], "dst_port": [] }
                
        for rule in ruleset.get_rules():
            for condition in unique.iterkeys():
                try:
                    if not (rule.get_condition(condition) in unique[condition]):
                        unique[condition].append(rule.get_condition(condition))
                except KeyError:
                    pass

        for condition in unique.iterkeys():
            mean += len(unique[condition])
        mean /= 5.0
        
        return [field for field in unique.iterkeys() if len(unique[field]) > mean]
    
    def _optimum_nc(self, dimension, ruleset, region):
        """
        Returns optimum number of cuts in given dimension
        
        dimennsion: cut dimenstion
        ruleset: rules
        region: region to cut
        """
        # statistics about number of cuts
        nc = {"number": 0, "mean": 0, "max": 0, "empty": 0}
        # stack for future heuristics
        stack = []
        
        value, mask = region.get_data()
        offset = sum([1 for i in bin(mask) if i == "1"])
        domain_size = region.get_domain_size()
        
        for nc["number"] in xrange(1, domain_size - offset):
            cut = []
            for i in xrange(int(math.pow(2, nc["number"]))):
                cut.append(0)
                new_value = value >> (domain_size - offset - nc["number"])
                new_value |= i
                new_value = new_value << (domain_size - offset - nc["number"])
                new_mask = sum([int(math.pow(2, j)) for j in xrange(domain_size - offset - nc["number"], domain_size)])
                # region of current part of new cut
                new_region = MaskedInt(new_value, new_mask, domain_size)
                # number of rules in current part of new cut
                for rule in ruleset.get_rules():
                    try:
                        condition = rule.get_condition(dimension)
                        # PrefixSet
                        if isinstance(condition, PrefixSet):
                            if condition[0].covers(new_region) or new_region.covers(condition[0]):
                                cut[i] += 1
                        # Int
                        else:
                            if new_region.covers(MaskedInt(condition[0], None, domain_size)):
                                cut[i] += 1
                    except KeyError:
                        cut[i] += 1
            nc["mean"] = sum(cut) / math.pow(2, nc["number"])          
            nc["max"] = max(cut)
            nc["empty"] = len([x for x in cut if x == 0])
            
            # here comes comparing function, for now simple comparisons
            if len(stack) != 0:
                if float(stack[0]["empty"]) > 0 and nc["empty"] / float(stack[0]["empty"]) >= 2:
                    return stack[0]["number"]
                if nc["mean"] / float(stack[0]["mean"]) > 0.9:
                    return stack[0]["number"]
            stack.insert(0, copy.deepcopy(nc))
            
        return nc["number"]
    
    def _cut(self, region, cut, number_of_bits):
        """
        Make cut-th cut of region on selected number_of_bits
        
        region: region to cut
        cut: number of cut
        number_of_bits: bits used for cutting
        """        
        if number_of_bits == 0:
            return region
        
        value, mask = region.get_data()
        offset = sum([1 for i in bin(mask) if i == "1"])
        domain_size = region.get_domain_size()
        
        new_value = value >> (domain_size - offset - number_of_bits)
        new_value |= cut
        new_value = new_value << (domain_size - offset - number_of_bits)
        new_mask = sum([int(math.pow(2, i)) for i in xrange(domain_size - offset - number_of_bits, domain_size)])
        
        return MaskedInt(new_value, new_mask, domain_size)
    
    def _rules_for_region(self, region, ruleset):
        """
        Extract rules for specified region
        
        region: region to cover
        ruleset: rules to extract
        """
        new_ruleset = RuleSet()
        
        for rule in ruleset.get_rules():
            acc_cond = 0
            for dimension in region.iterkeys():
                try:
                    condition = rule.get_condition(dimension)
                    # PrefixSet
                    if isinstance(condition, PrefixSet):
                        if condition[0].covers(region[dimension]) or region[dimension].covers(condition[0]):
                            acc_cond += 1
                    # Int
                    else:
                        if region[dimension].covers(MaskedInt(condition[0], None, region[dimension].get_domain_size())):
                            acc_cond += 1
                except KeyError:
                    acc_cond += 1
            if acc_cond == 5:
                new_ruleset.add_rule(rule)
                
        return new_ruleset        

    def _merge_nodes(self):
        """
        Merges same nodes together
        """    
        merged = []
        for i in xrange(len(self._leafs)):
            if self._leafs[i][0] in merged:
                continue
            current_rules = self._leafs[i][0].get_rules()
            for j in xrange(i + 1, len(self._leafs)):
                if current_rules.count_rules() == self._leafs[j][0].get_rules().count_rules():
                    to_merge = True
                    similar_rules = self._leafs[j][0].get_rules()
                    for k in xrange(current_rules.count_rules()):
                        if current_rules.get_rule(k) != similar_rules.get_rule(k):
                            to_merge = False
                            break
                    # Merge nodes
                    if to_merge:
                        # Change parents pointer
                        self._leafs[j][1].get_childs()[self._leafs[j][2]] = self._leafs[i][0]
                        self._statistics["leafs"] -= 1
                        self._statistics["size"] -= Trie.LEAF_SIZE + current_rules.count_rules() * Trie.RULEP_SIZE
                        self._statistics["merged_nodes"] += 1
                        merged.append(self._leafs[j][0])
        # update list of leafs, just in case
        self._leafs = [leaf for leaf in self._leafs if leaf[0] not in merged]

    def create_node(self, ruleset, region, depth = 0):
        """
        Creates node and its childs recursively
        """
        node = Node(region)
        
        # if there are too few rules, we consider node a leaf 
        if ruleset.count_rules() <= self._bucket_size:
            if ruleset.count_rules() == 0:
                node = self._empty_node
                return node, 1
            node.set_leaf()
            node.add_rules(ruleset)
            self._statistics["leafs"] += 1
            self._statistics["size"] += Trie.LEAF_SIZE + Trie.RULEP_SIZE * ruleset.count_rules()
            return node, 1
        
        # choose dimensions to cut
        dim = self._choose_dimensions(ruleset)

        # keeping rules in this node, if they cover whole region
        if self._heuristics["push_rules"]:
            pushed_rules = node.keep_rules(ruleset, dim, self._bucket_size)
            self._statistics["pushed_rules"] += pushed_rules
            self._statistics["size"] += pushed_rules * Trie.RULEP_SIZE
        
        # determine number of cuts in chosen dimensions
        max_childs = self._spfac * math.sqrt(ruleset.count_rules())
        
        # optimal number of cuts in each dimension
        nc = { "srcipv4": 0, "dstipv4": 0, "protocol": 0,
                        "src_port": 0, "dst_port": 0 }
        # number of cuts
        n = 1
        #determine optimal number of cuts (in our case bits)
        for i in dim:
            nc[i] = self._optimum_nc(i, ruleset, region[i])
            n *= math.pow(2, nc[i])
            
        # remove excess cuts
        while n > max_childs:
            for i in dim:
                if nc[i] >= 1:
                    n /= math.pow(2, nc[i])
                    nc[i] -= 1
                    n *= math.pow(2, nc[i])
                if n <= max_childs:
                    break

        # save number of cuts
        node.set_number_of_cuts(nc)
        
        # cutting current region into the pieces
        new_region = {}
        for i in xrange(int(math.pow(2, nc["srcipv4"]))):
            new_region["srcipv4"] = self._cut(region["srcipv4"], i, nc["srcipv4"])            
            for j in xrange(int(math.pow(2, nc["dstipv4"]))):
                new_region["dstipv4"] = self._cut(region["dstipv4"], j, nc["dstipv4"])
                for k in xrange(int(math.pow(2, nc["protocol"]))):
                    new_region["protocol"] = self._cut(region["protocol"], k, nc["protocol"])
                    for l in xrange(int(math.pow(2, nc["src_port"]))):
                        new_region["src_port"] = self._cut(region["src_port"], l, nc["src_port"])
                        for m in xrange(int(math.pow(2, nc["dst_port"]))):
                            new_region["dst_port"] = self._cut(region["dst_port"], m, nc["dst_port"])
                            new_ruleset = self._rules_for_region(new_region, ruleset)
                            child, childs_depth = self.create_node(new_ruleset, new_region)
                            if childs_depth > depth:
                                depth = childs_depth
                            node.add_child(child)
                            # for node merging
                            if child.is_leaf() and child is not self._empty_node:
                                self._leafs.append((child, node, node.count_childs() - 1))
        
        # actualize statistics
        self._statistics["size"] += Trie.NODE_SIZE + n * Trie.CHILDP_SIZE
        self._statistics["n_of_nodes"] += 1
        
        depth += 1
        
        return node, depth
    
    def build_trie(self, ruleset):
        """
        Builds Trie for HyperCuts
        
        ruleset: rules for trie
        """
        self._reset_stats()
        
        # Preprocess ruleset
        ruleset.expand_prefixsets()
        ruleset.remove_covered()
        self._statistics["size"] += ruleset.count_rules() * Trie.RULE_SIZE
        
        # Definition of initial region
        initial_region = { "srcipv4": MaskedInt(0, 0, 32), "dstipv4": MaskedInt(0, 0, 32),
                           "protocol": MaskedInt(0, 0, 8), "src_port": MaskedInt(0, 0, 16),
                           "dst_port": MaskedInt(0, 0, 16) }
        
        self._root, self._statistics["depth"] = self.create_node(ruleset, initial_region)
        
        # Node merging
        if self._heuristics["node_merge"]:
            self._merge_nodes()
            
    def process_packet_header(self, packetheader):
        """
        Preprocess packet header for classification
        
        packetheader: packet to preprocess
        """
        header = {}
        
        if packetheader.get_header("ipv4") is None:
            return None
        else:
            header["srcipv4"] = MaskedInt(packetheader["ipv4"].get_field("src_addr"), 0, 32)
            header["dstipv4"] = MaskedInt(packetheader["ipv4"].get_field("dst_addr"), 0, 32)
            header["protocol"] = MaskedInt(packetheader["ipv4"].get_field("protocol"), 0, 8)
            protocol_field = packetheader["ipv4"].get_field("protocol").toStr().lower()
            header["dst_port"] = MaskedInt(packetheader[protocol_field].get_field("dst_port"), 0, 16)
            header["src_port"] = MaskedInt(packetheader[protocol_field].get_field("src_port"), 0, 16)
            return header
            
    def search_trie(self, packet_header):
        """
        Searchs Trie for coressponding rules
        
        packet_header: packet header to search matching rules
        """
        matching_rules = []
        packet_region = self.process_packet_header(packet_header)
        if packet_region is None:
            return matching_rules
        current_node = self._root
        
        while not current_node.is_leaf():
            child_index = 0
            matching_rules = current_node.match(packet_header)
            cuts = current_node.get_cuts()
            for dimension in Trie.fields_in_order:
                if cuts[dimension] > 0:
                    value, mask = packet_region[dimension].get_data()
                    offset = sum([1 for i in bin(mask) if i == "1"])
                    value &= ~mask
                    value >>= (packet_region[dimension].get_domain_size() - offset - cuts[dimension])
                    child_index <<= cuts[dimension]
                    child_index |= value
            current_node = current_node.get_child(child_index)                    
        
        matching_rules.extend(current_node.match(packet_header))
        
        return matching_rules
    
    def set_heuristic(self, name, activate = True):
        """
        Sets heuristic on (or off if activate is False)
        
        name: name of heuristic, e.g. push_rules
        activate: True to turn on, False to turn off
        """
        self._heuristics[name] = activate
    
    def get_stats(self):
        """
        Returns statistics about Trie
        """
        return self._statistics

#######################################################################
# HyperCuts algorithm class
#######################################################################
class HyperCuts (bclassification.BClassification):
    """
    HyperCuts algorithm for packet classification
    """    
    def __init__(self, spfac = 2, bucket_size = 8):
        """
        Constructor
        
        spfac: space factor
        bucket_size: maximum number of rules that can be stored in node
        """       
        # Rules with wildcards in IP destination or source field
        self._wset = RuleSet()
        
        # Other rules, from which decision tree will be built
        self._rset = RuleSet()
        
        # Root node of RuleSet with wildcards
        self._wtrie = Trie(spfac, bucket_size)
        
        # Root node of regular RuleSet
        self._rtrie = Trie(spfac, bucket_size)
    
    def load_ruleset(self, ruleset):
        """
        Load ruleset and thus create tree for HyperCuts algorithm
        """
        ruleset.expand_prefixsets()
        rules = ruleset.get_rules()
        
        # Store wildcards into different RuleSet
        for rule in rules:
            if rule.get_condition("srcipv4").get_prefixes()[0].is_universal() \
            and rule.get_condition("dstipv4").get_prefixes()[0].is_universal():
                self._wset.add_rule(rule)
            else:
                self._rset.add_rule(rule)
                
        # Build tree for regular rules
        self._rtrie.build_trie(self._rset)
        self._wtrie.build_trie(self._wset)        
        
    def classify(self, packetheader):
        """
        Classifies packet using HyperCuts algorithm
        
        return: list of rules corresponding to packetheader
        """
        rules = self._rtrie.search_trie(packetheader)
        rules.extend(self._wtrie.search_trie(packetheader))
        
        rules.sort(cmp=lambda x,y: cmp(x.get_priority(), y.get_priority()))
        
        return rules
    
    def report_memory(self):
        """
        One can modify properties for memory requirements calculations 
        by changing desired constants in Trie class
        """
        print "Trie for rules with wildcards in ip fields"
        print "================================================"
        print "Number of nodes:  " + str(self._wtrie.get_stats()["n_of_nodes"] + self._wtrie.get_stats()["leafs"])
        print "Depth of Trie:    " + str(self._wtrie.get_stats()["depth"])
        print "Approximate size: " + str(int(self._wtrie.get_stats()["size"])) + " bits"
        if self._wtrie.get_stats()["pushed_rules"] > 0 or self._wtrie.get_stats()["merged_nodes"] > 0:
            print "=Heuristics====================================="
            print "Pushed Rules:     " + str(self._wtrie.get_stats()["pushed_rules"])
            print "Merged Leafs:     " + str(self._wtrie.get_stats()["merged_nodes"])
        print "================================================"
        print "Trie with regular rules"
        print "================================================"
        print "Number of nodes:  " + str(self._rtrie.get_stats()["n_of_nodes"] + self._rtrie.get_stats()["leafs"]) + "(" + str(self._rtrie.get_stats()["leafs"]) + ")"
        print "Depth of Trie:    " + str(self._rtrie.get_stats()["depth"])
        print "Approximate size: " + str(int(self._rtrie.get_stats()["size"])) + " bits"
        if self._rtrie.get_stats()["pushed_rules"] > 0 or self._rtrie.get_stats()["merged_nodes"] > 0:
            print "=Heuristics====================================="
            print "Pushed Rules:     " + str(self._rtrie.get_stats()["pushed_rules"])
            print "Merged Leafs:     " + str(self._rtrie.get_stats()["merged_nodes"])
        print "================================================"
        print "Total size:       " + str(int(self._rtrie.get_stats()["size"] + self._wtrie.get_stats()["size"])) + " bits"
