###############################################################################
#  mspcca.py: Module for experiments with combined classification algorithms
#  Copyright (C) 2011 Brno University of Technology, ANT @ FIT
#  Author(s): Viktor Pus <ipus@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

"""
Module for experiments with combined classification algorithms
"""

import sys
import math
from math import log,ceil

from netbench.classification.rule import *
from netbench.classification.ruleset import *
from netbench.classification.prefix import *
from netbench.classification.prefixset import *
from netbench.classification.common import *

from netbench.classification.algorithms.msca import *
from netbench.classification.algorithms.pcca import *

from netbench.common.packetheader import *
from struct import *
import bclassification

class MSPCCA(bclassification.BClassification):
    """
    Multi-subset Crossproduct + Prefix Coloring Classification Algorithm.
    """
    def __init__(self, subset_count = 3, remove_spoilers = 8, colors = 8):
        self._subset_count = subset_count
        self._cond_names = []
        self._prefixes = []
        self._subset = []
        self._pcca = []
        self._orig_rules = 0
        self._remove_spoilers = remove_spoilers
        self._spoilers = []
        self._colors = colors

    def load_ruleset(self, ruleset):
        self._filters = []
        self._table = []
        self._table_size = []

        ###################### Rules prepsocessing #######################
        # copy rules (this operation will remove duplicates and rules
        #             covered by other rules)
        rules = RuleSet()
        for rule in ruleset.get_rules():
            rules.add_rule(rule, check=True)

        self._rules_count = rules.count_rules()

        print "Number of rules after input filtering: %d" % \
            self._rules_count

        # If there is an universal rule, store it and delete it from rulesets
        self._universalrule = None
        if rules.get_rule(-1).is_universal():
            print "Universal rule detected:",
            rules.get_rule(-1).display()
            self._universalrule = rules.get_rule(-1)
            rules.del_rule(-1)

        ruleset = rules

        ruleset.add_universal_prefixes()
        self._orig_rules = ruleset.count_rules()
        print "Number of rules:",ruleset.count_rules()
        ruleset.expand_prefixsets()
        print "After expanding prefixsets:",ruleset.count_rules()
        if ruleset.count_rules() > self._remove_spoilers:
            print "Removing %d rules, that generates most pseudorules..." % \
                self._remove_spoilers
            s = ruleset.remove_spoilers(self._remove_spoilers)
            print "Removed rules priorities:", [r.get_priority() for r in s]
            self._spoilers = s


        ######################## Division into subsets ######################
        self._subset = subset_NLTSS(ruleset, self._subset_count)
        self._cond_names = ruleset.get_field_names()

        self._headerSize = {}
        for c in self._cond_names:
            fp = cond2field(c)
            if (fp == None):
                raise NotImplemented("Condition \"" + c + "\" is not specified in cond2field function.")
            else:
                fp = fp[0]
                if (fp[2] == "ipv4str" or fp[2] == "int"):
                    self._headerSize[c] = 32
                elif (fp[2] == "int16"):
                    self._headerSize[c] = 16
                elif (fp[2] == "int8"):
                    self._headerSize[c] = 8
                elif (fp[2] == "macstr"):
                    self._headerSize[c] = 48
                else:
                    raise NotImplemented("Size of header type \"" + c + "\" is not specified.")

        for s in range(self._subset_count): # for each subset

            print "###### Subset ", s, " ######"
            ## Prefix coloring, pseudorules and perfect hash tab generation ##
            pcca = PCCA()
            print "### PCCA ###"
            pcca.load_ruleset(self._subset[s], remove_spoilers=0,\
                              colors=self._colors)
            self._pcca.append(pcca)

            # Report pseudorules
            print "Before: ", self._subset[s].count_rules()
            self._subset[s] = pcca._rules
            print "After: " , self._subset[s].count_rules()

            ################## Bloom filters population ##################
            print "### Bloom filter ###"
            # FIXME: Change table size and Bloom filter size
            self._table_size.append(self._subset[s].count_rules())
            self._table.append([ [] for i in range(self._table_size[s]) ])

            # Create Bloom filter
            self._filters.append(BloomFilter(self._table_size[s]))
            self._prefixes.append(self._subset[s].get_prefixes())
            
            # Store each rule into Bloom filter
            for r in self._subset[s].get_rules():
                # Create tuple from all used fields in rule
                t = []
                for k in self._cond_names:
                    if k == 'protocol':
                        rr = ( r.get_condition(k)[0].get_value(), r.get_condition(k)[0].get_value())
                    else:
                        rr = r.get_condition(k)[0].get_range()
                    t.append(rr[0])
                    t.append(rr[1])
                # Store tuple in Bloom filter
                self._filters[s].add(t)
            


    def classify(self, packetheader):
        """
        Classify single packet.
        """
        rule = None
        prefix = [ {} for i in range(self._subset_count)] # list of dicts

        ############ Parse packet ###############
        for c in self._cond_names: # c is condition name
            try:
                # Get value from packet and convert it to integer
                field = extract_field(packetheader, c)
            except FieldNotFoundError:
                raise BadPacketError("Field '%s' not found."%c)
            
        ########### Check spoilers ############
        for r in self._spoilers:
            if (r.match(packetheader)):
                if (rule is None) or \
                   (r.get_priority() < rule.get_priority()):
                    rule = r

        for s in range(self._subset_count): # s is subset number
            ############ LPM, Color processing, Perfect hash ################
            # (Bloom Filters skipped in the model)
            pcca_rule = self._pcca[s].classify(packetheader)
            if len(pcca_rule) > 0:
                if (rule is None) or \
                   (pcca_rule[0].get_priority() < rule.get_priority()):
                    rule = pcca_rule[0]

        if rule:
            return [rule]
        elif self._universalrule is not None:
            return [self._universalrule]
        else:
            return []

    def report_memory(self):
        """
        Print detailed info about algorithm memory requirements.
        """
        print "========== MSPCCA memory report =========="
        rulesCount = sum([s.count_rules() for s in self._subset])
        bloomFilter = sum([f.get_array_size() for f in self._filters])
        headerSize = 127 #sum(self._headerSize[c] for c in self._cond_names) +...
        rulesInMemory = sum([p.count_rules() for p in self._subset]) * headerSize

        prefixSize = 0
        for c in self._cond_names:
            ps = 0
            for s in range(self._subset_count):
                if(self._prefixes[s].has_key(c)):
                    ps = ps + len(self._prefixes[s][c]) * math.ceil(self._headerSize[c]+1)
            prefixSize = prefixSize + ps

        graphsize = 0
        vsize = 16
        minsize = ceil(log(self._orig_rules, 2))
        lpm_added_sum = 0
        for p in self._pcca:
            graphsize += len(p._nodetable)
            lpm_added = {}
            for dim in p._prefix_cond_names: # For each dimension
                lpm_added[dim] = ceil(log(p._colours_info[dim], 2))
                # Store own color
                for odim in p._prefix_cond_names: # For all other dimensions
                    if dim != odim:
                        lpm_added[dim] += p._colours_info[odim] # Store bitmap
                lpm_added[dim] *= len(p._prefixsets[dim])
                # Times number of prefixes
                lpm_added_sum += lpm_added[dim]


        print "Number of all rules:                ", self._orig_rules
        print "Header size (bits):                 ", headerSize
        print "Size of all rules (bits):           ", headerSize*self._orig_rules
        print "Size of Bloom filter table (bits):  ", bloomFilter
        print "Minimal vertex size (bits):         ", minsize
        print "Size of minimal Vertex Table (bits):", graphsize*minsize
        print "Size of 16 b Vertex Table (bits):   ", graphsize*vsize
        print "------------------------------------"
        print "MSCA Size in memory (bloom+rules):  ", rulesInMemory + bloomFilter
        print "MSPCCA Size in memory (16 b vertex):", graphsize*vsize + bloomFilter
        print "MSPCCA Size in memory (min vertex): ", graphsize*minsize + bloomFilter
        print "Added bits to LPM by MSCA:          ", prefixSize
        print "Added bits to LPM by PCCA:          ", lpm_added_sum

