###############################################################################
#  PFCA.py: Module for experiments with the PFCA algorithm
#  Copyright (C) 2011 Brno University of Technology, ANT @ FIT
#  Author(s): Viktor Pus <ipus@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id: dcfl.py 485 2010-09-21 13:54:44Z xbarto11 $

"""
Module for experiments with the Prefix Filtering Classification Algorithm.
"""

import sys
import copy
import math
from math import log, ceil
import bclassification
from netbench.classification.rule import *
from netbench.classification.ruleset import *
from netbench.classification.prefix import *
from netbench.classification.prefixset import *
from netbench.classification.common import *
from netbench.common.packetheader import *

#######################################################################
# PFCA
#######################################################################

class PFCA(bclassification.BClassification):
    """
    Prefix Filtering Classification Algorithm.
    """

    def __init__(self):
        """Constructor"""
        
        # Given ruleset
        self._rules = None
        
        # List of all condition names (dimensions) used in ruleset
        self._prefix_cond_names = []
        
        # List of all prefixes (one PrefixSet for each dimension)
        self._prefixes = {}

        # Universal rule (rule that match all packets)
        self._universalrule = None

        # Generalization rules
        self._grules = []

        # Table of weights of graph nodes
        self._nodetable = []

        # Rules handled by TCAM (generate most of the pseudorules)
        self._spoilers = []

    def load_ruleset(self, ruleset, remove_spoilers = 8):
        """
        Load rules and create generalization rules.
        
        ruleset - Instance of the Ruleset class
        """
        
        # -v---- Common preprocessing ----v-
        orig_count = ruleset.count_rules()
        
        if (orig_count == 0):
            print "ERROR: Ruleset is empty"
            return
        
        # Copy ruleset
        self._rules = RuleSet()
        for rule in ruleset.get_rules():
            if rule.is_universal():
                self._universalrule = rule
            else:
                self._rules.add_rule(rule)
        
        # Filter out duplicates and rules covered by other rules
        self._rules.remove_covered()
        
        print "Number of rules:", self._rules.count_rules(),
        print "(%d loaded, %d filtered out)"%(orig_count,orig_count-self._rules.count_rules())
        
        self._orig_rules_count = self._rules.count_rules()
        
        # Get names of fields used in the ruleset
        #cond_names = self._rules.get_field_names()
        cond_names = ['protocol','src_port','dst_port','srcipv4','dstipv4']
        self._prefix_cond_names = cond_names

        # Add zero-length prefixes to non-specified fields and expand prefixsets 
        self._rules.add_universal_prefixes(cond_names)
        self._rules.expand_prefixsets()
        print "After expanding prefixsets:",self._rules.count_rules()
        print
        
        # -^---- Common preprocessing ----^-

        # remove "worst" rules
        if (self._rules.count_rules() > remove_spoilers):
            self._spoilers = self._rules.remove_spoilers(remove_spoilers)

        print "Number of rules after spoilers removal: %d" % \
            self._rules.count_rules()

        # Perfect hash will search in these rules only
        self._search_rules = self._rules

        # This will probably take a long time
        self._rules.expand_pseudorules()

        # Sort to get pseudorules AFTER rules
        self._rules._rules.sort(key=Rule.get_sorting_priority)
        print "Sorted pseudorules:"
        #self._rules.display()
        print self._rules.count_rules()

        # New ruleset for filtered pseudorules
        nrs = RuleSet()
        
        # Create dictionary with empty list for each dimension
        found_prefixes = {}
        for cond in cond_names:
            found_prefixes[cond] = set()

        # Dimension is forbidden if rule with ANY in that dimension was found
        forbidden_dims = set()

        # Traverse all pseudorules to get GRs
        print "Generating GRs, filtering pseudorules:"
        for r in self._rules.get_rules():
            #r.display()

            # Check if the pseudorule r is not covered by existing GR egr
            filtered = False
            for egr in self._grules:
                if egr.applies_to(r): # If yes, skip
                    #print "Filtered by " + str(gr)
                    filtered = True
                    #print "Filtered by", egr

            # If the pseudorule is covered, then it is deleted
            # (not added to the final list of pseudorules, and not processed
            # by the following code)
            if filtered:
                continue
            else:   # If no, it will appear in the final pseudoruleset
                nrs._rules.append(r) 
                # add_rule does sorting, but rules are already sorted

            has_universal = [] #Rules without univ. prefix aren't interesting
            candidate_dim = []
            candidate_pref = []
            for cond in cond_names: # For each condition
                cond_prefix = r.get_condition(cond).get_prefixes()[0] 

                # Gather universal prefixes in this rule
                if cond_prefix.is_universal():
                    has_universal.append(cond) # Store dims with univ prefs
                else:
                    # Check whether this rule can generate a GR
                    if cond_prefix in found_prefixes[cond] or \
                       cond in forbidden_dims:
                        #print "Prefix was met before: "+ cond+" "+str(cond_prefix)
                        pass # Do nothing
                    else:
                        candidate_dim.append(cond)
                        candidate_pref.append(cond_prefix)

                # Update set of observed prefixes
                found_prefixes[cond].add(cond_prefix)

                # Update the set of dimensions where ANY was observed
                if cond_prefix.is_universal():
                    forbidden_dims.add(cond)


            if len(candidate_dim) > 0 and len(has_universal) > 0:
                # We can create new GR
                gr = GRule()
                index = self._pick_dimension(candidate_dim)
                gr._index = candidate_dim[index]   # Could be more options here
                gr._prefix = candidate_pref[index] # (future work)
                gr._discard = has_universal
                #print r.get_priority()
                #print gr
                self._grules.append(gr)

        # Store final pseudoruleset
        self._rules = nrs

        print "Final rules and pseudorules:"
        #self._rules.display()
        print self._rules.count_rules()
        #print "Generalization rules:"
        #for gr in self._grules:
            #print gr
        
        # Gather prefixsets
        self._prefixsets = self._rules.get_prefixes(self._prefix_cond_names)
        
        self.create_graph()
        print "Number of graph vertices: %d" % len(self._nodetable)


    def _pick_dimension(self, dimlist):
        """Pick steering dimension for GR from the list. Return index."""
        if "srcipv4" in dimlist:
            return dimlist.index("srcipv4")
        if "dstipv4" in dimlist:
            return dimlist.index("dstipv4")
        return 0


    def create_graph(self):
        """
        Creates acyclic graph for perfect hash function
        """

        #----------------------------------------------------------------------
        # Create graph

        # For each node there is an array of its edges - (node, weight) pairs
        # [[(node1, index_of_rule),(node2, index_of_rule2)],...]
        graph = []
        num_of_keys = len(self._rules.get_rules())
        size = 1.5
        hc = 1    # Hash constant
        print "Trying to create acyclic graph"
        last = [0,0,0]
        looped = False
        while (True):
            N = int(num_of_keys * size)
            graph = [ [] for i in range(N)]
            print "Size of graph:", N, "...",

            # For every pseudorule, create a key, compute its hashes h1 and h2
            # and add an edge to the graph
            for r in self._rules.get_rules(): # For each rule
                # Put prefixes in all dimensions into one list
                key = []
                for c in self._prefix_cond_names:  # For all dimensions
                    # Get data of the first (and only) prefix in prefixset
                    #v, m = r.get_condition(c)[0].get_data()
                    #key.append(v + (v >> 6)^hc)
                    #key.append(m)
                    key.append(self._prefixsets[c].get_prefixes().index(r.get_condition(c)[0]))
                key.append(hc)

                # Compute hash functions
                h1 = hash(tuple(key)) % N
                key.append(hc)
                h2 = hash(tuple(key)) % N

                #print key, h1, h2, r._priority
                
                dedge = False
                for (h,p) in graph[h1]:
                    if (h == h2): # This edge already exists in the graph
                        print "Double edge (rules:",p,r.get_priority(),") ->",
                        dedge = True
                if dedge:
                    break
                
                # Add edge between h1 and h2 to graph
                graph[h1].append((h2, r.get_priority()))
                graph[h2].append((h1, r.get_priority()))
            
            else:
                # Determine, if graph is acyclic
                if acyclic(graph):
                    print "OK"
                    break
            
            print "cycle detected"
            hc += 1
            size *= 1.05

        # Set weights of graph nodes in such way that sum of two nodes
        # is weight of edge between them.
        weights = [None for i in range(N)] # Array of graph nodes weights
        stack = []
        for start_node in range(N):
            if (weights[start_node] != None):
                continue
            weights[start_node] = 0
            stack = [start_node]
            while (len(stack) > 0):
                cur_node = stack.pop()
                for n, w in graph[cur_node]:
                    if (weights[n] == None):
                        weights[n] = w - weights[cur_node]
                        stack.append(n)

        self._pseudorules_count = len(self._rules.get_rules())
        self._nodetable = weights
        self._hash_const = hc



    def classify(self, packetheader):
        """
        Classify packet by the algorithm.
        Return the list of all rules that matche given packet.
        
        packetheader: Instance of the PacketHeader class.
        """

        if (self._nodetable == []):
            print "ERROR: No rules are loaded, call load_ruleset() method first()."
            return []

        # Do LPM search on every packet header field and 
        # concatenate results to one key
        key = []
        hc = self._hash_const

        matching_prefixes = {}

        for c in self._prefix_cond_names: # For every dimension (condition)
            try:
                # Get value from packet and convert it to integer
                field = extract_field(packetheader, c)
            except FieldNotFoundError:
                raise BadPacketError("Field '%s' not found."%c)

            matching_prefixes[c] = self._prefixsets[c].match_longest(field)


        # Check if some of the removed spoilers matches the packet
        spoiler = None
        for r in self._spoilers:
            if (r.match(packetheader)):
                if (spoiler is None) or \
                   (r.get_priority() < spoiler.get_priority()):
                    spoiler = r

        # check if there is some prefix in every condition
        # check if every condition has been matched, if some has no match,
        # packet matches either universal rule (if defined), spoiler or no rule
        for cond in matching_prefixes.keys():
            if (matching_prefixes[cond] is None):
                # found condition with no matching prefix
                if spoiler is not None: # Spoiler has higher priority than univ
                    return [spoiler]
                else:
                    if (self._universalrule):
                        return [self._universalrule]

                    else:
                        return []

        # Filter the LPM vector by GRs
        for gr in self._grules:
            gr.apply_to(matching_prefixes)

        ## Perfect hash
        key = []
        # Create key
        for c in self._prefix_cond_names:
            key.append(self._prefixsets[c].get_prefixes().index(matching_prefixes[c]))
        
        # Hash 1
        key.append(self._hash_const)
        h1 = hash(tuple(key)) % len(self._nodetable)
        # Hash 2
        key.append(self._hash_const)
        h2 = hash(tuple(key)) % len(self._nodetable)

        # Perfect hash table access
        rulepriority = self._nodetable[h1] + self._nodetable[h2]

        # Find rule with given priority
        rule = None
        for r in self._search_rules.get_rules():
            if (r.get_priority() == rulepriority):
                rule = r
                break
        
        # Check if found rule really matches the packet
        if rule is not None:
            if not rule.match(packetheader):
                rule = None
        
        if spoiler is not None:
            if rule is not None:
                if spoiler.get_priority() < rule.get_priority():
                    rule = spoiler
            else:
                rule = spoiler
        
        if rule is not None:
            return [rule]
        elif self._universalrule is not None:
            return [self._universalrule]
        else:
            return []

    def report_memory(self):
        """
        Print detailed info about algorithm memory requirements.
        """
        
        print "========== PFCA memory report =========="
        print "Number of spoilers: %d" % len(self._spoilers)
        print "Number of prefixes in each dimension (no spoilers):"
        for cond, prefixsets in self._prefixsets.iteritems():
            print cond + ": %d" % len(prefixsets),
        print

        print "Number of Generalization Rules in each dimension:"
        gd = {}
        for gr in self._grules:
            if gr._index in gd.keys():
                gd[gr._index] = gd[gr._index]+1
            else:
                gd[gr._index] = 1
        print gd

        bits_per_gr = 32
        print "GR table size (for 32 bit per GR):",\
            len(self._grules) * bits_per_gr, "bits"

        print "Perfect hash table: %d items" % len(self._nodetable)
        print 'Perfect hash table: {0} bits supposing one vertex takes {1} bits'\
            .format(len(self._nodetable) * ceil(log(self._orig_rules_count, 2)),
            ceil(log(self._orig_rules_count, 2)))

        print 'Perfect hash table: {0} bits supposing one vertex takes 16 bits'\
            .format(len(self._nodetable) * 16)

        bits_per_rule = 64
        print 'Rule table supposing one rule occupies {0} bits: {1} bits'\
            .format(bits_per_rule, bits_per_rule*self._orig_rules_count)

#######################################################################
# Generalization Rule
#######################################################################

class GRule:
    """
    Generalization Rule for the PFCA.
    """

    def __init__(self):
        """Constructor"""

        # Name of the steering dimension
        self._index = ""

        # Prefix condition
        self._prefix = None

        # List of dimension names to be substituted by ANY
        self._discard = []

    def __str__(self):
        """Printing method"""

        return "GR:If " + self._index + "=" + str(self._prefix) + \
            " then set LPM" + str(self._discard) + "=ANY."

    def applies_to(self, rule):
        """Return true if this GR can be applied to the given rule"""

        # If the condition applies ...
        if (rule.get_condition(self._index).get_prefixes()[0] == self._prefix):
            for dim in self._discard:
                # And if the rule has some non-ANY cond in the substitued field
                if not rule.get_condition(dim).is_universal():
                    # Then it applies
                    return True

        # Else it doesn't
        return False

    def apply_to(self, lpmvect):
        """
        Generalize the LPM vector. 
        Do not modify if GR does not apply
        lpmvect: dictionary of prefixes
        """
        # If the condition applies ...
        if (lpmvect[self._index] == self._prefix):
            for dim in self._discard:
                # and if there is non-ANY cond in the substitued field
                if not lpmvect[dim].is_universal():
                    # Substitute by ANY
                    lpmvect[dim].set_mask(0)


#######################################################################
# Other stuff
#######################################################################

def acyclic(graph):
    """
    Returns True, if given graph is acyclic.

    graph: It must be undirected graph given as list of lists of connected
    nodes and edge weights (weights are ignored):
    [[(node1,weight1),(node2,weight2),...],...]
    """

    visits = [0 for i in range(len(graph))] # Mark all nodes as not yet visited
    next_nodes = [] # List of (next node to visit, previous node)

    for start_node in range(len(graph)):
        if (visits[start_node] > 0):
            continue
        next_nodes = [(start_node, None)]
        while (len(next_nodes) > 0):
            cur_node, prev_node = next_nodes.pop(0)
            if (visits[cur_node] > 0): # If node has been visited already
                return False           # there must be a cycle
            visits[cur_node] += 1 # Count this visit
            # For all nodes connected to current node
            for n, w in graph[cur_node]:
                if (n != prev_node):
                    next_nodes.append((n,cur_node)) # Add node to list of nodes to visit

    return True # All nodes has been visited and no cycle has been detected - graph is acyclic
