#!/usr/bin/env python

import string
import time
import struct

import packetheader
import protoheader
import netbench.common.headerfields as hf # There must be full path to avoid problems when comparing types

class Packet(object):
    """Class for Packet storage and packet header parsing.
    
    Parsed packet can contain this fields:
      ProtoHeader     fields
      -----------  ------------
       ethernet    typelength
                   src_mac
                   dst_mac
       vlan        type
       ipv4        version
                   header_len
                   tos
                   total_len
                   id
                   flags
                   fragment_offset
                   ttl
                   protocol
                   checksum
                   src_addr
                   dst_addr
                   options
       tcp         src_port
                   dst_port
                   data_offset
                   tcp_flags
                   tcp_window
       udp         src_port
                   dst_port
                   udp_len
                   udp_checksum
    """

    def __init__(self):
        self.parsed = False
        self.pckthdr = packetheader.PacketHeader()
        self.raw_data = []
        self.payload_offset = []
        tmphdr = protoheader.ProtoHeader("info")
        self.pckthdr.push_header(tmphdr)
    ""
    def dumphex(self):
        """Print out raw data of a packet in hexadecimal format."""
        s = self.raw_data
        bytes = map(lambda x: '%.2x' % x, map(ord, s))
        for i in xrange(0,len(bytes)/16):
            print '        %s' % string.join(bytes[i*16:(i+1)*16],' ')
        print '        %s' % string.join(bytes[(i+1)*16:],' ')

    def decode_l2(self, s, l2header):
        """Dissector for Ethernet header"""
        next_header = None
        try:
            x = struct.unpack('!HI', s[0:6])
            x = (x[0] << 32) + x[1]
            l2header.set_field('dst_mac', hf.Eth_DstMac(x))
            x = struct.unpack('!HI', s[6:12])
            x = (x[0] << 32) + x[1]
            l2header.set_field('src_mac', hf.Eth_SrcMac(x))
            x = struct.unpack('!H', s[12:14])[0]
            l2header.set_field('typelength', hf.Eth_TypeLength(x))
            if x == 0x8100:
                next_header = 'vlan'
            elif x == 0x800:
                next_header = 'ipv4'
            return 14, next_header
        except IndexError: # Header is not complete
            return len(s), next_header 

    def decode_ipv4(self, s, ipheader):
        """Dissector for IPv4 header"""
        next_header = None # string of next_header
        loaded = 0 # number of read bytes
        try:
            ipheader.set_field('version', hf.IP4_Version((ord(s[0]) & 0xf0) >> 4))
            ipheader.set_field('header_len', hf.IP4_HeaderLen(ord(s[0]) & 0x0f))
            ipheader.set_field('tos', hf.IP4_ToS(ord(s[1])))
            ipheader.set_field('total_len', hf.IP4_TotalLen(struct.unpack('!H',s[2:4])[0]))
            ipheader.set_field('id', hf.IP4_ID(struct.unpack('!H',s[4:6])[0]))
            ipheader.set_field('flags', hf.IP4_Flags((ord(s[6]) & 0xe0) >> 5))
            ipheader.set_field('fragment_offset', hf.IP4_FragmentOffset(struct.unpack('!H',s[6:8])[0]) & 0x1fff)
            ipheader.set_field('ttl', hf.IP4_TTL(ord(s[8])))
            ipheader.set_field('protocol', hf.IP4_Protocol(ord(s[9])))
            try:
                next_header = hf.IP4_Protocol.names[ord(s[9])].lower()
            except IndexError:
                next_header = None
            ipheader.set_field('checksum', hf.IP4_Checksum(struct.unpack('!H',s[10:12])[0]))
            ipheader.set_field('src_addr', hf.IP4_SrcAddress(struct.unpack('!I',s[12:16])[0]))
            ipheader.set_field('dst_addr', hf.IP4_DstAddress(struct.unpack('!I',s[16:20])[0]))
            loaded = 20
            if ipheader.get_field('header_len')>5:
                length = ipheader.get_field('header_len')-5 # length of otpions (in 4-byte words)
                opt_data = struct.unpack('!'+'I'*length,s[20:20+4*length]) # get data as tuple of 4-byte integers
                loaded += 4*length
                opt_data = reduce(lambda x,y: (x<<32)+y, opt_data) # convert to one long integer
                ipheader.set_field('options', hf.IP4_Options(opt_data))
            else:
                ipheader.set_field('options', None)
        
        except IndexError: # Header is not complete
            loaded = len(s)
        
        return loaded, next_header
    
    def decode_vlan(self, s, vlanheader):
        """Dissector for 802.1Q Virtual LAN header. It currently decodes next header type only."""
        if s < 4:
            return 0, None
        typenum = struct.unpack('!H',s[2:4])[0]
        vlanheader.set_field('type', hf.VLAN_Type(typenum))
        if typenum == 0x800:
            next_header = 'ipv4'
        else:
            next_header = None
        return 4, next_header
        

    def decode_udp(self, s, udpheader):
        """Dissector for UDP header"""
        try:
            processed = 0
            if (len(s) >= 4):
                udpheader.set_field('src_port', hf.UDP_SrcPort(struct.unpack('!H',s[0:2])[0]))
                udpheader.set_field('dst_port', hf.UDP_DstPort(struct.unpack('!H',s[2:4])[0]))
                processed = 4
            
            if (len(s) >= 8):
                udpheader.set_field('udp_len',hf.UDP_Length(struct.unpack('!H',s[4:6])[0]))
                udpheader.set_field('udp_checksum',hf.UDP_Checksum(struct.unpack('!H',s[6:8])[0]))
            return processed, None
        except IndexError:
            return len(s), None

    def decode_tcp(self, s, tcpheader):
        """Dissector for TCP header"""
        try:
            if (len(s) <= 16):
                return 0, None
 
            tcpheader.set_field('src_port', hf.TCP_SrcPort(struct.unpack('!H',s[0:2])[0]))
            tcpheader.set_field('dst_port', hf.TCP_DstPort(struct.unpack('!H',s[2:4])[0]))
            tcpheader.set_field('data_offset', hf.TCP_DataOffset((ord(s[12]) & 0xf0) >> 4))
            tcpheader.set_field('tcp_flags', hf.TCP_Flags(ord(s[13]) & 0x3f))
            tcpheader.set_field('tcp_window', hf.TCP_Window(struct.unpack('!H',s[14:16])[0]))

            return 4*tcpheader.get_field('data_offset'), None
        except IndexError:
            return len(s), None


    def parse(self, level=10):
        """Parse packet header up to a specified level.

           level
               Specifies the maximum level a packet header should be parsed to.
        """
        decoders = {
            "ethernet" : self.decode_l2,
            "ipv4" : self.decode_ipv4,
            "vlan" : self.decode_vlan,
            "udp"  : self.decode_udp,
            "tcp"  : self.decode_tcp,
        }
        
        self.parsed = True
        endofparsing = 0
        next_header = "ethernet"
        while next_header and level > 0:
            self.payload_offset.append(endofparsing)
            tmphdr = protoheader.ProtoHeader(next_header)
            try:
                length, next_header = decoders[next_header](self.raw_data[endofparsing:], tmphdr)
            except KeyError:
                break
            endofparsing += length
            self.payload_offset.append(endofparsing)
            self.pckthdr.push_header(tmphdr)
            level -= 1
        
        return self.pckthdr


    def get_payload(self,type):
        """Returns payload of a packet belonging to a given protocol layer."""
        if not self.parsed:
            self.parse()
        level = self.pckthdr.get_headers_type().index(type)
        return self.raw_data[self.payload_offset[level]:]

    def get_packet_header(self):
        """Returns packet header"""
        if not self.parsed:
            self.parse()
        return self.pckthdr



