#! /usr/bin/env python

###############################################################################
#  pcapfile.py: Pcap files parser
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Vaclav Bartos <xbarto11@stud.fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id: pcapfile.py 1565 2012-01-02 08:54:34Z izadnik $

from struct import unpack
from packet import Packet


class NotOpenedError(Exception):
    def __str__():
        return "No file is opened."
class NotIndexedError(Exception):
    def __str__():
        return "File is not indexed. Call index() method first."

class PcapFile:
    """
    Simple parser of pcap files. Extracts raw packet data from a given file.
    
    You can open file by passing its name to constructor or by calling open() 
    method.
    
    For sequential reading of packets, call next_packet() method in a loop. 
    It always returns next packet if the file (an instance of Packet class),
    or None if there aren't any packets left.
    
    To enable random access to packets, you must call index() method first.
    Than you can access packets in the file using indexes (ie. pcap[i]).
    Note that file is not loaded to memory by calling index(), only positions
    of packets are saved, so every access means reading from disk.
    
    Writing of pcap files is not implemented (yet).
    """
    
    def __init__(self, filename = None):
        """
        Contructor of a PcapFile object. If you pass name of a file, it's
        opened instantly.
        """
        # Public attributes
        self.filename = None
        self.indexed = None
        self.num_of_packets = None
        
        self._file = None
        self._index = None
        if filename is not None:
            self.open(filename)
        
    
    def open(self, filename, mode = "r"):
        """
        Opens a given file in a given mode.
        Default (and only currently supported) mode is 'r', which means reading.
        """
        if mode == "w":
            raise NotImplemented("Writing of Pcap files not implemented yet.")
        elif mode != "r":
            raise ValueError("Unknown mode: '"+mode+"'")
        
        if self._file:
            self._file.close()
        
        f = open(filename, 'rb')
        self._file = f
        
        hdr = f.read(24) # Read global header
        
        # Read magic number and determine its endianness
        magic_number = unpack('<I',hdr[0:4])[0] # read as little-endian
        if (magic_number == 0xa1b2c3d4):
            e = '<' # it's ok, use little-endian
        elif (magic_number == 0xd4c3b2a1):
            e = '>' # it's swapped, use big-endian
        else:
            self.close()
            raise IOError("Not valid PCAP file.")
        
        # Parse global header
        self._file_hdr = unpack(e+'IHHiIII',hdr)
        self._e = e
        
        self.filename = filename
        self.indexed = False
    
    def close(self):
        """
        Closes the file.
        """
        if self._file:
            self._file.close()
        self._file = None
        self.filename = None
    
    def __getattr__(self, attr):
        if attr == "opened":
            return self._file != None
        if attr == "closed":
            return self._file == None
    
    def next_packet(self):
        """
        Returns next packet in the file. Returns None when EOF is reached.
        """
        if not self._file:
            raise NotOpenedError()
        
        # Read and parse header of packet
        phdr = self._file.read(16)
        if len(phdr) < 16:
            return None
        
        ts_sec, ts_usec, incl_len, orig_len = unpack(self._e+'IIII',phdr)
        
        # Read raw data of packet
        data = self._file.read(incl_len)
        
        # Create Packet instance and fill it with data
        pkt = Packet()
        #pkt.pckthdr['info'].set_field("timestamp", ts_sec*1000000+ts_usec)
        pkt.pckthdr['info'].set_field("timestamp", ts_sec+ts_usec/1000000.0)
        pkt.pckthdr['info'].set_field("len", incl_len)
        pkt.pckthdr['info'].set_field("orig_len", orig_len)
        pkt.raw_data = data

        return pkt
    
    def __iter__(self):
        """
        Returns itself. It allows to use this class as iterator.
        """
        return self
    
    def next(self):
        """
        Returns next packet in the file. Raises StopIteration exception when
        EOF is reached.
        This method is called by Python to iterate in "for" loop.
        """
        np = self.next_packet()
        if np is None:
            raise StopIteration()
        return np
    
    def rewind(self):
        """
        Rewinds file to the begining so next read returns the first packet.
        """
        if not self._file:
            raise NotOpenedError()
        
        self._file.seek(24)
    
    
    def index(self):
        """
        Reads whole file and save positions of all packets. Also counts number 
        of packets in the file.
        """
        if not self._file:
            raise NotOpenedError()
        self._indexes = []
        
        while True:
            phdr = self._file.read(16)
            if len(phdr) < 16:
                break
            self._indexes.append(self._file.tell() - 16)
            self._file.seek(unpack(self._e+'I',phdr[8:12])[0], 1)
        
        self.indexed = True
        
    def __len__(self):
        """
        Returns number of packets in the file.
        It's needed to call index() first.
        """
        if not self._file:
            raise NotOpenedError()
        if not self.indexed:
            raise NotIndexedError()
        
        return len(self._indexes)
        
    def __getitem__(self, i):
        """
        Returns i-th packet in the file.
        It's needed to call index() first.
        
        Raises IndexError if i >= number of packets in the file.
        """
        if not self._file:
            raise NotOpenedError()
        if not self.indexed:
            raise NotIndexedError()

        self._file.seek(self._indexes[i])
        return self.next_packet()
