###############################################################################
#  hash_tbm.py: Module with Hash Tree Bitmap LPM algorithm
#  Copyright (C) 2011 Brno University of Technology, ANT @ FIT
#  Author(s): Lukas Kekely <xkekel00@stud.fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id: hash_tbm.py 1553 2011-12-22 18:06:39Z xsedla76 $

"""
Module with Hash Tree Bitmap LPM algorithm
"""

import sys
import blpm
from netbench.classification import prefixset
from netbench.classification import maskedint
from netbench.lpm.algorithms import treebitmap


class HashTBM(blpm.BLPM):
    """
    Hash Tree Bitmap LPM algorithm for prefix lookups
    Prefixes from prefixset will be divided into classes by their length
    Length span for classes can be configured
    TBM algorithm stride can also be configured (max height of TBM tree will be roundup(span/stride))
    """

    def __init__(self, lengths=[0], stride = 4):
        """
        Constructor
            lenghts = lengths of used hash tables (length 0 added if not present)
            stride  = stride for TBM algorithm 
        """
        blpm.BLPM.__init__(self)
        self.stride = stride
        self.tables = []
        # hash tables ordered from longest to shortest
        lengths.sort()
        lengths.reverse()
        # length 0 added if missing (to be able to represent all lengths)
        if lengths[-1] != 0:
            lengths.append(0)
        for l in lengths:
            self.tables.append(Table(l))

    def report_memory(self):
        """
        Return detailed info about algorithm memory requirements.
        """
        report = {}
        suma = 0
        trees = 0
        nodes = 0
        report['tables'] = []
        for t in self.tables:
            pom = len(t.tab)
            suma += pom
            pomn = 0
            pomt = 0
            for r in t.tab:
                if t.tab[r].mark:
                    pomt += 1
                    pomn += t.tab[r].ID.tree.nodes
            nodes += pomn
            trees += pomt
            report['tables'].append({'length': t.len, 'starts': pom, 'nodes': pomn, 'trees': pomt})
        report['starts'] = suma
        report['trees'] = trees
        report['nodes'] = nodes

        return report      

    def load_prefixset(self, prefixset):
        """
        Load prefixes and generate all necessary data structures. 
        """
        self.prefixset = prefixset
        for p in self.prefixset.get_prefixes():
            # binary representation of prefix
            key = ""
            if p.get_length():
                key = bin(p.get_value() >> (p.get_domain_size() - p.get_length()))[2:]
                key = (p.get_length() - len(key))*"0" + key

            # go throught hash tables 'till first shorter than prefix length found
            for t in self.tables:
                if t.len <= p.get_length():
                    # divide prefix into START and END piece
                    start=key[0:t.len]
                    end  =key[t.len:]
                    # add prefix into data structures depending on situation 

                    # record with same prefix start doesn't exists
                    if start not in t.tab:
                        pom = Record()
                        # new prefix continues into TBM
                        if end:
                            pom.ID   = Tree_record(treebitmap.Tree(self.stride))
                            pom.ID.tree.add_prefix(end,p)
                            pom.mark = True
                        # new prefix ends in hash table (prefix length equals table length)
                        else:
                            pom.ID   = p
                            pom.mark = False
                        t.tab[start]= pom

                    # record with same prefix start already exists
                    else:
                        # existing record already marked (already continuos)
                        if t.tab[start].mark:
                            t.tab[start].ID.tree.add_prefix(end,p)
                        # unmarked record => change to marked and add Tree
                        else:
                            if end:
                                pom = t.tab[start].ID
                                t.tab[start].ID = Tree_record(treebitmap.Tree(self.stride))
                                t.tab[start].ID.tree.add_prefix("",pom)
                                t.tab[start].ID.tree.add_prefix(end,p)
                                t.tab[start].mark = True
                            else:
                                t.tab[start].ID = p
                                
                    break

        # prepare default LPM values for continuous prefixes
        # only if end in TBM with key "" doesn't exist
        count = len(self.tables)-1
        t = -2
        # traverse tables from second shortest (there isn't shorter LPM for length 0) to longest
        while count:
            for r in self.tables[t].tab:
                # only records with marks (continuos)
                if self.tables[t].tab[r].mark:
                    # and only if start isn't prefix by it self (end "" doesn't exist)
                    if not self.tables[t].tab[r].ID.tree.root.prefix_bitmap[0]:
                        # basic lookup, but only in directly shorter table
                        start = r[0:self.tables[t+1].len]
                        end   = r[self.tables[t+1].len:]
                        if start in self.tables[t+1].tab:
                            if self.tables[t+1].tab[start].mark:
                                self.tables[t].tab[r].ID.default_LPM = TBM_lookup(self.tables[t+1].tab[start].ID, end)
                            else:
                                self.tables[t].tab[r].ID.default_LPM = self.tables[t+1].tab[start].ID
            count-=1
            t-=1
            
    def lookup(self, ip):
        """
        Longest prefix match lookup for IP address.
            NOTE: only longest prefix (or None) is returned !!!
        """
        # transform IP into binary notation
        key = bin(ip)[2:]
        key = (2**maskedint.log2(len(key)) - len(key))*"0" + key
        # traverse hash tables from longest and stop at first matching start
        # NOTE: in HW implementation, by the use of paralelism, record from each table is checked at the same time in one step
        #       then longest from found records is chosen and used
        for t in self.tables:
            if len(key) >= t.len:
                # divide IP adress into start and end
                start=key[0:t.len]
                end  =key[t.len:]
                # search start in hash table
                if start in t.tab:
                    if t.tab[start].mark:
                        # search end in TBM tree if TBM tree exists 
                        return TBM_lookup(t.tab[start].ID, end)
                    else:
                        return t.tab[start].ID
        return None
                    

    def display(self):
        """Display the structure of the tree"""
        for t in self.tables:
            print 'TABLE:', t.len, '###########################################'
            for r in t.tab:
                if t.tab[r].mark:
                    print 'start:', '"'+r+'"' , '-> TBM tree (default LPM: ' + str(t.tab[r].ID.default_LPM) + ')'
                    TBM_display(t.tab[r].ID.tree.root)
                else:
                    print 'start:', '"'+r+'"' , '-> LPM:' ,t.tab[r].ID
            if not len(t.tab):
                print '< empty >'
            print



class Table(object):
    """
    Class representing hash table.
        len = length of table (prefix starts in table)
        tab = hash table data
    """
    def __init__(self, length):
        self.len = length
        self.tab = {}


class Record(object):
    """
    Class representing one field in hash table.
        ID   = prefix or TBM tree, depending on mark
        mark = is prefix continuing into TBM tree?
    """
    def __init__(self):
        self.ID   = False
        self.mark = False

class Tree_record(object):
    """
    Class encapsulating TBM tree structure, just adding default LPM used in case that no match found within TBM.
        tree        = TBM tree
        default_LPM = result of lookup in case that no match is found in TBM
    """
    def __init__(self, tree):
        self.tree        = tree
        self.default_LPM = None

# BASED ON FUNCTION display FROM treebitmap
def TBM_display(node):
        """Display the structure of the tree"""
        
        # traverse the tree in pre-order and print the stucture of the tree
        if node:
            print node.depth *"   ", "   ---- NODE --------------------"
            print node.depth *"   ", "   - prefix bitmap ", node.prefix_bitmap
            print node.depth *"   ", "   - children bitmap ", node.child_bitmap
            print node.depth *"   ", "   - prefixes in node:",
            for prefix in node.prefixes:
                print prefix[1],
            print "\n",node.depth *"   ", "   ------------------------------"
            for i in node.children:
                TBM_display(i[1])

# BASED ON FUNCTION lookup FROM treebitmap
def TBM_lookup(tree, key):
        """
        Lookup longest prefix that match ip.
        Return only the longest matched prefix. 
        If ip matches no prefix, default_LPM is returned.
         
        ip: value of the ip address according to maskedint.py
        """
        list = tree.default_LPM
        current_node = tree.tree.root
        seq = ""
        # the key will be manipulated in parts. Size of the part depends on the stride
        for i in range(0, len(key), tree.tree.stride):
            # one part of the key with the size of the stride
            seq = key[i:i+tree.tree.stride]
            # it the part has the full length, it will be continued with the next part
            if len(seq) == tree.tree.stride:
                # number to the children-bitmap
                number = int(seq,2)
                # there is a node in the desired way
                if current_node.child_bitmap[number] == 1:
                    # store all valid prefixes to the list
                    for j in range(0, len(seq)):
                        tmpseq = seq[0:j]
                        if len(tmpseq) == 0:
                            position = 0
                        else:
                            position = 2**j - 1 + int(tmpseq,2)
                        if current_node.prefix_bitmap[position] == 1:
                            count = 0
                            for k in range(0, position):
                                if current_node.prefix_bitmap[k] == 1:
                                    count += 1
                            list = current_node.prefixes[count][1]
                # there is no more node in the desired way                 
                else:
                    break
                # compute the position of the next Node
                count = 0
                for j in range(0,number):
                    if current_node.child_bitmap[j] == 1:
                        count += 1
                seq = ""
                # move on the next Node
                current_node = current_node.children[count][1]
        
        # we handled as much of the key as possible
        # in the current_node is now the last node of the tree in desired way
        
        # check the last node and store all valid prefixes to the list
        if len(seq) == 0:
            if current_node.prefix_bitmap[0] == 1:
                list = current_node.prefixes[0][1]
        else:
            for i in range(0, len(seq)):
                tmpseq = seq[0:i]
                if len(tmpseq) == 0:
                    position = 0
                else:
                    position = 2**i - 1 + int(tmpseq,2)
                if current_node.prefix_bitmap[position] == 1:
                    count = 0
                    for j in range(0,position):
                        if current_node.prefix_bitmap[j] == 1:
                            count += 1
                    list = current_node.prefixes[count][1]
        
        # reverse list to be sorted in descentant order    
        return list

        

        
