###############################################################################
#  lctrie.py: Module for LC Trie algorithm
#  Copyright (C) 2011 Brno University of Technology, ANT @ FIT
#  Author(s): Frantisek Sedlar <xsedla76@stud.fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

"""
Module for LC-Trie algorithm
"""

import sys
import blpm
from netbench.classification import prefixset
from netbench.classification import maskedint

class LCTrie(blpm.BLPM):
    """
    LC-Trie LPM algorithm for lookup prefixes
    """

    def __init__(self):
        """Constructor"""
        blpm.BLPM.__init__(self)


    def report_memory(self):
        """
        Return detailed info about algorithm memory requirements.
        """
        report = {
                  'prefixes': self.trie.prefixes,      # prefixes in trie
                  'nodes': self.trie.nodes,            # nodes in trie
                  'internal_nodes': self.trie.nodes - self.trie.edge_nodes,   # internal nodes in trie
                  'edge_nodes': self.trie.edge_nodes,  # edge nodes in trie
                  'pointers': self.trie.nodes,         # pointers in trie
                  'depth': self.trie.max_depth,        # max depth of trie
                  'prefix_table_len': len(self.prefix_table.table),   # prefixes in prefix_table
                  'max_branch': self.trie.max_branch,  # maximum branch value
                  'max_skip': self.trie.max_skip       # maximum skip value
                 }
        return report

    
    def build_prefix_table(self):
        """
        Finds prefixes of other prefixes, remove them from prefixset
        and add them into the prefix table, because these are not
        going to be in the trie.        
        """

        # Start with the last and the last but one prefix
        if self.prefixset.__len__() < 2:
            return
        i = self.prefixset.__len__() - 2
        last_prefix = self.prefixset.get_prefixes()[i+1]

        while i >= 0:
            current_prefix = self.prefixset.get_prefixes()[i]
            if last_prefix.get_binary()[0:len(current_prefix.get_binary())] == current_prefix.get_binary():
                # found prefix of other prefix, move it
                self.prefix_table.append(current_prefix)
                self.prefixset.remove_prefix(current_prefix)
                i-=1
            else:
                i -= 1
                last_prefix = self.prefixset.get_prefixes()[i+1]
            

    
    def load_prefixset(self, prefixset):
        """
        Load prefixes and generate all necessary data structures. 
        
        Note that it changes prefixset - it removes prefixes, which are prefixes of other (longer) prefixes
        """

        self.prefixset = prefixset

        # we need sorted prefixset to build the tree
        self.prefixset.sortbinary()

        #print sorted prefixes
        # for prefix in prefixset.get_prefixes():
        #     print(prefix)


        # remove prefixes of other prefixes and build prefix table
        self.prefix_table = Prefix_table()
        self.build_prefix_table()

        # building trie
        self.trie = Trie(prefixset)
        self.trie.build(0, self.prefixset.__len__(), 0, 0)
        
        #print what we have done
        # self.prefix_table.get_table()
        # self.trie.get_table()
        # self.trie.display()
        



    def lookup(self, ip):
        """
        Lookup prefixes that match ip.
        Return the list of matched prefixes. 
        If ip matches no prefix, the list will be empty.
        
        ip: value of the ip address according to maskedint.py
        """
	# does prefixset contain any prefixes?
	if len(self.prefixset.get_prefixes())==0:
	  return []

        list = []
        node = self.trie.trie[0]
        pos = node.skip
        branch = node.branch
        adr = node.adr
        while (branch != 0):
            node = self.trie.trie[adr + self.trie.extract(pos, branch, bin(ip)[2:])]
            pos = pos + branch + node.skip
            branch = node.branch
            adr = node.adr

        # We found edge node. Because many bits could be skipped, we need to check
        if (self.prefixset.get_prefixes()[node.adr].match(ip)):
            list.append(self.prefixset.get_prefixes()[node.adr])

        # There may be other prefixes of the ip in the prefix table
        for current_prefix in self.prefix_table.table:
            if current_prefix.prefix.match(ip):
                list.append(current_prefix.prefix)

        return list


    def display(self):
        """Display the structure of the tree"""
        self.trie.display(0,0,0,0)



class Prefix_table(object):
    """
    List of prefixes, which are prefixes of other (longer) prefixes. These were removed from the prefixset
    """

    def __init__(self):
        self.table = []

    def append(self, prefix):
        """
        Add a prefix to the prefix table
        """
        self.table.append(Prefix_table_entry(prefix))



class Trie(object):
    """
    Trie consists of Nodes and represents the whole tree in which will be
    looking up the prefixes
    """
    def __init__(self, prefixset):
        self.trie = []
        self.prefixes = 0
        self.max_depth = 1
        self.nodes = 1
        self.edge_nodes = 0
        self.trie.append(Node("?", "?", "?"))
        self.prefixset = prefixset
        self.actual_depth = 0
        self.max_branch = 0
        self.max_skip = 0
        #self.pos = 1

    def display(self, actual_node, depth, branch, skip):
        """
        Display the structure of the tree
    
        Displays the trie starting at the actual node to the depth passed by argument.
        Branch represents how many children the actual node has and skip represents the skip value of the actual node.
        """
        done = []

        if self.nodes == 1:
            # trie has just the root
            print "ROOT"
            return

        for i in range(2**branch):
            # we have to work with all children in cycle
            i = (  str( bin(i)[2:] )  )
            if len(i) < branch:
                i = (branch - len(i)) * "0" + i
            node = self.trie[actual_node]
            if node.branch==0:
                # this is edge node, print prefix number
                print (4*depth) * " " + "-" + skip*"*" + str(i) + " (P" + str(node.adr) + ")"
                actual_node += 1
            else:
                # this is not edge node
                if branch == 0:
                    # this is root, let us print it
                    print "ROOT"
                else:
                    # this internal node
                    print (4*depth) * " " + "-" + skip*"*" + str(i)
                # because this is not an edge node, we have to work up it's children
                self.display(self.trie[actual_node].adr, depth+1, node.branch, node.skip)
                actual_node += 1
                
                
                

    def checkBranch(self, branch, pre, first, n, skip):
        """
        Check if the "n" prefixes in prefixset, starting at prefix "first"
        with skip value "skip" can have branching factor "2^branch".
        "pre" bits of these prefixes are not considered.

        Return "branch" if it is possible, otherwise return 0
        """

        prefixy = []
        i = first
        while i < first + n:
            current_prefix = self.prefixset.get_prefixes()[i].get_binary()
            if (current_prefix[pre:branch+pre] not in prefixy) and (len(current_prefix) >= branch+pre):
                prefixy.append(current_prefix[pre:branch+pre])
            i+=1

        if len(prefixy) == (2**branch):
            return branch
        else:
            return 0


    def computeBranch(self, pre, first, n, skip):
        """
        Compute branching factor for the "n" prefixes in prefixset, starting at prefix "first"
        with skip value "skip". "pre" bits of these prefixes are not considered.

        Return "branch" (branching factor is 2^branch)
        """

        branch = 1
        while (self.checkBranch(branch+1, pre, first, n, skip)):
            branch+=1
        if branch > self.max_branch:
            self.max_branch = branch
        return branch



    def computeSkip(self, pre, first, n):
        """
        Compute skip value for the "n" prefixes in prefixset, starting at prefix "first"
        "pre" bits of these prefixes are not considered.

        Return skip value
        """

        first_prefix = self.prefixset.get_prefixes()[first].get_binary()
        last_prefix = self.prefixset.get_prefixes()[first+n-1].get_binary()

        i = pre
        while i < len(first_prefix) and i < len(last_prefix) and (first_prefix[i] == last_prefix[i]):
            i+=1

        if i-pre > self.max_skip:
            self.max_skip = i-pre
        return i-pre


    def extract(self, pos, n, vector):
        """
        Extract "n" bits from prefix "vector", starting at position "pos"
        Return extracted string
        """
        text = vector[pos:n+pos]
        if text == "":
            return 99
        else:
            return (int(vector[pos:n+pos], 2))
        

    def build(self, first, n, pre, pos):
        """
        Used to build LC-Trie. Build part of trie - use "n" prefixes in prefixset,
        starting at prefix "first", "pre" bits of these prefixes are not considered.
        "pos" is index of first free position in the table representing the trie

        This is called recursively
        """
        if self.prefixset.__len__() < 1:
            # no prefixes in prefixset
            return

        first_prefix = self.prefixset.get_prefixes()[first].get_binary()

        skip = self.computeSkip(pre, first, n)

        if n == 1:
            # just one prefix - creating edge node
            current_node = self.trie[pos]
            current_node.reinit(0, skip, first)
            self.prefixes += 1
            self.edge_nodes += 1
            return

        # more prefixes - compute branch
        branch = self.computeBranch(pre, first, n, skip)

        # adr = pointer to the first position of the memory alocated below
        adr = len(self.trie)
        for i in range(2**branch):    # "alocating" memory for (2^branch) new nodes
            self.trie.append(Node("?", "?", "?"))
            self.nodes += 1

        current_node = self.trie[pos]
        current_node.reinit(branch, skip, adr)


        p = first
        for bitpat in range(2**branch):
            # building individual branches
            self.actual_depth += 1
            if self.actual_depth > self.max_depth:
                self.max_depth = self.actual_depth

            # determine how many prefixes will this branch contain
            k = 0
            while (p+k < first + n) and (self.extract(pre+skip, branch, self.prefixset.get_prefixes()[p+k].get_binary()) == bitpat):
                k+=1

            # build the branch
            self.build(p, k, pre+branch+skip, adr+bitpat)

            p = p+k
            self.actual_depth -= 1
        return


    def get_table(self):
        """
        Print table representation of the trie
        [branch    skip    pointer]
        """
        x = 0
        for i in self.trie:
            print(str(x) + ":    " + str(i.branch) + "   " + str(i.skip) + "   " + str(i.adr))
            x+=1
        return
        

        
class Prefix_table_entry(object):
    """
    Entry in prefix table. Represents a prefix of other (longer) prefix. These prefixes are removed from prefixset
    """
    def __init__(self, prefix, pointer = -1):
        self.prefix = prefix
        self.length = len(prefix.get_binary())


class Node(object):
    """
    Common class necessary for the class Trie
    """
    def __init__(self, branch, skip, adr):
        self.branch = branch
        self.skip = skip
        self.adr = adr

    def reinit(self, branch, skip, adr):
        self.branch = branch
        self.skip = skip
        self.adr = adr


