###############################################################################
#  dcam.py: Module for PATTERN MATCH - DCAM
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Vlastimil Kosar <ikosar@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

# Implements DCAM (Pre-decoded CAM) as described in Ioannis Sourdis and Dionisios Pnevmatikatos: Pre-decoded CAMs for Efficient and High-Speed NIDS Pattern Matching
# Note: Partitioning and multiplr char acception is currently not supported
from netbench.pattern_match import b_ptrn_match
from netbench.pattern_match import aux_func
import re
import copy

class dcam(b_ptrn_match.b_ptrn_match):
    """ 
        Implements DCAM (Pre-decoded CAM) as described in Ioannis Sourdis and Dionisios Pnevmatikatos: Pre-decoded CAMs for Efficient and High-Speed NIDS Pattern Matching.
        
        Note: Partitioning and multiple char acception is currently not supported
    """
    def __init__(self):
        """ 
            Class constructor 
        """
        b_ptrn_match.b_ptrn_match.__init__(self)
        self.width = 8
        self.template = aux_func.getPatternMatchDir() + "/algorithms/experimental/dcam/vhdl/sourdis_dcam.vhd"
        self._LUTInputs = 6
        self._luts = 0
        self._strings = list()

    def load_from_file(self, file_name):
        """ 
            Loads strings from specified file.
            
            :param file_name: Name of specified file.  One pattern on each line.
            :type file_name: string
        """
        f = open(file_name,"rb");    #Opens file automat
        blob = f.read()
        lines = list()
        i = 0;
        trueLine = str()
        while i < len(blob):
            if (blob[i] != '\n'):
                trueLine = trueLine + blob[i]
            else:
                lines.insert(len(lines), copy.deepcopy(trueLine))
                trueLine = str()
            i = i + 1
        f.close();
        
        if len(trueLine) != 0:
            lines.insert(len(lines), copy.deepcopy(trueLine))
        
        self._strings = lines
    
    def load_from_string(self, string):
        """ 
            Loads strings from specified string which contains the strings.

            :param string: String with string patterns. One pattern on each line.
            :type string: string
        """
        lines = list()
        i = 0;
        trueLine = str()
        while i < len(string):
            if (string[i] != '\n'):
                trueLine = trueLine + string[i]
            else:
                trueLine = trueLine + string[i]
                lines.insert(len(lines), copy.deepcopy(trueLine))
                trueLine = str()
            i = i + 1
        
        if len(trueLine) != 0:
            lines.insert(len(lines), copy.deepcopy(trueLine))
        
        self._strings = lines
    
    def load_from_list(self, slist):
        """ 
            Loads strings from specified list of strings.
            
            :param slist: The list of string patterns. One pattern in one list element.
            :type slist: list(string)
        """
        self._strings = slist
    
    def compute(self):
        """
            Compute DCAM mapping into FPGA.
        """
        self.get_HDL()
    
    def get_HDL(self):
        """ 
            Returns VHDL representation of DCAM. 
            
            :returns: VHDL description of DCAM unit.
            :rtype: string            
        """
        f = open(self.template, "rb");    #Opens file automat
        blob = f.read()
        tmp = re.split("%\$%", blob)
        
        self._luts = 0
               
        chdec = self._get_char_dec_HDL()
        dcam = self._get_dcam_HDL()
        final = self._get_final_HDL()
        
        dataSignal = chdec[0] + dcam[0] + final[0]
        dataImplementation = chdec[1] + dcam[1] + final[1]
        
        textSignal = "-- --------------- GENERATED BY DCAM.PY ----------------\n"
        textImplementation = "-- --------------- GENERATED BY DCAM.PY ----------------\n"
        
        for element in dataSignal:
            textSignal += element
        for element in dataImplementation:
            textImplementation += element
        textSignal += "-- --------------- END ----------------\n"
        textImplementation += "-- --------------- END ----------------\n"
        
        result = tmp[0] + str(self.width) + tmp[1] + str(self._fStateNum) + tmp[2] + textSignal + tmp[3] + textImplementation + tmp[4]
        return result
        
    def _get_char_dec_HDL(self):
        """
            Return HDL description of shared char decoder as (signals, description). 
                       
            :returns: HDL description of shared char decoder as (signals, description).
            :rtype: tuple(list(string),list(string))
        """
        
        allChars = set()
        
        for string in self._strings:
            for char in string:
                allChars.add(char)
                
        signalList = list()
        descriptionList = list()
                    
        for char in allChars:
            signalList.append("    signal char_" + str(ord(char)) + " : std_logic;\n")
            chrDec  = "    VD_" + str(ord(char)) + ": entity work.VALUE_DECODER\n" 
            chrDec += "    generic map(\n"
            chrDec += "        DATA_WIDTH => " + str(self.width) + ",\n"
            chrDec += "        VALUE      => " + str(ord(char)) + "\n"
            chrDec += "    )\n"
            chrDec += "    port map(\n"
            chrDec += "        INPUT  => DATA,\n"
            chrDec += "        OUTPUT => char_" + str(ord(char)) + "\n"
            chrDec += "    );\n\n"
            descriptionList.append(chrDec)
            self._luts += self._countLUTsForInpunts(self.width)
                            
        return (signalList, descriptionList)
        
    def _get_dcam_HDL(self):
        """
            Return HDL description of DCAM. 
            
            :returns: HDL description of DCAM logic as (signals, description).
            :rtype: tuple(list(string),list(string)) 
        """
               
        delayed_chars = set()
        
        for string in self._strings:
            for i in range(0, len(string)):
                delayed_chars.add((string[i], len(string) - 1 - i))
        
        signalList = list()
        descriptionList = list()
        
        for char in delayed_chars:
            signalList.append("    signal delayed_char_" + str(ord(char[0])) + "_" + str(char[1]) + " : std_logic;\n")
            if char[1] > 0:
                dcam_element  = "    DCAM_DELAY_" + str(ord(char[0])) + "_" + str(char[1]) + ": entity work.sh_reg\n"
                dcam_element += "    generic map(\n"
                dcam_element += "        NUM_BITS    => " + str(char[1]) + "\n";   
                dcam_element += "    )\n"
                dcam_element += "    port map(\n"
                dcam_element += "        CLK      => CLK,\n"
                dcam_element += "        DIN      => char_" + str(ord(char[0])) + ",\n"
                dcam_element += "        CE       => we,\n"
                dcam_element += "        DOUT     => delayed_char_" + str(ord(char[0])) + "_" + str(char[1]) + "\n"
                dcam_element += "    );\n\n"
                descriptionList.append(dcam_element)
                self._luts += char[1] / 32 + 1;
            else:
                descriptionList.append("    delayed_char_" + str(ord(char[0])) + "_" + str(char[1]) + " <= char_" + str(ord(char[0])) + ";\n")
        
        for j in range(0, len(self._strings)):
            signalList.append("    signal symbol_" + str(j) + " : std_logic;\n")
            text = "    symbol_" + str(j) + " <= "
            first = True
            self._luts += self._countLUTsForInpunts(len(string))
            for i in range(0, len(self._strings[j])):
                if first == True:
                    text += "delayed_char_" + str(ord(self._strings[j][i])) + "_" + str(len(self._strings[j]) - 1 - i)
                    first = False
                else:
                    text += " and delayed_char_" + str(ord(self._strings[j][i])) + "_" + str(len(self._strings[j]) - 1 - i)
            text += ";\n"
            descriptionList.append(text)
            
        return (signalList, descriptionList)

    def _get_final_HDL(self):
        """
            Return HDL description of found strings bitmap. 
            
            :returns: HDL description of found strings bitmap as (signals, description).
            :rtype: tuple(list(string),list(string))
        """
        signalList = list()
        descriptionList = list()
               
        self._fStateNum = len(self._strings)
        
        signalList.append("    signal bitmap_in : std_logic_vector(" + str(len(self._strings)) + " - 1 downto 0);\n")
        for i in range(0, len(self._strings)):
            signalList.append("    signal final_" + str(i) + " : std_logic;\n")
            descriptionList.append("    final_" + str(i) + " <= symbol_" + str(i) + ";\n")
            descriptionList.append("    bitmap_in(" + str(i) + ") <= final_" + str(i) + ";\n")
                
        return (signalList, descriptionList)

    def report_logic(self):
        """ 
            Reports amount of logic consumed by the algorithm. 
            
            :returns: Amount of logic consumed by the algorithm. (LUTs, FlipFlops, BRAMs)
            :rtype: tuple(int, int, int)
        """
        if self._luts == 0:
            self.get_HDL()
        ffSize = len(self._strings) + 1        
        bramSize = 0        
        return (self._luts, ffSize, bramSize)
        
    def _countLUTsForInpunts(self, inputs):
        """
            Counts amount of LUTs necessary for implenentation of inputs-input logic function.
            
            :param inputs: Number of logic function inputs
            :type inputs: int
            
            :returns: Amount of LUTs necessary for implenentation of inputs-input logic function.
            :rtype: int
        """
        i = inputs
        res = 0;
        if i <= self._LUTInputs:
            res = 1;
        else:
            while i > self._LUTInputs:
                if i/self._LUTInputs == round(i/self._LUTInputs):
                    res = res + int(round((i / self._LUTInputs), 0))
                    i = int(round((i / self._LUTInputs), 0))
                else:
                    res = res + int(round((i / self._LUTInputs) + 0.5, 0))
                    i = int(round((i / self._LUTInputs) + 0.5, 0))
            res = res + 1;
        return res
