############################################################################
#  dga.py: Module for PATTERN MATCH algorithm Deterministic Generalized
#          Automata
#  Copyright (C) 2011 Brno University of Technology, ANT @ FIT
#  Author(s): Martin Soka <xsokam00@stud.fit.vutbr.cz>
############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is, and any express or
#  implied warranties, including, but not limited to, the implied warranties
#  of merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

from netbench.pattern_match.b_dfa import b_dfa
from netbench.pattern_match import sym_string
from netbench.pattern_match.b_automaton import b_Automaton

class DGA(b_dfa):
    """
        Class for Deterministic Generalized Automata.
        
        Based on:
           "Algorithms to accelerate multiple regular expressions matching 
           for deep packet inspection"
           ISBN: 1-59593-308-5
           URL: http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.52.7243
           
        A generalized automaton (GA) is a finite automaton where the single 
        transitions are defined on words rather than on single letters.
    """
    
    def __init__(self):
        
        b_dfa.__init__(self)
        
        
    def compute(self):
        """
            This method computes Deterministic Generalized Automata.
            
            :raises: COMPUTE_ERROR
        """
        
        b_dfa.compute(self)
        
        #Check if dfa was computed.
        if not self.get_compute():
            raise COMPUTE_ERROR        
        
        self._make_dga()
    
    def _make_dga(self):
        """
            This method changes Deterministic Finite Automaton into 
            Deterministic Generalized Automata.
            
            :raises: DETERMINISTIC_ERROR
        """
        
        # Check if the automat is deterministic.
        if not self.has_flag("Deterministic") \
           or self.get_flag("Deterministic") == False:
            raise DETERMINISTIC_ERROR
        
        # Creation list of superfluous states. 
        superf = list()
        superf = self._get_superf()
        
        # Creation dictionary of labels of path in automaton.
        # Labels are the lists of unique numbers.
        self.xtrans = {}
        for trans in self._automaton.transitions:
            self.xtrans[trans[1]] = []
            self.xtrans[trans[1]].append(trans[1])
        
        # Check if automaton contains superfluous states.
        if superf != None:
            deleteS = list()
            deleteS = self._max_acyclic_subgraph(superf)
        
            b_Automaton.resolve_alphabet(self)

            self._reduction(deleteS)
        
        # Set automat flag.
        self._automaton.Flags["Generalized automaton"] = True
        self._compute = True

    def _get_superf(self):
        """
            This method search superfluous states in automaton.
            
            :returns: List of superfluous states.
            :rtype: list(int)
        """
        
        superf = list()
        
        a = self._automaton
        
        for state in a.states:
            # Check if state is initial.
            if state == a.start:
                continue
            
            # Check if state is final.
            if state in a.final:
                continue
            
            flag = 0
            for trans in a.transitions:
                # Check if state contains self-loop.
                if trans[0] == state and trans[2] == state:
                    flag = 1
                    break
            if flag:
                continue
            
            superf.append(state)
        
        if len(superf) == 0:
            return None

        return superf
        
    def _search_graph(self, superf, lst, state, ac_sg):
        """
            This is a recursive method which looks for a maximal acyclic subgraph.
            
            :param superf: List of superfluous states
            :type superf: list(int)
            :param lst: List of states, which induces acyklic subgraph
            :type lst: list(int)
            :param state: State, from which search next states.
            :type state: int
            :param ac_sg: List of states, which induces actual maximal 
                          acyklic subgraph
            :type ac_sg: list(int)
        """
        
        tr = self._automaton.transitions
        
        # List of new states.
        new_q = list()
        
        # List of testing states.
        test = list()
        
        for trans in tr:
            if trans[0] == state and trans[2] in superf and trans[2] not in lst:
                flag = 1
                test.append(trans[2])
                for trans2 in tr:
                    # Chcek if a cycle does not arise after adding a state.
                    if trans2[0] == trans[2] and trans2[2] in lst:
                        flag = 0
                        break
                if flag:
                    new_q.append(trans[2])
            
            if trans[2] == state and trans[0] in superf and trans[0] not in lst:
                flag = 1
                test.append(trans[0])
                for trans2 in tr:
                    # Chcek if a cycle does not arise after adding a state
                    if trans2[2] == trans[0] and trans2[0] in lst:
                        flag = 0
                        break
                if flag:
                    new_q.append(trans[0])
                    
        # Check if new states was searching.
        if len(new_q) != 0:
            for i in range(len(new_q)):
                new_lst = list()
                new_lst.extend(lst)
                new_lst.append(new_q[i])
                
                if len(new_lst) > len(ac_sg):
                    del ac_sg[:]
                    ac_sg.extend(new_lst)
                
                # Check if the maximal acyclical subgraph is induced 
                # by all of the superfluous states.
                if len(ac_sg) == len(superf):
                    return
                
                self._search_graph(superf, new_lst, new_q[i], ac_sg)
                
        # Searching other state.
        else:
            for q in superf:
                if q not in lst and q not in test:
                    new_lst = list()
                    new_lst.extend(lst)
                    new_lst.append(q)
                    
                    if len(new_lst) > len(ac_sg):
                        del ac_sg[:]
                        ac_sg.extend(new_lst)
                    
                    # Check if the maximal acyclical subgraph is induced 
                    # by all of the superfluous states.
                    if len(ac_sg) == len(superf):
                        return
                    
                    self._search_graph(superf, new_lst, q, ac_sg)



    def _max_acyclic_subgraph(self, superf):
        """
            This method looks for a maximal acyklic subgraph.
            
            :param superf: List of superfluous states
            :type superf: list(int)
            :returns: List of states, which induces actual maximal 
                      acyklic subgraph
            :rtype: list(int)
        """
        
        # Check if superfluous state is only one.
        if len(superf) == 1:
            return superf
        
        # Creation of list of states, which induces maximal acyklic subgraph
        ac_sg = list()
        
        for state in superf:
            lst = list()
            lst.append(state)

            # Calling recursive method.
            self._search_graph(superf,lst,state,ac_sg)
            
            if len(lst) > len(ac_sg):
                del ac_sg[:]
                ac_sg.extend(a)
    
            # Check if the maximal acyclical subgraph is induced 
            # by all of the superfluous states.
            if len(ac_sg) == len(superf):
                return ac_sg
        
        return ac_sg        
    
    def _reduction(self, deleteS):
        """
            This method reduces number of states.
            
            :param deleteS: List of superfluous states
            :type deleteS: list(int)
        """
        
        states = self._automaton.states
        transitions = self._automaton.transitions
        
        xtrans = {}
        for trans in transitions:
            xtrans[trans[1]] = []
            xtrans[trans[1]].append(trans[1])
        
        while len(deleteS) != 0:
            delete = deleteS.pop(0)
            
            st = dict()
            tr = set()

            for trans in transitions:
                # Check if transition contains actual uperfluous state.
                if delete != trans[0] and delete != trans[2]:
                    tr.add(trans)
                    st[trans[0]] = states[trans[0]]
                    st[trans[2]] = states[trans[2]]
                else:
                    if delete == trans[2]:
                        for trans2 in transitions:
                            if trans2[0] == delete:
                                st[trans[0]] = states[trans[0]]
                                st[trans2[2]] = states[trans2[2]]
                                
                                # Creation of new transition
                                inx = max(xtrans.keys()) + 1
                                xtrans[inx] = []
                                xtrans[inx].extend(xtrans[trans[1]])
                                xtrans[inx].extend(xtrans[trans2[1]])
                                tr.add((trans[0], inx, trans2[2]))                                
                
            states = st
            transitions = tr 

        # Deleting of unused labels
        for a in xtrans.keys():
            count = 0
            for trans in transitions:
                if a == trans[1]:
                    count += 1
            if count == 0:
                del xtrans[a]

        self._automaton.states = states
        self._automaton.transitions = transitions
        self.xtrans = xtrans
        
    def search(self, input_string):
        """
            This function will find patterns in the given string.

            :param input_string: Input string.
            :param input_string: string
            :returns: Bitmap of matched regular expressions.
            :rtype: list(int)
        """
        
        # Stack of states. State is list (tupple) consisting of state 
        # and unprocessed part of input string.
        Stack = list()
        # Set of actual states.
        ActState = set()

        # Create start state.
        ActState.add((self._automaton.start, input_string, False))
        # Add start state to stack.0
        Stack.append(ActState)

        # Create mapping between reg. exp. number and coresponding final 
        # states.
        sameFinal = dict()
        for fstate in self._automaton.final:
            rnums = self._automaton.states[fstate].get_regexp_number()
            for rnum in rnums:
                if rnum in sameFinal:
                    sameFinal[rnum].add(fstate)
                else:
                    sameFinal[rnum] = set()
                    sameFinal[rnum].add(fstate)

        # Compute number of reg. exp.
        rules = len(sameFinal)
        # dictionary is no longer needed.
        sameFinal = None

        # Init bitmap to zeros.
        bitmap = [0] * rules

        # If needed create mapping between states and outgoing transitions.
        _mapper = dict()
        for transition in self._automaton.transitions:
            if transition[0] in _mapper:
                _mapper[transition[0]].add((transition[1], transition[2]))
            else:
                _mapper[transition[0]] = set()
                _mapper[transition[0]].add((transition[1], transition[2]))

        # Until stack is empty, search
        while len(Stack) != 0:
            # Pop state from stack.
            ActState = Stack.pop()
            newActState = set()
            
            # Create new state. Accept char if possible and add state 
            # to new state.
            for state in ActState:
                if _mapper.has_key(state[0]):
                    for transition in _mapper[state[0]]:
                        number = 0
                        res = state[1]
                        for symbol in self.xtrans[transition[0]]:
                            try:
                                res = self._automaton.alphabet[symbol].accept(res)
                            except:
                                pass
                            else:
                                number += 1
                        
                        if number == len(self.xtrans[transition[0]]):
                            newActState.add((transition[1],res))
                # If in final state, set coresponding bitmap field to 1.
                if self._automaton.states[state[0]].is_final() == True:
                    for rnum in self._automaton.states[state[0]].get_regexp_number():
                        bitmap[rnum] = 1
            # If possible add new state to stack.
            if len(newActState) > 0:
                Stack.append(newActState)

        # Return bitmap.
        return bitmap

        
    def show(self, fileName, sizeStr=" size=\"8,5\"\n"):
        """
            This method can't be used.
        """
        pass

           
       