###############################################################################
#  test_ddfa.py: Test module for DGA
#  Copyright (C) 2011 Brno University of Technology, ANT @ FIT
#  Author(s): Martin Soka <xsokam00@stud.fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is, and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

from dga import DGA
from netbench.pattern_match.pcre_parser import pcre_parser
from netbench.pattern_match.nfa_data import nfa_data
from netbench.pattern_match.b_dfa import b_dfa
from netbench.pattern_match.b_nfa import b_nfa
from netbench.pattern_match.b_symbol import b_Symbol, DEF_SYMBOLS
from netbench.pattern_match.sym_char import b_Sym_char
from netbench.pattern_match.b_state import b_State
from netbench.pattern_match.b_automaton import b_Automaton
import unittest

class test_dga(unittest.TestCase):
    
    def test_compute(self):
        """compute()"""
        
        #Test with simple expression - DGA doesn`t reduce any states.
        #For example regular expression /a+b+/.
        self._test_compute1()
            
        #Test with simple expression - DGA reduce some states.
        #For example regular expression /^abcd/.
        self._test_compute2()
        
        #Test with two simple expressions.
        #For example regular expressions /^POST/\n/^GET/.
        self._test_compute3()
        
        #Test with two expressions, which contain char classes.
        #For example regular expressions /^[a-e]+bc/\n/^[x-z]ab+/.
        self._test_compute4()
            
    def _test_compute1(self):
            
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text("/a+b+/")
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        dfa.create_by_parser(parser)
        dfa.compute()
        
        a = dga.get_automaton()
        b = dfa.get_automaton()
        
        
        self.assertEqual(a.states.keys(), b.states.keys())
        self.assertEqual(a.start, b.start)
        self.assertEqual(a.final, b.final)
        self.assertEqual(a.alphabet, b.alphabet)
        self.assertEqual(a.transitions, b.transitions)
        self.assertTrue(a.Flags["Generalized automaton"])     
        
    def _test_compute2(self):
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text("/^abcd/")
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        
        dfa = b_dfa()
        dfa.create_by_parser(parser)
        dfa.compute()
        
        a = dga.get_automaton()
        b = dfa.get_automaton()       
        
        self.assertEqual(len(a.states), 2)
        self.assertEqual(len(a.transitions), 1)
        self.assertEqual(a.start, b.start)
        self.assertEqual(a.final, b.final)
        self.assertEqual(a.alphabet, b.alphabet)
        self.assertTrue(a.Flags["Generalized automaton"])
        
    def _test_compute3(self):
        
        re = "/^POST/\n/^GET/"
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text(re)
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        
        parser = pcre_parser()
        parser.set_text(re)
        dfa.create_by_parser(parser)
        
        dfa.compute()
        
        a = dga.get_automaton()
        b = dfa.get_automaton()       
        
        self.assertEqual(len(a.states), 3)
        self.assertEqual(len(a.transitions), 2)
        self.assertEqual(a.start, b.start)
        self.assertEqual(a.final, b.final)
        self.assertEqual(a.alphabet, b.alphabet)
        self.assertTrue(a.Flags["Generalized automaton"])
        
    def _test_compute4(self):
        
        re = "/^[a-e]+bc/\n/^[x-z]ab+/"
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text(re)
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        
        parser = pcre_parser()
        parser.set_text(re)
        dfa.create_by_parser(parser)
        
        dfa.compute()
        b_Automaton.resolve_alphabet(dfa)
        
        a = dga.get_automaton()
        b = dfa.get_automaton()       
        
        self.assertEqual(len(a.states), 5)
        self.assertEqual(len(a.transitions), 18)
        self.assertEqual(a.start, b.start)
        self.assertEqual(a.final, b.final)
        self.assertEqual(a.alphabet, b.alphabet)
        self.assertTrue(a.Flags["Generalized automaton"])
    
    def test_search(self):
        """search()"""
        
        #Tests a set of regular expressions and compares the results with
        #results form DFA.
        self._test_search1()
        
        self._test_search2()
        
        self._test_search3()
        
        self._test_search4()
        
    def _test_search1(self):
        
        re = "/^POST/\n/^GET/"
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text(re)
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        
        parser = pcre_parser()
        parser.set_text(re)
        dfa.create_by_parser(parser)
        
        dfa.compute()     
        
        self.assertEqual(dga.search("GET"), dfa.search("GET"))
        self.assertEqual(dga.search("POST"), dfa.search("POST"))
        self.assertEqual(dga.search(" GET"), dfa.search(" GET"))
    
    def _test_search2(self):
        
        re = "/abcd/\n/efgh/"
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text(re)
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        
        parser = pcre_parser()
        parser.set_text(re)
        dfa.create_by_parser(parser)
        
        dfa.compute()     
        
        self.assertEqual(dga.search("abcdefgh"), dfa.search("abcdefgh"))
        self.assertEqual(dga.search("xxxabcd"), dfa.search("xxxabcd"))
        self.assertEqual(dga.search("abc"), dfa.search("abc"))
        self.assertEqual(dga.search("cdefgh"), dfa.search("cdefgh"))
        
    def _test_search3(self):
        
        re = "/forum=.*/"
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text(re)
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        
        parser = pcre_parser()
        parser.set_text(re)
        dfa.create_by_parser(parser)
        
        dfa.compute()     
        
        self.assertEqual(dga.search("aforum="), dfa.search("aforum="))
        self.assertEqual(dga.search("forum=abc"), dfa.search("forum=abc"))
        self.assertEqual(dga.search("forum"), dfa.search("forum"))
    
    def _test_search4(self):
        
        re = "/abcd.*/\n/^[a-e]+[c-e]*[a-e]{2}/"
        
        dga = DGA()
        
        parser = pcre_parser()
        parser.set_text(re)
        dga.create_by_parser(parser)
        
        dga.compute()
        self.assertTrue(dga.get_compute())
        
        dfa = b_dfa()
        
        parser = pcre_parser()
        parser.set_text(re)
        dfa.create_by_parser(parser)
        
        dfa.compute()     
        
        self.assertEqual(dga.search("abcdxxx"), dfa.search("abcdxxx"))
        self.assertEqual(dga.search("eaee"), dfa.search("eaee"))
        self.assertEqual(dga.search("abccaaabcd"), dfa.search("abccaaabcd"))
        self.assertEqual(dga.search("aaa"), dfa.search("aaa"))
        self.assertEqual(dga.search("aeee"), dfa.search("aeee"))
        self.assertEqual(dga.search("abcdabcd"), dfa.search("abcdabcd"))
    
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(test_dga)
    unittest.TextTestRunner(verbosity=2).run(suite)