/***************************************************************************
############################################################################
#  extend_fa.c: Algorithm for Extend FA (transit through automat)
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Jaroslav Suchodol
############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or
#  implied warranties, including, but not limited to, the implied warranties
#  of merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$
***************************************************************************/

/*
 * extend_fa.c    ver. 1.0
 *
 * Algorithm for Extend FA (transit through automat)
 * =================================================
 *
 * Jaroslav Suchodol,   September 2010
 *
 */

#include <stdio.h>
#include <stdlib.h>

# define false 0
# define true 1

typedef unsigned int u_i;

// FLAGS (actions)
// 0 = allocation, 1 = hold in this state, 2 = de-allocation, 5 = not set

// Transition for Extend FA.
typedef struct t_e_t {
  u_i *c;   // transition char(s)
  u_i c_c;  // count char for transition
  u_i d;    // destination state of transition
  u_i f;    // flag associated with transition (action)
  u_i ctr;  // counter associated with flag
} T_E_T;

// State for Extend FA.
typedef struct t_e_s {
  T_E_T *t; // transitions for state
  u_i t_c;  // t_c=trans_count
  u_i a;    // indication of activate state (true, false)
  u_i in;   // incomed tran. to this state
  u_i f;    // mark of the final state
} T_E_S;

// Extend FA.
typedef struct t_e {
  T_E_S *s;   // states of E-FA
  u_i s_c;    // states count
  u_i s_s;    // start state
} T_E;    // T_EXTEND_FA

// Counter.
typedef struct t_c {
  u_i m;  // maximum
  u_i c;  // actual count
  u_i c_a;// count auxiliary
  u_i a;  // activation (true, false)
  u_i *d; // delta of ther counters
  u_i b;  // begin index of delta
  u_i e;  // end index of delta
  u_i in; // incomed tran. to state which is "hold in this state"
} T_C;

// global variables
T_E e;      // Extend automaton
T_C *ctr;   // counters
u_i ctr_c;  // counters count


/*
 * Parse input file (first argument of program) and
 * make structure from it for later use.
 */
void parse_file(char *FileName)
{
  FILE *fr;
  u_i i, j, k;

  // open parse file
  if ((fr = fopen(FileName, "r")) == NULL) {
    fprintf(stderr, "Error: file \"%s\" could not be open for reading!\n",
    FileName);
    exit(1);
  }
  // fill data structures
  // parse COUNT OF STATES
  fscanf(fr, "%u\n", &e.s_c);
  e.s = (T_E_S *) malloc(e.s_c * sizeof(T_E_S));
  if (e.s == NULL) {
    fprintf(stderr, "Error: not enough memory!\n");
    exit(1);
  }
  // parse STARTING STATE
  fscanf(fr, "%u\n", &e.s_s);
  // parse TRANSITIONS between states
  for (i = 0; i < e.s_c; i++) {
    // set all state for not final
    e.s[i].f = false;
    // deactivate all states
    e.s[i].a = false;
    e.s[i].in = false;
    // count transitions for given state
    fscanf(fr, "t_c: %u\n", &e.s[i].t_c);
    e.s[i].t = (T_E_T *) malloc(e.s[i].t_c * sizeof(T_E_T));
    if (e.s[i].t == NULL) {
      fprintf(stderr, "Error: not enough memory!\n");
      exit(1);
    }
    // fill chars, count chars, destination state
    for (j = 0; j < e.s[i].t_c; j++) {
      fscanf(fr, "c_c: %u, d: %u\n", &e.s[i].t[j].c_c, &e.s[i].t[j].d);
      e.s[i].t[j].c = (u_i *) malloc(e.s[i].t[j].c_c * sizeof(u_i));
      if (e.s[i].t[j].c == NULL) {
        fprintf(stderr, "Error: not enough memory!\n");
        exit(1);
      }
      for (k = 0; k < e.s[i].t[j].c_c; k++) {
        fscanf(fr, "%u|", &e.s[i].t[j].c[k]);
      }
      fscanf(fr, "\n");
      // fill flag (action) for transition
      fscanf(fr, "f: %u, ctr: %u\n", &e.s[i].t[j].f, &e.s[i].t[j].ctr);
    }
  }
  // parse FINAL STATES
  u_i s_f_c;
  fscanf(fr, "%u\n", &s_f_c);
  for (i = 0; i < s_f_c; i++) {
    fscanf(fr, "%u|", &j);
    e.s[j].f = true;
  }
  fscanf(fr, "\n");
  // parse COUNTERS
  fscanf(fr, "ctr_c: %u\n", &ctr_c);
  ctr = (T_C *) malloc(ctr_c * sizeof(T_C));
  if (ctr == NULL) {
    fprintf(stderr, "Error: not enough memory!\n");
    exit(1);
  }
  // parse maximum of counter and initialize T_C structure
  for (i = 0; i < ctr_c; i++) {
    ctr[i].a = false; // inactive counter
    fscanf(fr, "%u|", &ctr[i].m);
    // m - 1 because the oldest counter is in ctr[i].c
    ctr[i].d = (u_i *) malloc((ctr[i].m - 1) * sizeof(u_i));
    if (ctr[i].d == NULL) {
      fprintf(stderr, "Error: not enough memory!\n");
      exit(1);
    }
  }
  fscanf(fr, "\n");
  // close parse file
  if (fclose(fr) == EOF) {
    fprintf(stderr, "Error: file \"%s\" could not be closed!\n", FileName);
  }
}


/*
 * Passing through automat with chars from file
 * which is second argument of program.
 */
void passing_automat(char *FileName)
{
  FILE *fr;
  int c;        // one input char from file
  u_i i, j, k;  // helpful variables
  u_i c_i;      // counter index

  e.s[e.s_s].a = true;  // activate start state
  // open data file
  if ((fr = fopen(FileName, "rb")) == NULL) {
    fprintf(stderr, "Error: file \"%s\" could not be open for reading!\n",
    FileName);
    exit(1);
  }
  // Follow algorithm for way through automat.
  while ((c = getc(fr)) != EOF) {
    // look for activated state
    for (i = 0; i < e.s_c; i++) {
      if (e.s[i].a == true) {
        // in current state try all transitions
        for (j = 0; j < e.s[i].t_c; j++) {
          for (k = 0; k < e.s[i].t[j].c_c; k++) {
            // found transition for working char
            if (c == e.s[i].t[j].c[k]) {
              c_i = e.s[i].t[j].ctr;

              // allocation
              if (e.s[i].t[j].f == 0) {
                if (ctr[c_i].a == true) {
                  ctr[c_i].d[ctr[c_i].e] = ctr[c_i].c;
                  ctr[c_i].e = (ctr[c_i].e + 1) % (ctr[c_i].m - 1);
                }
                // counter not active
                else {
                  // activate counter + initialization
                  ctr[c_i].a = true;
                  ctr[c_i].c_a = 1; ctr[c_i].b = 0; ctr[c_i].e = 0;
                }
                ctr[c_i].in = true;
              }

              // hold in this state
              else if (e.s[i].t[j].f == 1 && ctr[c_i].a == true) {
                // increment
                if (ctr[c_i].c < ctr[c_i].m) {
                  ctr[c_i].c_a += 1;
                  ctr[c_i].in = true;
                }
                // count overflow
                else {
                  // activate next count and increment
                  if (ctr[c_i].b != ctr[c_i].e) {
                    ctr[c_i].c_a = ctr[c_i].c + 1 - ctr[c_i].d[ctr[c_i].b];
                    ctr[c_i].b = (ctr[c_i].b + 1) % (ctr[c_i].m - 1);
                    ctr[c_i].in = true;
                  }
                }
              }

              // de-allocation (done in hold in this state (else branch))
              else if (e.s[i].t[j].f == 2 && ctr[c_i].a == true) {
                if (ctr[c_i].m != ctr[c_i].c) {
                  // do not use this transition
                  break;
                }
              }

              // execute transition
              if (e.s[i].t[j].f == 5
                  || (  (e.s[i].t[j].f >= 0 && e.s[i].t[j].f <= 2)
                        && ctr[c_i].a == true)  ) {
                // deactivate current state (automaticly if not in)
                // activate destinaton state
                e.s[e.s[i].t[j].d].in = true;
                // check for final state
                if (e.s[e.s[i].t[j].d].f == true) {
                  printf("\n*** Result: match ***\n\n");
                  goto finish_after_match;
                }
                // skip other chars
                break;
              }
            }
          }
        }
      }
    }
    // after process char, change the activation of states
    for (i = 0; i < e.s_c; i++) {
      if (e.s[i].in == true) {
        e.s[i].a = true;
      } else {
        e.s[i].a = false;
      }
      e.s[i].in = false;
    }
    // after process char, change the activation of counters and count
    for (i = 0; i < ctr_c; i++) {
      if (ctr[i].in != true) {
        ctr[i].a = false;
      }
      ctr[i].in = false;
      ctr[i].c = ctr[i].c_a;
    }
  }
  // algorithm was NOT anytime in finite state
  printf("\n*** Result: NOT match ***\n\n");
  finish_after_match:
  // close data file
  if (fclose(fr) == EOF) {
    fprintf(stderr, "Error: file \"%s\" could not be closed!\n", FileName);
  }
}


/*
 * Main function.
 */
int main(int argc, char *argv[])
{
  if (argc == 3) {
    parse_file(argv[1]);
    passing_automat(argv[2]);
  } else {
    fprintf(stderr, "Error: bad arguments!\n");
    fprintf(stderr, "Example: ./extend_fa parse_file data_file\n");
    exit(1);
  }
  return 0;
}

