###############################################################################
#  nfa_split.py: Module for PATTERN MATCH
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Vlastimil Kosar <ikosar@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

# Pattern Match modules
from netbench.pattern_match import b_state
from netbench.pattern_match.b_nfa import b_nfa
from netbench.pattern_match.b_dfa import b_dfa
from netbench.pattern_match import nfa_data
from netbench.pattern_match import sym_char
from netbench.pattern_match import sym_char_class
from netbench.pattern_match import b_symbol
from netbench.pattern_match.algorithms.clark_nfa import clark_nfa
from netbench.pattern_match.pattern_exceptions import COMPUTE_ERROR, empty_automaton_exception
from netbench.pattern_match import aux_func
from nfa_split_det_part import nfa_det_part
from nfa_split_nondet_part import nfa_nondet_part

# General modules
import random
import gc
import re
import cPickle
from collections import deque
import os
import sys
import math
import copy

            
# -----------------------------------------------------------------------
#
#   NFA Split architecture - reduction of logic resources 
#
# -----------------------------------------------------------------------

class nfa_split(b_nfa):
    """
        The NFA split class implements NFA split architecture. This            \
        architecture consists from nondeterministic part and from one ore more \
        deterministic parts. Construction algorithm determinates which states  \
        can be active together and then the states which can't be active       \
        together are mapped into DFA parts as more than one set of nonactive   \
        states is possible - each set coresponds to one deterministic part.    \
        States which can be active together are implemented in the             \
        nondeterministic part.
        
        This class is based on:
        
        Korenek Jan, Kosar Vlastimil: NFA Split Architecture for Fast Regular  \
        Expression Matching, In: Proceedings of the 6th ACM/IEEE Symposium on  \
        Architectures for Networking and Communications Systems, La Jolla, US, \
        ACM, 2010, p. 2, ISBN 978-1-4503-0379-8
        
        Korenek Jan, Kosar Vlastimil: Efficient Mapping of Nondeterministic    \
        Automata to FPGA for Fast Regular Expression Matching, In: Proceedings \
        of the 13th IEEE International Symposium on Design and Diagnostics of  \
        Electronic Circuits and Systems DDECS 2010, Vienna, AT, IEEE CS, 2010, \
        s. 6, ISBN 978-1-4244-6610-8
        
    """
    def __init__(self):
        b_nfa.__init__(self)

        self._det_parts  = list();  # deterministic parts - list of
                                    # nfa_det_part classes
        self._nfa_part   = None;    # non-deterministic part is 
                                    # defined as a reference to Clark
                                    # NFA approach
        
        self._allNoncollSets = None
        self.statesCollisions = dict()
        self._statistics = dict() # statistics, key - string name
        self.template = aux_func.getPatternMatchDir() + "/algorithms/experimental/nfa_split/vhdl/unit.vhd"
        self._reduced = False
        self.width = 8
        self._fStateNum = 0
    
    def concurrently_reachable(self):
        """
            Compute set of state's tuples, which can ba active in same time.

            :returns: Set of concurrently active states.
            :rtype: set(tuple(int, int))
        """
        stack = list()
        reachable = set()
        
        mapper = dict()
        for state in self._automaton.states.keys():
            mapper[state] = dict()
        for transition in self._automaton.transitions:
            if mapper[transition[0]].has_key(transition[1]) == True:
                mapper[transition[0]][transition[1]].add(transition[2])
            else:
                mapper[transition[0]][transition[1]] = set()
                mapper[transition[0]][transition[1]].add(transition[2])
                
        stack.append((self._automaton.start, self._automaton.start))
        reachable.add((self._automaton.start, self._automaton.start))
        while len(stack) > 0:
            states = stack.pop()
            q1 = states[0]
            q2 = states[1]
            for a in self._automaton.alphabet.keys():
                if a in mapper[q1].keys():
                    for q3 in mapper[q1][a]:
                        if a in mapper[q2].keys():
                            for q4 in mapper[q2][a]:
                                if q3 < q4:
                                    q5q6 = (q3, q4)
                                else:
                                    q5q6 = (q4, q3)
                                if q5q6 not in reachable:
                                    reachable.add(q5q6)
                                    stack.append(q5q6)
        removeable = set()
        for states in reachable:
            if states[0] == states[1]:
                removeable.add(states)
        return reachable - removeable
    
    # ----------- Deterministic parts creation --------
    def create_det_parts(self, number_of_parts=1, table_mapping_method=0,      \
                         useCharClasses = False, concurrency_method=0):
        """
            Create deterministic parts of NFA. Number of parts, table mapping  \
            algorithm, usage of charclasses and mathod of determinating state  \
            concurrecy can be specified.
            
            :param number_of_parts: Number of deterministic parts to create.   \
                                    Less number parts can be generated if      \
                                    requested number of parts is higher than   \
                                    number of parts which can be generated.    
            :type number_of_parts: int
            :param table_mapping_method: Specifies transition table mapping    \
                                         algorithm. Use 0 for heuristic and 1  \
                                         for genetic algorithm. Usage of 0 is  \
                                         recommended.
            :type table_mapping_method: int
            :param useCharClasses: Specifies if character class are used in DFA\
                                   parts.
            :type useCharClasses: boolean
            :param concurrency_method: Specifies algorithm which is used to    \
                                       determinate state concurrecy. Use 0 for \
                                       determinisation and 1 for               \
                                       concurrently_reachable() method. The    \
                                       concurrently_reachable() method has time\
                                       complexity O(n^2).
            :type concurrency_method: int
            
            :returns: State of DFA parts construction. True means that all is  \
                      OK and False indicate error during construction.
            :rtype: boolean
        """
        # If no automaton is loaded, return False
        if self._automaton == None or len(self._automaton.transitions) == 0:
            return False
        
        # If sets of non collison states are not computed, compute them
        if self._allNoncollSets == None:
            # Use determinisation for generation of concurency informations
            if concurrency_method == 0:
                # Remove epsilons
                self.remove_epsilons()
                # Update statistics
                self._statistics["NFA States"] = len(self._automaton.states)
                self._statistics["NFA Transitions"] = len(self._automaton.transitions)
                
                # Create DFA
                dfa = b_dfa()
                dfa.create_from_nfa_data(self._automaton)
                # Create char classes to simplify determinisation
                dfa.create_char_classes()
                # Determinise the automaton to get concurrecy informations
                dfa.determinise(True)
                
                # Update statistics
                self._statistics["DFA States"] = len(dfa._automaton.states)
                self._statistics["DFA Transitions"] = len(dfa._automaton.transitions)
                
                # Get for each state its collision states
                # Init data structures
                self.statesCollisions = dict()
                for state in self._automaton.states.keys():
                    self.statesCollisions[state] = set()
                
                # Init strucures
                activeStatesSizeList = [0]*len(dfa._state_representation)
                
                # Compute the collision states for all states
                for detState in range(0, len(dfa._state_representation)):
                    activeStatesSizeList[detState] = len(dfa._state_representation[detState])
                    for state in dfa._state_representation[detState]:
                        mappedStates = copy.deepcopy(dfa._state_representation[detState])
                        mappedStates.remove(state)
                        if len(mappedStates) > 0:
                            self.statesCollisions[state].update(mappedStates)
                
                # Update statistics
                maxActiveStates = max(activeStatesSizeList) 
                self._statistics["Max Active States"] = maxActiveStates
                
                # free memory
                dfa = None
                activeStatesSizeList = None
                
                # Generate split into sets of noncollision states
                self._allNoncollSets = self._AllNonCollisionSets(self.statesCollisions)
            else:
                # Use concurrently_reachable() for generation of concurency informations
                # Remove epsilons
                self.remove_epsilons()
                # Update statistics
                self._statistics["NFA States"] = len(self._automaton.states)
                self._statistics["NFA Transitions"] = len(self._automaton.transitions)
                
                # Compute the list of concurrently reachable states
                self.remove_char_classes()
                concurrent = self.concurrently_reachable()
                self.create_char_classes()
                
                # Update statistics - we do not use DFA for this method
                self._statistics["DFA States"] = -1
                self._statistics["DFA Transitions"] = -1
                                
                # Get for each state its collision states
                # Init data structures
                self.statesCollisions = dict()
                for state in self._automaton.states.keys():
                    self.statesCollisions[state] = set()
                
                # Compute the collision states for all states
                for collision in concurrent:
                    for state in collision:
                        mappedStates = list(collision)
                        mappedStates.remove(state)
                        if len(mappedStates) > 0:
                            self.statesCollisions[state].update(mappedStates)
                
                # Update statistics - we do not use DFA for this method
                self._statistics["Max Active States"] = -1
                
                # free memory
                nfa = None
                
                # Generate split into sets of noncollision states
                self._allNoncollSets = self._AllNonCollisionSets(self.statesCollisions)
        else:
            # Init list of deterministic states
            self._det_parts = list()
        
        # Init variables
        nonColSize = list()
        colSize = 0
        
        # Generate deterministic parts, number of parts is:
        #min(number_of_parts, len(self._allNoncollSets))
        for i in range (0, min(number_of_parts, len(self._allNoncollSets))):
            # Create empty deterministic part
            self._det_parts.append(nfa_det_part())
            # Get list of nocollision sets with oidex i
            nonColSize.append(len(self._allNoncollSets[i]))
            # Copy states from original automaton into deterministic part
            for state in self._automaton.states.keys():
                if state in self._allNoncollSets[i]:
                    self._det_parts[i]._automaton.states[state] = self._automaton.states[state]
            
            # Copy final states from original automaton into deterministic part
            for fstate in self._automaton.final:
                if fstate in self._allNoncollSets[i]:
                    self._det_parts[i]._automaton.final.add(fstate)
            
            # Copy transitions from original automaton into deterministic part,
            # generate insets and outsets (transitions between deterministic 
            # part and nondeterministic part) and copy symbols from original 
            # automaton into deterministic part
            for transition in self._automaton.transitions:
                if transition[0] in self._allNoncollSets[i] and transition[2] in self._allNoncollSets[i]:
                    self._det_parts[i]._automaton.transitions.add(transition)
                    self._det_parts[i]._automaton.alphabet[transition[1]] = self._automaton.alphabet[transition[1]]
                if transition[0] in self._allNoncollSets[i] and transition[2] not in self._allNoncollSets[i]:
                    self._det_parts[i]._outset.add(transition[2])
                if transition[0] not in self._allNoncollSets[i] and transition[2] in self._allNoncollSets[i]:
                    self._det_parts[i]._inset.add(transition[0])
                    self._det_parts[i]._automaton.transitions.add(transition)
                    self._det_parts[i]._automaton.alphabet[transition[1]] = self._automaton.alphabet[transition[1]]
                    self._det_parts[i]._automaton.states[transition[0]] = self._automaton.states[transition[0]]

            # Set flags of DFA part
            self._det_parts[i]._automaton.Flags["Deterministic"] = True
            self._det_parts[i]._automaton.Flags["Epsilon Free"] = True
            # remove character classes of deterministic part
            self._det_parts[i].remove_char_classes()
            
            # If DFA char classes are used generate them and and set flag to True
            if useCharClasses == True:
                self._det_parts[i].reduce_alphabet()
                self._reduced = True
            # otherwise only set flag to False
            else:
                self._reduced = False
            
            # If transition table mapping method is 0, call heuristic
            if table_mapping_method == 0:
                # Generate transition table mapping by heuristic
                res = self._det_parts[i].getTableByHeuristic()
                # Set sollution of mapping
                self._det_parts[i]._solution = res[0]
                # Set memory size
                self._det_parts[i]._mem_size = res[1]
                # Set state size to acording to memory size
                self._det_parts[i]._stateWidth = int(math.ceil(math.log(self._det_parts[i]._mem_size,2)))
                #print(res[1])
            #  If transition table mapping method is 1, call genetic algorithm
            if table_mapping_method == 1:
                # Set default memory size to number of transitions
                res = None
                size = len(self._det_parts[i]._automaton.transitions)
                # While sucessfull mapping is not reached, run GA
                while res == None:
                    # run GA
                    res = self._det_parts[i].getTableByGA(size)
                    # if no solution was found, enlarge memory space
                    if res == None:
                        size = 2*size
                # Set sollution of mapping
                self._det_parts[i]._solution = res[0]
                # Set memory size
                self._det_parts[i]._mem_size = res[1]
                self._det_parts[i]._stateWidth = int(math.ceil(math.log(self._det_parts[i]._mem_size,2)))
        
        # Generate set of collision states, this set is created from unused 
        # sets of collision states
        collSet = set()
        for i in range (min(number_of_parts, len(self._allNoncollSets)), len(self._allNoncollSets)):
            collSet |= self._allNoncollSets[i]
        
        # Get size of collision set
        colSize = len(collSet)
        
        # Update statistics
        self._statistics["NCA Sizes"] = nonColSize
        self._statistics["N Size"] = colSize
        
        # Compute number of transitions and states in DFA parts, input and 
        # output transitions 
        detSizeTrans = list()
        detSizeStates = list()
        detSizeInsets = list()
        detSizeOutsets = list()
        for i in range(0, len(self._det_parts)):
            detSizeTrans.append(len(self._det_parts[i]._automaton.transitions))
            detSizeStates.append(len(self._det_parts[i]._automaton.states))
            detSizeInsets.append(len(self._det_parts[i]._inset))
            detSizeOutsets.append(len(self._det_parts[i]._outset))
            self._det_parts[i]._index = i
        
        # Update statistics
        self._statistics["DFA Parts States"] = detSizeStates
        self._statistics["DFA Parts Transitions"] = detSizeTrans
        self._statistics["DFA Parts Inset"] = detSizeInsets
        self._statistics["DFA Parts Outset"] = detSizeOutsets
            
        # Create empty nondeterministic part    
        self._nfa_part = nfa_nondet_part()
        
        # Create sets of input and output states
        self._nfa_part._inset       = set()     # set of starting states
        self._nfa_part._outset      = set()     # set of outgoing states
        
        # Copy states from original automaton into nondeterministic part
        for state in self._automaton.states.keys():
            if state in collSet:
                self._nfa_part._automaton.states[state] = self._automaton.states[state]
        
        # Copy final states from original automaton into nondeterministic part
        for fstate in self._automaton.final:
            if fstate in collSet:
                self._nfa_part._automaton.final.add(fstate)

        # Copy transitions from original automaton into nondeterministic part,
        # generate insets and outsets (transitions between nondeterministic 
        # part and deterministic parts) and copy symbols from original 
        # automaton into nondeterministic part
        for transition in self._automaton.transitions:
            if transition[0] in collSet and transition[2] in collSet:
                self._nfa_part._automaton.transitions.add(transition)
                self._nfa_part._automaton.alphabet[transition[1]] = self._automaton.alphabet[transition[1]]
            if transition[0] in collSet and transition[2] not in collSet:
                self._nfa_part._outset.add(transition[2])
                self._nfa_part._automaton.transitions.add(transition)
                self._nfa_part._automaton.alphabet[transition[1]] = self._automaton.alphabet[transition[1]]
            if transition[0] not in collSet and transition[2] in collSet:
                self._nfa_part._inset.add(transition[2])

        # if DFA char classes are used, generate them and mark appropriate flag
        if useCharClasses == True:
            self._nfa_part.reduce_alphabet()
            self._reduced = True
        # Otherwise just set the flag
        else:
            self._reduced = False

        # Update statistics
        self._statistics["NFA Part States"] = len(self._nfa_part._automaton.states)
        self._statistics["NFA Part Transitions"] = len(self._nfa_part._automaton.transitions)
        self._statistics["NFA2DFAs Target states"] = len(self._nfa_part._outset)
         
          
    # ----------------- Internal functions -----------
    def _GetCollisionCard(self, ModstatesCollisions):
        """
            Compute cardinality of dict of collision sets.
            
            :param ModstatesCollisions: Dictionary containing sets of collision\
                                        states for each state.
            :type ModstatesCollisions: dict(int -> set(int))
            
            :returns: Dictionary of cardinality for each state
            :rtype: dict(int -> int)
        """
        collisionCard = dict()
        for state in ModstatesCollisions.keys():
            collisionCard[state] = len(ModstatesCollisions[state])
        return collisionCard
        
    def _GetRemovedCollisions(self, state, ModstatesCollisions):
        """
            Heuristic 2 - for given state compute how much collisions will     \
                          removed if we will use this state.
                          
            :param state: Selected state for which the number of removed       \
                          collisons is computed.
            :type state: int
            :param ModstatesCollisions: Collision states for all states
            :type ModstatesCollisions: dict(int->set(int))
            
            :returns: Number of removed collisions.
            :rtype: int
        """
        collisions = ModstatesCollisions[state]
        num = 0
        for i in collisions:
            num += len(ModstatesCollisions[i])
        return num

    def _minSet(self, collisionCard):
        """
            Find set of states with minimal number of collisions.
            
            :param collisionCard: Dictionary of cardinality of collision set   \
                                  for each state.
            :type collisionCard: dict(int -> int)
            
            :returns: Returns set of states with minimal number of collisions.
            :rtype: set(int)
        """
        minimalSet = set()
        keys = collisionCard.keys()
        if len(collisionCard) > 0:
            #minimum = collisionCard[keys[0]]
            minimum = min(collisionCard.values())
            for k in collisionCard.keys():
                if collisionCard[k] < minimum:
                    minimum = collisionCard[i]
            for k in collisionCard.keys():
                if collisionCard[k] == minimum:
                    minimalSet.add(k)
        return minimalSet

    def _findMaximumHeur(self, minimalSet, ModstatesCollisions):
        """
            Find one best state - acording heuristic 2.
            
            :param minimalSet: Set of states with minimal number of collisions.
            :type minimalSet: set(int)
            :param ModstatesCollisions: Collision states for all states.
            :type ModstatesCollisions: dict(int->set(int))
            
            :returns: Best state with maximal number of removed collisions.
            :rtype: int
            
        """
        maximalSet = set()
        # Get number of removed collisions
        maximum = self._GetRemovedCollisions(copy.deepcopy(minimalSet).pop(), ModstatesCollisions)
        # Compute the maximum
        for i in minimalSet:
            if self._GetRemovedCollisions(i, ModstatesCollisions) > maximum:
                maximum = self._GetRemovedCollisions(i, ModstatesCollisions)
        # Get set ob states with maximal number of removed collisions
        for i in minimalSet:
            if self._GetRemovedCollisions(i, ModstatesCollisions) == maximum:
                maximalSet.add(i)
        #if len(maximalSet) > 1:
            # do it recursive sometimes
        # Return the best state
        return maximalSet.pop()
            
    def _removeStateFromstatesCollisions(self, state, ModstatesCollisions):
        """
            Remove collision from ModstatesCollisions.
            
            :param state: State which is to be removed from ModstatesCollisions.
            :type state: int
            :param ModstatesCollisions: Collision states for all states.
            :type ModstatesCollisions: dict(int->set(int))
            
            :returns: Collision states for all states with removed state.
            :rtype: dict(int->set(int))
        """
        for k in ModstatesCollisions.keys():
            if state in ModstatesCollisions[k]:
                ModstatesCollisions[k].remove(state)
        return ModstatesCollisions
            
    def _NonCollisionSet(self, ModstatesCollisions):
        """
            Compute set of states which aren't in collision. Two states are in \
            collision if they can be active together. Function is changing     \
            the ModstatesCollisions.
            
            :param ModstatesCollisions: Collision states for all states.
            :type ModstatesCollisions: dict(int->set(int))
            
            :returns: Set of states which aren't in collision.
            :rtype: set(int)
            
        """
        nonCollision = set()
        # Get cardinality of collisions
        collisionCard = self._GetCollisionCard(ModstatesCollisions)
        # we don't need to compute heuristic, etc. for states where are 0 collisios
        delset = set() 
        for i in ModstatesCollisions.keys():
            if collisionCard[i] == 0:
                nonCollision.add(i)
                delset.add(i)
        # Delete the states with 0 collisions from ModstatesCollisions and
        # from collisionCard
        for i in delset:
            del ModstatesCollisions[i]
            del collisionCard[i]
        # While there is some state with collisions
        while len(ModstatesCollisions) > 0:
            # Get set of states with minimal number of collisions
            minimalSet = self._minSet(collisionCard)
            # If there is more than one state in min set, find one with largest 
            # number of removed collisions if used
            if len(minimalSet) > 1:
                state = self._findMaximumHeur(minimalSet, ModstatesCollisions)
            else:
                state = minimalSet.pop()
            # Add it to non collision set
            nonCollision.add(state)
            # Get copy of collisions for this state
            collisions = copy.deepcopy(ModstatesCollisions[state])
            # Remove this state from all relevent structures
            for i in collisions:
                del ModstatesCollisions[i]
                del collisionCard[i]
                ModstatesCollisions = self._removeStateFromstatesCollisions(i, ModstatesCollisions)
            del ModstatesCollisions[state]
            del collisionCard[state]
            ModstatesCollisions = self._removeStateFromstatesCollisions(state, ModstatesCollisions)
            # Update cardinality of collisions
            collisionCard = self._GetCollisionCard(ModstatesCollisions)
        # Return set of states without collision
        return nonCollision

    def _modSet(self, workStatesCollisions, ModstatesCollisions, usedStates):
        """
            Update dictionaries.
            
            :param workStatesCollisions: Working copy of ModstatesCollisions.
            :type workStatesCollisions: dict(int->set(int))
            
            :param ModstatesCollisions: Collision states for all states.
            :type ModstatesCollisions: dict(int->set(int))
            
            :param usedStates: Set of states used in last iteration.
            :type usedStates: set(int)
            
            :returns: Updated collision states for all states.
            :rtype: dict(int->set(int))
            
        """
        workStatesCollisions = copy.deepcopy(ModstatesCollisions)
        for i in usedStates:
            del workStatesCollisions[i]
            workStatesCollisions = self._removeStateFromstatesCollisions(i, workStatesCollisions)
        return workStatesCollisions

    def _AllNonCollisionSets(self, ModstatesCollisions):
        """
            Compute all noncollision sets - sets are in collision, but elements\
            of set are not in collisions.
            
            :param ModstatesCollisions: Collision states for all states.
            :type ModstatesCollisions: dict(int->set(int))
            
            :returns: List of noncollision sets.
            :rtype: list(set(int))
        """
        # Init variables
        allSets = list()
        usedStates = set()
        workStatesCollisions = copy.deepcopy(ModstatesCollisions)
        # Get first noncollision set
        noncoll = self._NonCollisionSet(workStatesCollisions)
        # Add this set to the result
        allSets.insert(len(allSets),noncoll)
        # Update set of resolved states
        usedStates |= noncoll
        # Update working copy of ModstatesCollisions.
        workStatesCollisions = self._modSet(workStatesCollisions, ModstatesCollisions, usedStates)
        # Until all noncollision sets are computed, compute next noncollision set
        while len(workStatesCollisions)> 0:
            # Get next noncollision set
            noncoll = self._NonCollisionSet(workStatesCollisions)
            # Add this set to the result
            allSets.insert(len(allSets),noncoll)
            # Update set of resolved states
            usedStates |= noncoll
            # Update working copy of ModstatesCollisions.
            workStatesCollisions = self._modSet(workStatesCollisions, ModstatesCollisions, usedStates)
        # Return the result
        return allSets
                        
    # ----------------- NFA estimation ---------------  
    def report_nfa_logic(self):
        """
            Return estimation of LUTs, FFs and BRAMs consumed by               \
            nondeterministic parts.
            
            :returns: Amount of logic consumed by the algorithm. Returns       \ 
                      in (LUT, FF, BRAMS) format.
            :rtype: list(int, int, int)
        """
        self._nfa_part.get_HDL()
        return self._nfa_part.report_logic()

    # ---------------- Det parts representation ------------      
    def report_dfa_logic(self, bram_size = 2048, bram_word = 8 ):
        """
            Return estimation of LUTs, FFs and BRAMs consumed by all           \
            deterministic parts.
            
            :param bram_size: Size of BRAM in words
            :type bram_size: int
            
            :param bram_word: Size of BRAM word in bits
            :type bram_word: int
            
            :returns: Amount of logic consumed by the algorithm. Returns       \ 
                      in (LUT, FF, BRAMS) format.
            :rtype: list(int, int, int)
        """
        lut = 0
        ff = 0
        mem = 0
        for dfa in self._det_parts:
             res = dfa.report_logic(bram_size, bram_word)
             lut += res[0]
             ff += res[1]
             mem += res[2]
        
        return (lut, ff, mem)

    # --------------- OverAll statistics ----------------
    def report_memory(self):
        """
            Return estimation of consumed memory (in Bytes).
            
            :returns: Estimation of consumed memory in bytes.
            :rtype: int
        """
        res = 0
        for dfa in self._det_parts:
             res += dfa.ttable_overlap()
        return res
        
    def report_logic(self, bram_size = 2048, bram_word = 8 ):
        """
            Return estimation of LUTs, FFs and BRAMs consumed by whole         \
            nfa_spli architecture.
            
            :param bram_size: Size of BRAM in words
            :type bram_size: int
            
            :param bram_word: Size of BRAM word in bits
            :type bram_word: int
            
            :returns: Amount of logic consumed by the algorithm. Returns       \ 
                      in (LUT, FF, BRAMS) format.
            :rtype: list(int, int, int)
        """
        
        nfa = self.report_nfa_logic()
        dfa = self.report_dfa_logic(bram_size, bram_word)
        preDecoderSize = 0
        if self._reduced == True:
            preDecoderSize = int(math.ceil((len(self._det_parts)+1) / 8))
        return (nfa[0] + dfa[0], nfa[1] + dfa[1], nfa[2] + dfa[2] + preDecoderSize)
        
    # --------- OverAll HDL generation -------
    def get_HDL(self):
        """ 
            Generate VHDL code representing whole nfa split architecture.
            
            :returns: File names and VHDL code for all nfa split units.
            :rtype: dict(string -> string)
        """
        # Init variables
        hdls = dict()
        
        # Generarte code for nondeterministic part
        hdls["nondet_part.vhd"] = self._nfa_part.get_HDL()
        # Generate code for all deterministic parts
        for i in range(0, len(self._det_parts)):
            # Generate code for deterministic part
            hdls["det_part" + str(i) + ".vhd"] = self._det_parts[i].detunit_HDL()
            # Generate encoder code
            hdls["encoder" + str(i) + ".vhd"] = self._det_parts[i].encoder_HDL()
            # Generate decoder code
            hdls["decoder" + str(i) + ".vhd"] = self._det_parts[i].decoder_HDL()
            # Generate code for memory of deterministic part and RAM initialisation file
            res = self._det_parts[i]._memory_HDL()
            hdls["memory" + str(i) + ".vhd"] = res[0]
            hdls[res[2]] = res[1]
        
        # Open templete file
        f = open(self.template, "rb");
        # Read the template
        blob = f.read()
        # Split the templete acording to %$%, for later adition of generated 
        # code
        tmp = re.split("%\$%", blob)
        # Close the file
        f.close()

        # Generate the code of the whole unit
        result = tmp[0] + str(8) + tmp[1] + str(self._nfa_part._fStateNum) + tmp[2] + str(len(self._nfa_part._inset)) + tmp[3] + str(len(self._nfa_part._outset)) + tmp[4]
        
        # Add the generated code to the dict
        hdls["nfa_split.vhd"] = result
        
        # Set number of final states
        self._fStateNum = self._automaton.final
        
        # Return dict of generated code
        return hdls;

    def split_det_parts(self, k):
        """
            Split deterministic parts in order to reduce complexity of
            input encoder. Each part is split to k-parts.
            
            :param k: Each deterministic part is split into k-parts.
            :type k: int
        """
        # List of new deterministic parts
        newDetPartsList = list()
        
        # Split each deterministic part into k parts
        for i in range(0, len(self._det_parts)):
            parts = list()
            #self._det_parts[i]._createCharClasses()
            parts = self._det_parts[i].split_det_part(k)
            newDetPartsList += parts
        
        # Update list of all deterministic parts
        self._det_parts = newDetPartsList
        
        #for i in range(0, len(self._det_parts)):
        #    self._det_parts[i]._removeCharClasses()
        
        # Update insets, outsets, transitions and states for all new parts and
        # perform new mapping of transition tables
        detSizeTrans = list()
        detSizeStates = list()
        detSizeInsets = list()
        detSizeOutsets = list()
        for i in range(0, len(self._det_parts)):
            detSizeTrans.append(len(self._det_parts[i]._automaton.transitions))
            detSizeStates.append(len(self._det_parts[i]._automaton.states))
            detSizeInsets.append(len(self._det_parts[i]._inset))
            detSizeOutsets.append(len(self._det_parts[i]._outset))
            self._det_parts[i]._index = i
            res = self._det_parts[i].getTableByHeuristic()
            self._det_parts[i]._solution = res[0]
            self._det_parts[i]._mem_size = res[1]
            self._det_parts[i]._stateWidth = int(math.ceil(math.log(self._det_parts[i]._mem_size,2)))
        
        # Update statistics
        self._statistics["DFA Parts States"] = detSizeStates
        self._statistics["DFA Parts Transitions"] = detSizeTrans
        self._statistics["DFA Parts Inset"] = detSizeInsets
        self._statistics["DFA Parts Outset"] = detSizeOutsets
        
    def save(self, file):
        """ 
            Save nfa_split object using cPickle.
            
            :param file: File name.
            :type file: string
        """
        output = open(file, 'wb')
        cPickle.dump(self, output)
        output.close()
        
    def load(self, file):
        """ 
            Load nfa_split object using cPickle. 
            
            :param file: File name.
            :type file: string
            
            :returns: Loaded nfa_split object
            :rtype: nfa_split
        """
        pkl_file = open(file, 'rb')
        data = cPickle.load(pkl_file)
        pkl_file.close()
        return data
        