###############################################################################
#  nfa_split_det_part.py: Module for PATTERN MATCH - deterministic part of NFA
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Vlastimil Kosar <ikosar@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

# Import Netbench modules
from netbench.pattern_match import b_state
from netbench.pattern_match.b_nfa import b_nfa
from netbench.pattern_match.b_dfa import b_dfa
from netbench.pattern_match import nfa_data
from netbench.pattern_match import sym_char
from netbench.pattern_match import sym_char_class
from netbench.pattern_match import b_symbol
from netbench.pattern_match.algorithms.clark_nfa import clark_nfa
from netbench.pattern_match.pattern_exceptions import COMPUTE_ERROR, empty_automaton_exception
from netbench.pattern_match import aux_func

# Import bitstring module
from netbench.pattern_match.bin.library.bitstring import BitStream, BitArray

# Import general modules
import random
import gc
import re
import cPickle
from collections import deque
import os
import sys
import math
import copy

class nfa_det_part(b_dfa):
    """ 
        Deterministic part of NFA.
        
        Generated VHDL code depends on those modules: det_part.vhd,            \
        decoder.vhd, encoder.vhd and memory.vhd. Those modules are located in  \
        directory templates/vhdl.
        
        :param index: Number of deterministci part.
        :type index: int
        
    """
    def __init__(self, index = 0):
        """
            Class constructor.
        """
        b_dfa.__init__(self)
        self._inset       = set()     # set of starting states
        self._outset      = set()     # set of outgoing states
        self._stateWidth  = 16        # Data width of state
        self._statistic   = dict()    # Statistics dictionary
        self._LUTInputs   = 6         # Number of LUT inputs in FPGA
        self._index = index           # Number of deterministci part
        # Get base directory of pattern match
        base = aux_func.getPatternMatchDir()
        # VHDL template of whole deterministic unit
        self._template = base  + "/algorithms/experimental/nfa_split/vhdl/det_part.vhd" 
        # VHDL template of decoder
        self._decoder_template = base + "/vhdl/decoder.vhd"
        # VHDL template of encoder
        self._encoder_template = base + "/vhdl/encoder.vhd"
        # VHDL template of memory
        self._memory_template = base + "/algorithms/experimental/nfa_split/vhdl/nfa_split_memory.vhd"
        # Init the memory size to 0
        self._mem_size = 0
        # Init memory mapping solution to None
        self._solution = None


    def getTableByGA(self, maxSize):
        """
            Compute transition table mapping into list by heuristic.

            :param maxSize: Maximum size of memory useable for automata        \
                            implementation.
            :type maxSize: int

            :returns: Mapping of state id to new state id.
            :rtype: dict(int->int)
        """
        # Number of generations
        generations = 1000
        # Size of the population
        populationSize = 10
        # Probability of mutation
        mutationProbablity = 0.3
        # Rate of mutation
        mutationRate = 0.5
        # Final solution of mapping
        solution = None
        # Number of mutations per iteration
        mutationCount = 5

        # Generate mapping between source state and target states
        StateOutSymbols = dict()

        for transition in self._automaton.transitions:
            if StateOutSymbols.has_key(transition[0]) == True:
                StateOutSymbols[transition[0]].add((transition[1], transition[2]))
            else:
                StateOutSymbols[transition[0]] = set()
                StateOutSymbols[transition[0]].add((transition[1], transition[2]))
        
        # Init the random module by seed derived from time()
        random.seed()
        
        # Init the mapping function between indexes and state numbers
        mapper = [0] * len(StateOutSymbols)
        index = 0
        for state in StateOutSymbols.keys():
            mapper[index] = state
            index += 1

        # Generate the initial population
        population = list()
        for i in range(0, populationSize):
            subject = [0] * len(StateOutSymbols)
            index = 0
            for j in range(0, len(mapper)):
                subject[index] = random.randint(0, maxSize - 1)
                index += 1

            population.append(subject)

        # Run the evolution process
        for i in range(0, generations):
            # Get fitness vector for whole population
            fitnesses = self._evaluate(population, maxSize, StateOutSymbols, mapper)
            # Create new population
            population = self._create(population, mapper, fitnesses, mutationProbablity, mutationRate, mutationCount, maxSize)
            #print("Generation: " + str(i))

        # If minimal fitness is <= 0 we have reached a solution
        if min(fitnesses) <= 0:
            # Craete memory mapping solution
            solution = dict()
            for i in range(0, len(mapper)):
                solution[mapper[i]] = population[len(population) - 1][i]

            # Build memory map for check
            i = len(population) - 1
            memory = [0] * maxSize
            fitness = 0
            for j in range(0, len(mapper)):
                for transition in StateOutSymbols[mapper[j]]:
                    addr = (population[i][j] ^ ord(self._automaton.alphabet[transition[0]].char)) % maxSize
                    #print(str(population[i][j]) + " x " + str(ord(self._automaton.alphabet[transition[0]].char)) + "["+self._automaton.alphabet[transition[0]].char+"] = " + str(addr))
                    if memory[addr] != 0:
                        fitness += 1
                    memory[addr] += 1

        # Return evolved solution
        return solution

    def _evaluate(self, population, maxSize, StateOutSymbols, mapper):
        """
            Evaluates population.

            :param population: Population to be evaluated.
            :type population: list(list(int))
            :param maxSize: Maximum size of memory useable for automata implementation.
            :type maxSize: int
            :param StateOutSymbols: Mapping from state id to target state id via symbol id.
            :type StateOutSymbols: dict(int -> set(tupple(int, int)))
            :param mapper: Mapping from state id to index.
            :type mapper: list(int)
            :returns: Fitness for population.
            :rtype: list(int)
        """
        # Create and init fitness vector
        fitnesses = [0] * len(population)
        # Compute the fitness for whole population
        for i in range(0, len(population)):
            # Memory map for current subject
            memory = [0] * maxSize
            # Fitness
            fitness = 0
            # Max used address
            maxAddr = 0
            # Compute the memory map and compute the fitness
            for j in range(0, len(mapper)):
                for transition in StateOutSymbols[mapper[j]]:
                    addr = (population[i][j] ^ ord(self._automaton.alphabet[transition[0]].char)) % maxSize
                    if memory[addr] != 0:
                        fitness += 1
                    if addr > maxAddr:
                        maxAddr = addr
                    memory[addr] += 1

            # If fitness is 0 (no collision in mapping), compute optimisation 
            # variant of this fitness. This variant is used to optimise the size
            # of utilised memory
            if fitness == 0:
                fitness = fitness - (maxSize - 1) + maxAddr
            
            fitnesses[i] = fitness
        # Return fitness vector
        return fitnesses

    def _rank(self, rank, ssum):
        """
            Perform rank selection.

            :param rank: Rank of population.
            :type rank: list(int)
            :param ssum: Sum of all ranks.
            :type ssum: int
            :returns: Index of selected individual.
            :rtype: int
        """
        rnd = random.randint(0, ssum - 1)

        nsum = 0
        for i in range(0, len(rank)):
            nsum += rank[i]
            if nsum > rnd:
                return i
        return len(rank) - 1

    def _create(self, population, mapper, fitnesses, mutationProbablity, mutationRate, mutationCount, maxSize):
        """
            Creates new population from the old one.

            :param population: Old population.
            :type population: list(list(int))
            :param mapper: Mapping from state id to index.
            :type mapper: list(int)
            :param fitnesses: Fitness of old population.
            :type fitnesses: list(int)
            :param mutationProbablity: Mutation probability.
            :type mutationProbablity: float
            :param mutationRate: Mutation rate.
            :type mutationRate: float
            :param mutationCount: Number of mutations.
            :type mutationCount: int
            :param maxSize: Maximum size of memory useable for automata        \
                            implementation.
            :type maxSize: int

            :returns: New population.
            :rtype: list(list(int))
        """
        # Create list for new population
        newPopulation = list()
        # Get size of population
        length = len(population)
        # Get probability for mutation and cross
        if mutationProbablity < 1:
            cross = max(int((1.0 - mutationProbablity) * length) - 1, 0)
            muta = length - cross - 1
        else:
            muta = length - 1
            cross = 0

        # Init rank vactor
        rank = [0] * length
        fit = copy.deepcopy(fitnesses)
        index = length
        ssum = 0
        # Compute the rank
        for i in range(0, length):
            minimum = fit[0]
            ind = 0
            for j in range(1, len(fit)):
                if (fit[j] < minimum):
                    minimum = fit[j]
                    ind = j
            rank[ind] = index
            if i == 0:
                best = ind
            ssum += index
            index -= 1
            fit[ind] = 999999999

        #sum = 0
        #for i in range(0, length):
            #print(i)
            #print(keys)
            #print(keys[i])
            #print(rankDict[keys[i]])
            #rank[rankDict[keys[i]]] = index
            #sum += index
            #index -= 1

        # Perform cross operation
        for i in range (0, cross):
            parent1 = self._rank(rank, ssum)
            parent2 = self._rank(rank, ssum)
            point = random.randint(0, len(mapper) - 1)
            child = [0] * len(mapper)
            for j in range(0, len(mapper)):
                if j < point:
                    child[j] = population[parent1][j]
                else:
                    child[j] = population[parent2][j]
            newPopulation.append(child)

        # Perform mutation
        for i in range (0, muta):
            child = copy.deepcopy(population[best])
            for j in range(0, mutationCount):
                child[random.randint(0, len(mapper) - 1)] = random.randint(0, maxSize - 1)
            newPopulation.append(child)

        # Add the bast one from last generation
        newPopulation.append(population[best])
        #print("Best: " + str(fitnesses[best]))
        # Return new population
        return newPopulation

    def getTableByHeuristic(self):
        """
            Compute transition table mapping into list by heuristic.

            :returns: Mapping of state id into new state id and size of memory \
                      needed for implementation.
            :rtype: list[dict(int->int), int]
        """
        # Memory mapping solution dictionary
        solution = dict()
        isUsed = set()

        # Mapping between source states and target states
        StateOutSymbols = dict()

        # Compute the mapping
        for transition in self._automaton.transitions:
            if StateOutSymbols.has_key(transition[0]) == True:
                StateOutSymbols[transition[0]].add((transition[1], transition[2]))
            else:
                StateOutSymbols[transition[0]] = set()
                StateOutSymbols[transition[0]].add((transition[1], transition[2]))

        # Get number of outging transitions for all states into list
        outCounts = list()
        for state in StateOutSymbols.keys():
            outCounts.append([len(StateOutSymbols[state]), state])

        # Sort the list
        outCounts.sort()

        # Init memory
        memory = [0] * len(self._automaton.transitions)
        # Is used region
        testIsUsed = [0] * 5000000
        minimum = 0
        # Minimal size of memory needed if memory overflow occured
        minOverFlow = len(memory) + 256
        index = 0
        # Compute the mapping
        while len(outCounts) > 0:
            #print(len(outCounts))
            # Get the state
            state = outCounts.pop()

            placed = False
            index = minimum
            # Place the state and its transitions into the memory space
            while placed == False and index < len(memory) + 256:
                #if index not in isUsed:
                if testIsUsed[index] == 0:
                    placed = True
                    #old = copy.copy(memory)
                    for transition in StateOutSymbols[state[1]]:
                        addr = (index ^ ord(self._automaton.alphabet[transition[0]].char))
                        if addr >= len(memory):
                            placed = False
                            if index < minOverFlow:
                                minOverFlow = index
                        else:
                            if memory[addr] != 0:
                                placed = False
                                memory[addr] += 1
                            else:
                                memory[addr] += 1
                    if placed == False:
                        for transition in StateOutSymbols[state[1]]:
                            addr = (index ^ ord(self._automaton.alphabet[transition[0]].char))
                            if addr < len(memory):
                                memory[addr] -= 1
                        #memory = old
                        index += 1
                    else:
                        if index == minimum:
                            minimum = index + 1
                        #isUsed.add(index)
                        testIsUsed[index] = 1
                        solution[state[1]] = index
                else:
                    index += 1

            # If placement is not possible change memory size and perform 
            # the placement again
            if placed == False:
                memory.extend([0] * 256)
                placed = False
                index = minOverFlow
                while placed == False and index < len(memory) + 256:
                    #if index not in isUsed:
                    if testIsUsed[index] == 0:
                        placed = True
                        #old = copy.copy(memory)
                        for transition in StateOutSymbols[state[1]]:
                            addr = (index ^ ord(self._automaton.alphabet[transition[0]].char))
                            if addr >= len(memory):
                                placed = False
                            else:
                                if memory[addr] != 0:
                                    placed = False
                                    memory[addr] += 1
                                else:
                                    memory[addr] += 1
                        if placed == False:
                            for transition in StateOutSymbols[state[1]]:
                                addr = (index ^ ord(self._automaton.alphabet[transition[0]].char))
                                if addr < len(memory):
                                    memory[addr] -= 1
                            #memory = old
                            index += 1
                        else:
                            if index == minimum:
                                minimum = index + 1
                            #isUsed.add(index)
                            testIsUsed[index] = 1
                            solution[state[1]] = index
                    else:
                        index += 1

        # Get the last used index in memory
        last = 0
        for i in range(0, len(memory)):
            if memory[i] != 0:
                last = i

        # Create final mapping
        for final in self._automaton.final:
            if solution.has_key(final) == False:
                solution[final] = index
                index += 1

        # Create placement for final or otherwise blind states
        for state in self._automaton.states.keys():
            if solution.has_key(state) == False:
                solution[state] = index
                index += 1

        # Return solution
        return [solution, last+256]

    # ---- Inset and Outset --------------
    def  starting_states(self):
        """ 
            Return set of starting (input) states.
            :returns: set of starting (input) states.
            :rtype: set(int)
        """
        return self._inset;

    def  out_states(self):
        """ 
            Return output states.
            
            Return set of output states.
            :returns: set of output states.
            :rtype: set(int)
        """
        return self._outset;


    # ---- Optimizatons ------------------
    def  split_det_part(self, k):
        """ 
            Split deterministic part in order to reduce complexity of          \
            input encoder. The det part will be split into k parts.
            
            :param k: Defines number of parts.
            :type k: int
            
            :returns: Splited parts.
            :rtype: list(nfa_det_part)
        """
        # List of new parts
        newDetPartsList = list()
        
        # List reacheable parts
        reachableStates = [0]*len(self._inset)
        
        # Create mapping between source and target states
        mapper = dict()
        for transition in self._automaton.transitions:
            if mapper.has_key(transition[0]) == True:
                mapper[transition[0]].add((transition[1], transition[2]))
            else:
                mapper[transition[0]] = set()
                mapper[transition[0]].add((transition[1], transition[2]))
        
        i = 0
        # Create list of reacheable states from input states. For each input
        # state following informations are computed:
        # - number reacheable states
        # - input state
        # - list of reacheable states
        # - list of reacheable trasitions
        # - list of symbols on reacheable trasitions
        # - list of reacheable output states
        for state in self._inset:
            states = list()
            transitionSet = list()
            alphabetSet = list()
            outsetSet = list()
            fifo = deque()
            fifo.append(state)
            used = set()
            while len(fifo) > 0:
                cstate = fifo.popleft()
                used.add(cstate)
                if mapper.has_key(cstate):
                    transitions = mapper[cstate]
                    for trans in transitions:
                        if trans[1] not in used and trans[1] not in fifo:
                            fifo.append(trans[1])
                        states.append(trans[1])
                        transitionSet.append((cstate, trans[0], trans[1]))
                        alphabetSet.append(trans[0])
                        if trans[1] in self._outset:
                            outsetSet.append(trans[1])
                        if cstate in self._outset:
                            outsetSet.append(cstate)
            
            reachableStates[i] = [len(states), state, states, transitionSet, alphabetSet, outsetSet]
            i += 1
            
               
        # Create new deterministic parts
        for i in range(0, k):
            newDetPartsList.append(nfa_det_part(self._index*k+i))
        # Sort the list of reacheable states
        reachableStates.sort()
        
        # Init index
        index = 0
        # Split the automaton into k parsts
        while index < len(reachableStates):
            # Split states
            for state in self._automaton.states.keys():
                if state in reachableStates[index][2]:
                    newDetPartsList[index % k]._automaton.states[state] = self._automaton.states[state]
            # Split final states        
            for fstate in self._automaton.final:
                if fstate in reachableStates[index][2]:
                    newDetPartsList[index % k]._automaton.final.add(fstate)
            # Split transitions
            for trans in reachableStates[index][3]:
                newDetPartsList[index % k]._automaton.transitions.add(trans)
            # Split symbols
            for symbol in self._automaton.alphabet.keys():
                if symbol in reachableStates[index][4]:
                    newDetPartsList[index % k]._automaton.alphabet[symbol] = self._automaton.alphabet[symbol]
            # Split input states
            newDetPartsList[index % k]._inset.add(reachableStates[index][1])
            # Split output states
            for outState in reachableStates[index][5]:
                newDetPartsList[index % k]._outset.add(outState)
                
            index += 1

        # Return k splited parts
        return newDetPartsList


    # ---- Transition table functions ---- 
    def  ttable_size(self):
        """
            Return  size of transition table which corresponds to              \
            deterministic part.
            
            :returns: Size of transition table.
            :rtype: int
        """
        return len(self._automaton.transitions)

    def  ttable_min(self):
        """
            Return size of memory (Bytes) which is needed for minimal          \
            representation of transition table. Character classes is           \
            used to reduce memory utilization.
            
            :returns: Minimal size of transition table in bytes.
            :rtype: int
        """
        return self.ttable_size() + self.ttable_size()*(self._stateWidth/8) + int(math.ceil(self.ttable_size() * 3 / 8)) 

    def  ttable_overlap(self):
        """
            Return size of memory (Bytes) for overlapping of tt rows.
            
            :returns: Size of memory in bytes.
            :rtype: int
        """
        return self._mem_size + self._mem_size * (self._stateWidth/8) + int(math.ceil(self._mem_size * 3 / 8))

    def  ttable_overlap_BRAMS(self, bram_size=2048, bram_word=8):
        """
            Return size of memory (BRAMS) for overlapping of tt rows.
            
            
            :param bram_size: Size of BRAM in words
            :type bram_size: int
            
            :param bram_word: Size of BRAM word in bits
            :type bram_word: int
            
            :returns: Size of memory in BRAMS.
            :rtype: int
        """
        #brams = (self._mem_size + bram_size / 2) / bram_size * bram_size
        rem = self._mem_size % bram_size
        res = self._mem_size - rem + bram_size
        #print(res)
        brams = res / bram_size
        #print(brams)
        mult = (1 + self._stateWidth/8) * 8
        #print(mult)
        #print(int(math.ceil(mult / bram_word)) * brams)
        return int(math.ceil(mult / bram_word)) * brams

    # ---- Input encoder and output decoder ---- 
    def  encoder_size(self):
        """
            Return size of input encoder in LUTs.
            
            :returns: Size of input encoder in LUTs.
            :rtype: int
        """
        if self._statistic.has_key("Encoder size") == True:
            return self._statistic["Encoder size"]
        else:
            self.encoder_HDL()
            return self._statistic["Encoder size"]

    def  encoder_HDL(self):
        """
            Return HDL description of input encoder.
            
            :returns: Returns VHDL implementation.
            :rtype: string
        """
        # List of input sets of states
        inList = list()
        
        # Open input encoder template
        f = open(self._encoder_template, "rb");
        # Read the template
        blob = f.read()
        # Split the template for later inserion of generated code
        tmp = re.split("%\$%", blob)
        
        # Populate the inset list
        for mset in self._inset:
            inList.append(mset)
            
        # Sort the list
        inList.sort()
        
        binList = [0] * len(inList)
        # Create binary representation of input list
        for i in range(0, len(inList)):
            binVal = bin(self._solution[inList[i]])

            newBin = str()
            
            for j in range(0, len(binVal)):
                if j > 1:
                    newBin += binVal[j]
                    
            binVal = newBin    
            
            newBin = str()
            
            if len(binVal) < self._stateWidth:
                for j in range(0, self._stateWidth - len(binVal)):
                    newBin += "0"
                    
            newBin += binVal
            
            binList[i] = newBin
         
        wires = list()
        # Generate list of interconection wires
        for i in range (0, self._stateWidth):
            wire = list()
            for j in range (0, len(binList)):
                if binList[j][self._stateWidth - 1 - i] == '1':
                    wire.append(j)
            wires.append(wire)
                    
        for i in range(0, len(wires)):
            wires[i].sort()
            
        #print(wires)
        wiresHDL = [str()] * self._stateWidth
        
        luts = 0
        stages = 0
        
        # Generate the VHDL code based on the wires list
        for i in range(0, len(wires)):
            # If no input wires just conect it to 0
            if len(wires[i]) < 1:
                wiresHDL[i] = "bin(" + str(i) + ") <= '0';\n"
            # Else generate the code based on the input wires
            else:
                wiresHDL[i] = "bin(" + str(i) + ") <= INPUT(" + str(wires[i][0]) + ")"
                for j in range(1, len(wires[i])):
                    wiresHDL[i] += " or INPUT(" + str(wires[i][j]) + ")"
                wiresHDL[i] += ";\n"
                # Get LUT count for this decoder
                luts += self._countLUTsForInpunts(len(wires[i]))
                # Get number of LUT stages
                stg = self._countLUTStagesForInpunts(len(wires[i]))
                if stg > stages:
                    stages = stg
        
        # Set statistics
        self._statistic["Encoder size"] = luts
        self._statistic["Encoder stages"] = stages
        
        result = str()
        
        # Concatenate the VHDL code for decoders
        for i in range(0, len(wiresHDL)):
            result += wiresHDL[i]
        
        # Generate final component representation based on the templete 
        # and generated code
        result = tmp[0] + str(self._index) + tmp[1] + str(len(self._inset)) + tmp[2] + str(self._stateWidth) + tmp[3] + str(self._index) + tmp[4] + str(self._index) + tmp[5] + result + tmp[6]
        
        # Return the VHDL code
        return result
        
    def  decoder_size(self):
        """
            Return size of output decoder in LUTs.
            
            :returns: Size of output decoder in LUTs.
            :rtype: int
        """
        if self._statistic.has_key("Decoder size") == True:
            return self._statistic["Decoder size"]
        else:
            self.decoder_HDL()
            return self._statistic["Decoder size"]
    
    def  decoder_HDL(self):
        """ 
            Return HDL description of ouput decoder.
            
            :returns: Returns VHDL implementation.
            :rtype: string
        """
        # List of output states
        outlist = list()
        
        # Open the decoder template
        f = open(self._decoder_template, "rb");
        # Read the template
        blob = f.read()
        # Split the template for later insertion of generated code
        tmp = re.split("%\$%", blob)
        
        # Create the list of output
        for set in self._outset:
            outlist.append(set)
            
        # Sort the list
        outlist.sort()
        
        # Number of utilised LUTs
        luts = 0
        
        # List of decoder descriptions
        descriptionList = list()
        
        # Generate the decoder
        for i in range(0, len(outlist)):
            # Update number of utilised LUTs 
            luts += self._countLUTsForInpunts(self._stateWidth)
            
            # Generate the decoder
            chrDec  = "    VD_" + str(i) + ": entity work.VALUE_DECODER\n" 
            chrDec += "    generic map(\n"
            chrDec += "        DATA_WIDTH => " + str(self._stateWidth) + ",\n"
            chrDec += "        VALUE      => " + str(outlist[i]) + "\n"
            chrDec += "    )\n"
            chrDec += "    port map(\n"
            chrDec += "        INPUT  => INPUT,\n"
            chrDec += "        OUTPUT => OUTPUT(" + str(i) + ")\n"
            chrDec += "    );\n\n"
            # Append the code to the list
            descriptionList.append(chrDec)
            
        result = str()
        
        # Concatenate the code
        for i in range(0, len(descriptionList)):
            result += descriptionList[i]
        
        # Update statistics
        self._statistic["Decoder size"] = luts
        
        # Generate final VHDL code based on the template and the generated code
        result = tmp[0] + str(self._index) + tmp[1] + str(self._stateWidth) + tmp[2] + str(len(self._outset)) + tmp[3] + str(self._index) + tmp[4] + str(self._index) + tmp[5] + result + tmp[6]
        
        # Return final VHDL code
        return result

    def _memory_HDL(self, wrap = False):
        """ 
            Return HDL description of deterministic unit memory. 
            
            :param wrap: Defines if memory wraping is used by the mapping..
            :type wrap: boolean
           
            :returns: Returns VHDL implementation, RAM content \
                      initialisation file and its file name.
            :rtype: (string, string, string)
        """
        # Memory description for target states part of the whole memory
        memoryTarget = [0] * self._mem_size
        # Memory description for symbol part of the whole memory
        memorySymbol = [0] * self._mem_size
        # Memory description for flag part of the whole memory
        memoryFlags = [BitArray('0b000') for i in xrange(0, self._mem_size)]
        
        # Generate memory description for the automaton
        for trans in self._automaton.transitions:
            # Get addres based on solution of memory mapping
            addr = self._solution[trans[0]] ^ ord(self._automaton.alphabet[trans[1]].char)
            # If wrap is set, recompute the addres with respect to wrapping
            if wrap == True:
                addr = addr % msize
            # If solution has this target state mapped into memory (is in DFA 
            # part), set the address to the memory one.
            if self._solution.has_key(trans[2]):
                memoryTarget[addr] = self._solution[trans[2]]
            # Otherwise use the original one (final state or in NFA)
            else:
                memoryTarget[addr] = trans[2]
            
            # Set the symbol for this address
            memorySymbol[addr] = ord(self._automaton.alphabet[trans[1]].char)
            
            # Set the flags
            # Address is valid (Transition is in transition table)
            memoryFlags[addr].invert(2)
            # Target state is in NFA part
            if trans[2] in self._outset:
                memoryFlags[addr].invert(1)
            # Target state is final
            if trans[2] in self._automaton.final:
                memoryFlags[addr].invert(0)
                
        # RAM content initialisation file
        init_ram_file = ""
        
        # Generate RAM content initialisation file
        for i in xrange(0, self._mem_size):
            # Create binary representation of target state, symbol and flags
            target_bits = BitArray(uint=memoryTarget[addr], length=self._stateWidth)
            symbol_bits = BitArray(uint=memorySymbol[addr], length=8)
            flag_bits = memoryFlags[i]
            # Concatenate them into result
            result = target_bits + symbol_bits + flag_bits
            # Add them into RAM content initialisation file
            init_ram_file += result.bin + "\n"
        
        # Open the memory template
        f = open(self._memory_template, "rb");
        # Read the template
        blob = f.read()
        # Split the template for later insertion of generated code
        tmp = re.split("%\$%", blob)
        f.close()
        # Set the init file name
        init_file_name = "memory_init_" + str(self._index) + ".txt"
        
        # Crate final VHDL code from the template and generated code
        result = tmp[0] + str(self._index) + tmp[1] + str(self._stateWidth) + tmp[2] + str(8) + tmp[3] + str(int(math.ceil(math.log(self._mem_size,2)))) + tmp[4] + str(self._mem_size) + tmp[5] + str(self._index) + tmp[6] +  str(self._index) + tmp[7] + init_file_name + tmp[8]
        
        # Return the final VHDl code, RAM content initialisation file and its file name
        return (result, init_ram_file, init_file_name)

    def memory_HDL(self, wrap = False, bram_size=2048, bram_word=8):
        """ 
            Return HDL description of deterministic unit memory. 
            
            :param wrap: Defines if memory wraping is used by the mapping..
            :type wrap: boolean
            :param bram_size: Size of BRAM in words
            :type bram_size: int
            :param bram_word: Size of word in bits
            :type bram_word: int
            
            :returns: Returns VHDL implementation.
            :rtype: string
        """
        # Check if requested size of BRAM and width of word is supported
        if (bram_size != 2048 or bram_word != 8):
            print("ERROR: WRONG BRAM SIZE - " + str(bram_size) + "x" + str(bram_word) + " - NOT IMPLEMENTED")
            raise Exception()
        
        # Get number of necessery BRAMS
        size = int(self.ttable_overlap_BRAMS() / 3)
        
        # Get nearest bigger power of 2 to size
        nearPwr = 1;
        while nearPwr < size:
            nearPwr = nearPwr * 2;
        
        # Get size of memory in rows
        msize = size * bram_size
        
        # Memory description for target states part of the whole memory
        memoryTarget = [0] * msize
        # Memory description for symbol part of the whole memory
        memorySymbol = [0] * msize
        # Memory description for flag part of the whole memory
        memoryFlags = [[0,0,0] for i in range(0, msize)]
        
        # Generate memory description for the automaton
        for trans in self._automaton.transitions:
            # Get addres based on solution of memory mapping
            addr = self._solution[trans[0]] ^ ord(self._automaton.alphabet[trans[1]].char)
            # If wrap is set, recompute the addres with respect to wrapping
            if wrap == True:
                addr = addr % msize
            # If solution has this target state mapped into memory (is in DFA 
            # part), set the address to the memory one.
            if self._solution.has_key(trans[2]):
                memoryTarget[addr] = self._solution[trans[2]]
            # Otherwise use the original one (final state or in NFA)
            else:
                memoryTarget[addr] = trans[2]
            
            # Set the symbol for this address
            memorySymbol[addr] = ord(self._automaton.alphabet[trans[1]].char)
            
            # Set the flags
            # Address is valid (Transition is in transition table)
            memoryFlags[addr][0] = 1
            # Target state is in NFA part
            if trans[2] in self._outset:
                memoryFlags[addr][1] = 1
            # Target state is final
            if trans[2] in self._automaton.final:
                memoryFlags[addr][2] = 1
        # Add BRAM address signal to the list of signals
        signals = "signal addr: std_logic_vector("+ str(int(math.log(bram_size, 2)) - 1) + " downto 0);\n"
        # Add the BRAM select signal to the list of signals
        signals += "signal sel: std_logic_vector("+ str(int(math.log(nearPwr, 2)) - 1) + " downto 0);\n"
        # Generate signals (flags, target and symbol)for each bram and add 
        # them to the list of signals
        for i in range(0, nearPwr):
            signals += "signal n_" + str(i) + ": std_logic;\n"
            signals += "signal f_" + str(i) + ": std_logic;\n"
            signals += "signal v_" + str(i) + ": std_logic;\n"
            signals += "signal target_" + str(i) + ": std_logic_vector(" + str(self._stateWidth - 1) + " downto 0);\n"
            signals += "signal symbol_" + str(i) + ": std_logic_vector(7 downto 0);\n"
        
        # This list stores the generated code for BRAMS
        logic = list()
        # Generate the VHDL code
        for i in range(0, size):
            logic.append(self._map2memory(memoryTarget, memorySymbol, memoryFlags, i*bram_size, (i+1)*bram_size, i))
        
        # MUX descripton for all BRAMs
        targetMux = str()
        symbolMux = str()
        nMux = str()
        fMux = str()
        vMux = str()
        
        # Generate target MUX VHDL code
        targetMux = "with sel select\n"
        auxStr = ""
        for i in range(0, nearPwr):
            auxStr += "target_" + str(i) + " when conv_std_logic_vector(" + str(i) + ", " + str(int(math.log(nearPwr, 2))) +"),\n"
        targetMux +="TARGET <= " + auxStr + "(others => '0') when others;\n"    
        
        # Generate symbol MUX VHDL code
        symbolMux = "with sel select\n"
        auxStr = ""
        for i in range(0, nearPwr):
            auxStr += "symbol_" + str(i) + " when conv_std_logic_vector(" + str(i) + ", " + str(int(math.log(nearPwr, 2))) +"),\n"
        symbolMux +="SYMBOL <= " + auxStr + "(others => '0') when others;\n" 
        
        # Generate n-flag MUX VHDL code
        nMux = "with sel select\n"
        auxStr = ""
        for i in range(0, nearPwr):
            auxStr += "n_" + str(i) + " when conv_std_logic_vector(" + str(i) + ", " + str(int(math.log(nearPwr, 2))) +"),\n"
        nMux +="N <= " + auxStr + "(others => '0') when others;\n" 
        
        # Generate f-flag MUX VHDL code
        fMux = "with sel select\n"
        auxStr = ""
        for i in range(0, nearPwr):
            auxStr += "f_" + str(i) + " when conv_std_logic_vector(" + str(i) + ", " + str(int(math.log(nearPwr, 2))) +"),\n"
        fMux +="F <= " + auxStr + "(others => '0') when others;\n" 
        
        # Generate v-flag MUX VHDL code
        vMux = "with sel select\n"
        auxStr = ""
        for i in range(0, nearPwr):
            auxStr += "v_" + str(i) + " when conv_std_logic_vector(" + str(i) + ", " + str(int(math.log(nearPwr, 2))) +"),\n"
        vMux +="V <= " + auxStr + "(others => '0') when others;\n" 
        
        # Add them to the logic list
        logic.append(targetMux)
        logic.append(symbolMux)
        logic.append(nMux)
        logic.append(fMux)
        logic.append(vMux)
        logic.append("addr <= MEMORY_ADDR(" + str(int(math.log(bram_size, 2)) - 1) + " downto 0);\n")
        logic.append("sel <= MEMORY_ADDR(" + str(int(math.log(bram_size, 2)) + int(math.log(nearPwr, 2)) - 1) + " downto " + str(int(math.log(bram_size, 2))) +");\n")
        
        # Open the memory template
        f = open(self._memory_template, "rb");
        # Read the template
        blob = f.read()
        # Split the template for later insertion of generated code
        tmp = re.split("%\$%", blob)
        f.close()
        
        # Concatenate generated VHDL code
        logicStr = ""
        for i in range(0, len(logic)):
            logicStr += logic[i]
        
        # Crate final VHDL code from the template and generated code
        result = tmp[0] + str(self._index) + tmp[1] + str(self._stateWidth) + tmp[2] + str(8) + tmp[3] + str(max([self._stateWidth, 8])) + tmp[4] + str(self._index) + tmp[5] + str(self._index) + tmp[6] +  signals + tmp[7] + logicStr + tmp[8]
        
        # Return the final VHDl code
        return result
            
    def _map2memory(self, memoryTarget, memorySymbol, memoryFlags, start, stop, index):
        """
            Generate BRAM VHDL representation for selected part of mapped \
            transition table (Fits exactly one BRAM). 
            
            :param memoryTarget: Memory description for target states part \
                                 of the whole memory.
            :type memoryTarget: list(int)
            :param memorySymbol: Memory description for symbol part \
                                 of the whole memory.
            :type memorySymbol: list(int)
            :param memoryFlags: Memory description for flag part \
                                 of the whole memory.
            :type memoryFlags: list(int)
            :param start: Start index of mapped part of memory.
            :type start: int
            :param stop: Final index of mapped part of memory.
            :type stop: int
            :param index: Index of generated BRAM.
            :type index: int
            
            :returns: Returns VHDL implementation.
            :rtype: string
        """
        # Get width of states in BRAM words
        targetBRAMs = int(math.ceil(self._stateWidth / 8))
        if targetBRAMs < 2:
            targetBRAMs = 2
        
        # Get the width for hex representation
        hexWidth = targetBRAMs * 2;
        
        # List of hexa representations of target states
        targetHexs = list()
        # List of hexa representations of symbols
        symbolHexs = list()
        # Generate the hexa representations of target states and symbols
        for i in range(start, stop):
            # preparing target part
            pHex = hex(memoryTarget[i])
            
            sHex = pHex[2:];
            
            
            if len(sHex) < hexWidth:
                nulStr = ""
                for j in range(0, hexWidth - len(sHex)):
                    nulStr += "0"
                rHex = nulStr + sHex
            else:
                rHex = sHex;
                    
            row = list()
            for i in range(0, hexWidth, 2):
                val = rHex[i] + rHex[i+1]
                row.append(val)
                    
            row.reverse()
            targetHexs.append(row)
            
            #preparing symbol part
            pHex = hex(memorySymbol[i])
            
            sHex = pHex[2:];
            
            if len(sHex) < 2:
                nulStr = ""
                for j in range(0, 2 - len(sHex)):
                    nulStr += "0"
                rHex = nulStr + sHex
            else:
                rHex = sHex
        
            symbolHexs.append(rHex)
           
        # Lists of hexa represenations of flags
        validHexs = list()
        nfaHexs = list()
        finalHexs = list()
        
        # Generate the lists of hexa represenations of flags
        for i in range(0, int((stop - start)/256)):
            cntV = long(0)
            cntN = long(0)
            cntF = long(0)
            for j in range(0, 256):
                numV = long(memoryFlags[i*256+j][0]) << j
                numN = long(memoryFlags[i*256+j][1]) << j
                numF = long(memoryFlags[i*256+j][2]) << j
                cntV = cntV | numV
                cntN = cntN | numN
                cntF = cntF | numF
            
            pHex = hex(cntV)
            
            sHex = pHex[2:];
            
            if len(sHex) < 64:
                nulStr = ""
                for j in range(0, 64 - len(sHex)):
                    nulStr += "0"
                rHex = nulStr + sHex
            else:
                rHex = sHex
            
            rrHex = ""
            for j in range(0, len(rHex) - 1):
                rrHex += rHex[i]
            
            validHexs.append(rrHex)
            
            pHex = hex(cntN)
            
            sHex = pHex[2:];
            
            if len(sHex) < 64:
                nulStr = ""
                for j in range(0, 64 - len(sHex)):
                    nulStr += "0"
                rHex = nulStr + sHex
            else:
                rHex = sHex
            
            rrHex = ""
            for j in range(0, len(rHex) - 1):
                rrHex += rHex[i]
            
            nfaHexs.append(rrHex)
            
            pHex = hex(cntF)
            
            sHex = pHex[2:];
            
            if len(sHex) < 64:
                nulStr = ""
                for j in range(0, 64 - len(sHex)):
                    nulStr += "0"
                rHex = nulStr + sHex
            else:
                rHex = sHex
            
            rrHex = ""
            for j in range(0, len(rHex) - 1):
                rrHex += rHex[i]
            
            finalHexs.append(rrHex)
        
        # Final list of target state's hexa representations formated for BRAM
        targetHexsFinal = list()
        for i in range(0, int((stop - start)*8/256)):
            hexStr = [str()]*len(targetHexs[0])
            for j in range(0, 32):
                for k in range(0, len(targetHexs[0])):
                    hexStr[k] = targetHexs[i*32+j][k] + hexStr[k]
            targetHexsFinal.append(hexStr)
        
        # Final list of symbol's hexa representations formated for BRAM
        symbolHexsFinal = list()
        for i in range(0, int((stop - start)*8/256)):
            hexStr = str()
            for j in range(0, 32):
                    hexStr = symbolHexs[i*32+j] + hexStr
            symbolHexsFinal.append(hexStr)
            
        # List of VHLD logic
        logicList = list()
        # VHDL BRAM representation for target state BRAM 0
        targetBram0 = str()
        # VHDL BRAM representation for target state BRAM 1
        targetBram1 = str()
        # VHDL BRAM representation for symbol BRAM 
        symbolBram = str()
        
        # Generate VHDL representation of target state BRAM 0
        targetBram0 = "BRAM_TARGET_0_" + str(index) + ": RAMB16_S9_S9\n"
        targetBram0 += "  -- synthesis translate_off\n"
        targetBram0 += "  generic map (\n"
        for i in range(0, len(targetHexsFinal)):
            hexNum = hex(int(i))
            sHex = hexNum[2:];
            if len(sHex) < 2:
                sHex = "0" + sHex
            targetBram0 += "INIT_" + sHex + " => X\"" + targetHexsFinal[i][0]
            if i != len(targetHexsFinal) - 1:
                targetBram0 += "\",\n"
            else:
                targetBram0 += "\"\n"
        for i in range(0, len(nfaHexs)):
            hexNum = hex(int(i))
            sHex = hexNum[2:];
            if len(sHex) < 2:
                sHex = "0" + sHex
            targetBram0 += "INITP_" + sHex + " => X\"" + nfaHexs[i]
            if i != len(nfaHexs) - 1:
                targetBram0 += "\",\n"
            else:
                targetBram0 += "\"\n"
        targetBram0 += ")\n"
        targetBram0 += "-- synthesis translate_on\n"
        targetBram0 += "port map (\n"
        targetBram0 += "DIA     => X\"00\",\n"
        targetBram0 += "DIPA    => '0',\n"
        targetBram0 += "ADDRA   => addr\n"
        targetBram0 += "ENA     => MEMORY_RQ\n"
        targetBram0 += "WEA     => '0',\n"
        targetBram0 += "SSRA    => '0',\n"
        targetBram0 += "CLKA    => CLK,\n"
        targetBram0 += "DOA     => target_"+str(index)+"(7 downto 0),\n"
        targetBram0 += "DOPA    => n_" + str(index) + ",\n"
        targetBram0 += "DIB     => X\"00\",\n"
        targetBram0 += "DIPB    => '0',\n"
        targetBram0 += "ADDRB   => addr\n"
        targetBram0 += "ENB     => '0'\n"
        targetBram0 += "WEB     => '0',\n"
        targetBram0 += "SSRB    => '0',\n"
        targetBram0 += "CLKB    => CLK,\n"
        targetBram0 += "DOB     => open,\n"
        targetBram0 += "DOPB    => open,\n"
        targetBram0 += ");\n\n"
        
        # Generate VHDL representation of target state BRAM 1
        targetBram1 = "BRAM_TARGET_1_" + str(index) + ": RAMB16_S9_S9\n"
        targetBram1 += "  -- synthesis translate_off\n"
        targetBram1 += "  generic map (\n"
        for i in range(0, len(targetHexsFinal)):
            hexNum = hex(int(i))
            sHex = hexNum[2:];
            if len(sHex) < 2:
                sHex = "0" + sHex
            targetBram1 += "INIT_" + sHex + " => X\"" + targetHexsFinal[i][1]
            if i != len(targetHexsFinal) - 1:
                targetBram1 += "\",\n"
            else:
                targetBram1 += "\"\n"
        for i in range(0, len(finalHexs)):
            hexNum = hex(int(i))
            sHex = hexNum[2:];
            if len(sHex) < 2:
                sHex = "0" + sHex
            targetBram1 += "INITP_" + sHex + " => X\"" + finalHexs[i]
            if i != len(nfaHexs) - 1:
                targetBram1 += "\",\n"
            else:
                targetBram1 += "\"\n"
        targetBram1 += ")\n"
        targetBram1 += "-- synthesis translate_on\n"
        targetBram1 += "port map (\n"
        targetBram1 += "DIA     => X\"00\",\n"
        targetBram1 += "DIPA    => '0',\n"
        targetBram1 += "ADDRA   => addr\n"
        targetBram1 += "ENA     => MEMORY_RQ\n"
        targetBram1 += "WEA     => '0',\n"
        targetBram1 += "SSRA    => '0',\n"
        targetBram1 += "CLKA    => CLK,\n"
        targetBram1 += "DOA     => target_"+str(index)+"(15 downto 8),\n"
        targetBram1 += "DOPA    => f_" + str(index) + ",\n"
        targetBram1 += "DIB     => X\"00\",\n"
        targetBram1 += "DIPB    => '0',\n"
        targetBram1 += "ADDRB   => addr\n"
        targetBram1 += "ENB     => '0'\n"
        targetBram1 += "WEB     => '0',\n"
        targetBram1 += "SSRB    => '0',\n"
        targetBram1 += "CLKB    => CLK,\n"
        targetBram1 += "DOB     => open,\n"
        targetBram1 += "DOPB    => open,\n"
        targetBram1 += ");\n\n"
        
        # Generate VHDL representation of symbol BRAM
        symbolBram = "BRAM_SYMBOL_" + str(index) + ": RAMB16_S9_S9\n"
        symbolBram += "  -- synthesis translate_off\n"
        symbolBram += "  generic map (\n"
        for i in range(0, len(symbolHexsFinal)):
            hexNum = hex(int(i))
            sHex = hexNum[2:];
            if len(sHex) < 2:
                sHex = "0" + sHex
            symbolBram += "INIT_" + sHex + " => X\"" + symbolHexsFinal[i]
            if i != len(symbolHexsFinal) - 1:
                symbolBram += "\",\n"
            else:
                symbolBram += "\"\n"
        for i in range(0, len(validHexs)):
            hexNum = hex(int(i))
            sHex = hexNum[2:];
            if len(sHex) < 2:
                sHex = "0" + sHex
            symbolBram += "INITP_" + sHex + " => X\"" + validHexs[i]
            if i != len(nfaHexs) - 1:
                symbolBram += "\",\n"
            else:
                symbolBram += "\"\n"
        symbolBram += ")\n"
        symbolBram += "-- synthesis translate_on\n"
        symbolBram += "port map (\n"
        symbolBram += "DIA     => X\"00\",\n"
        symbolBram += "DIPA    => '0',\n"
        symbolBram += "ADDRA   => addr\n"
        symbolBram += "ENA     => MEMORY_RQ\n"
        symbolBram += "WEA     => '0',\n"
        symbolBram += "SSRA    => '0',\n"
        symbolBram += "CLKA    => CLK,\n"
        symbolBram += "DOA     => symbol_"+str(index)+",\n"
        symbolBram += "DOPA    => v_" + str(index) + ",\n"
        symbolBram += "DIB     => X\"00\",\n"
        symbolBram += "DIPB    => '0',\n"
        symbolBram += "ADDRB   => addr\n"
        symbolBram += "ENB     => '0'\n"
        symbolBram += "WEB     => '0',\n"
        symbolBram += "SSRB    => '0',\n"
        symbolBram += "CLKB    => CLK,\n"
        symbolBram += "DOB     => open,\n"
        symbolBram += "DOPB    => open,\n"
        symbolBram += ");\n\n"
        
        # Return generated VHDL code
        return targetBram0 + targetBram1 + symbolBram
            

    # --------------- Det Unit ----------------
    def  detunit_HDL(self):
        """
            Return HDL description of deterministic unit.
            
            :returns: Returns VHDL implementation.
            :rtype: string
        """
        # List of generated signals
        signalList = list()
        # List of generated descriptions
        descriptionList = list()
        
        # Generate mapping between regexp numbers and final states
        sameFinal = dict()
        for fstate in self._automaton.final:
            for rnum in self._automaton.states[fstate].get_regexp_number():
                if sameFinal.has_key(rnum) == True:
                    sameFinal[rnum].add(fstate)
                else:
                    sameFinal[rnum] = set()
                    sameFinal[rnum].add(fstate)
        
        
        # Set number od final state
        self._fStateNum = len(sameFinal)
        # Number of LUTs
        self._luts = 0
        
        # Add signal description into the signalList
        signalList.append("    signal bitmap_in : std_logic_vector(" + str(len(sameFinal)) + " - 1 downto 0);\n")
        # Generate the bitmap of found REs
        for final in sameFinal.keys():
            # Add signal description into the signalList
            signalList.append("    signal final_" + str(final) + " : std_logic;\n")
            # If only one final state coresponds to the RE, just conect it
            if len(sameFinal[final]) == 1:
                # Get the state from the mapping
                state = sameFinal[final].pop()
                # Add it into description list
                descriptionList.append("    final_" + str(final) + " <= state_out_" + str(state) + ";\n")
            else:
                # Otherwise or signals from all final states
                first = True
                # Update the number of LUTs
                self._luts += self._countLUTsForInpunts(len(sameFinal[final]))
                # Generate th VHDL code
                for pfinal in sameFinal[final]:
                    if first == True:
                        text = "    final_" + str(final) + " <= state_out_" + str(pfinal)
                        first = False
                    else:
                        text += " or state_out_" + str(pfinal)
                text += ";\n"
                # Add it into description list
                descriptionList.append(text)
                
        # Generate mapping of bitmap inputs
        sfKeys = sameFinal.keys()
        for i in range(0, len(sfKeys)):
            descriptionList.append("    bitmap_in(" + str(i) + ") <= final_" + str(sfKeys[i]) + ";\n")
            
        # Read the template of deterministic unit
        f = open(self._template, "rb");
        # Read the template
        blob = f.read()
        # Split the template for later insertion of generated VHDL code
        tmp = re.split("%\$%", blob)
        # Close the file
        f.close()
        
        # Generate final VHDL from the template and generated VHDL code
        result = tmp[0] + str(self._index) + tmp[1] + str(8) + tmp[2] + str(len(self._inset)) + tmp[3] + str(len(self._outset)) + tmp[4] + str(self._stateWidth) + tmp[5] + str(max([self._stateWidth, 8])) + tmp[6] + str(len(self._automaton.final)) + tmp[7] + str(self._index) + tmp[8] + str(self._index) + tmp[9] + "" + tmp[10] + str(self._index) + tmp[11] + str(self._index) + tmp[12] + str(self._index) + tmp[13] + "" + tmp[14] 
        
        # Return the final VHDL code
        return result

    # --------------- OverAll statistics ----------------
    def report_memory(self):
        """
            Reports amount of the memory consumed by deterministic part of \
            nfa split algorithm. The returned number is in the bytes.
            
            :returns: Amount of the memory consumed/
            :rtype: int
        """
        return self.ttable_overlap()
        
    def report_logic(self, bram_size=2048, bram_word=8):
        """
            Reports amount of logic consumed by the deterministic unit of \
            nfa split algorithm. 
            
            :param bram_size: Size of BRAM in words
            :type bram_size: int
            
            :param bram_word: Size of BRAM word in bits
            :type bram_word: int
            
            :returns: Amount of logic consumed by the algorithm. Returns \ 
                      in (LUT, FF, BRAMS) format.
            :rtype: list(int, int, int)
        """
        # Compute the number of LUTs
        lut = self.encoder_size() + self.decoder_size() + self._stateWidth
        # Compute the number of FFs
        ff = 8 + len(self._automaton.final)
        # Compute the number of BRAMs
        mem = self.ttable_overlap_BRAMS(bram_size, bram_word)
        # Return the value
        return (lut, ff, mem)
    
    def _countLUTsForInpunts(self, inputs):
        """
            Counts amount of LUTs necessary for implenentation of inputs-input \
            logic function. This creates simple tree structure of n-input LUTs \
            for inputs-input logical function with 1 output. This method only  \
            gives upper bound of actual implementation by synthesis tool.
            
            :param inputs: Number of logic function inputs
            :type inputs: int
            
            :returns: Amount of LUTs necessary for implenentation of           \
                      inputs-input logic function.
            :rtype: int
        """
        # Set start values
        i = inputs
        res = 0;
        # Number of inputs is lower than number of LUT inputs, just return 1
        if i <= self._LUTInputs:
            res = 1;
        # Else compute simple tree implementation
        else:
            # While number of current level inputs is bigger than number of LUT
            # inputs create new layer of luts
            while i > self._LUTInputs:
                # Perform eventual rounds for border LUTs of current level
                if i/self._LUTInputs == round(i/self._LUTInputs):
                    res = res + int(round((i / self._LUTInputs), 0))
                    i = int(round((i / self._LUTInputs), 0))
                else:
                    res = res + int(round((i / self._LUTInputs) + 0.5, 0))
                    i = int(round((i / self._LUTInputs) + 0.5, 0))
            # Add final level LUT
            res = res + 1;
        # Return number of LUTs
        return res
                
    def _countLUTStagesForInpunts(self, inputs):
        """
            Counts number of LUT stages necessary for implenentation of     \
            inputs-input logic function. This creates simple tree structure \
            of n-input LUTs for inputs-input logical function with 1 output.\
            This method only gives upper bound of actual implementation by  \
            synthesis tool.
            
            :param inputs: Number of logic function inputs
            :type inputs: int
            
            :returns: Amount of LUT stages necessary for implenentation of \
                      inputs-input logic function.
            :rtype: int
        """
        # Set start values
        i = inputs
        res = 0;
        # Number of inputs is lower than number of LUT inputs, just return 1
        if i <= self._LUTInputs:
            res = 1;
        # Else compute simple tree implementation
        else:
            # While number of current level inputs is bigger than number of LUT
            # inputs create new layer of luts
            while i > self._LUTInputs:
                # Perform eventual rounds for border LUTs of current level
                if i/self._LUTInputs == round(i/self._LUTInputs):
                    res = res + 1
                    i = int(round((i / self._LUTInputs), 0))
                else:
                    res = res + 1
                    i = int(round((i / self._LUTInputs) + 0.5, 0))
            # Add final level LUT stage
            res = res + 1;
            
        # Return number of stages
        return res
