###############################################################################
#  nfa_split_nondet_part.py: Module for PATTERN MATCH - nondeterministic part
#                            of NFA
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Vlastimil Kosar <ikosar@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$

# Import Netbench modules
from netbench.pattern_match import b_state
from netbench.pattern_match.b_nfa import b_nfa
from netbench.pattern_match.b_dfa import b_dfa
from netbench.pattern_match import nfa_data
from netbench.pattern_match import sym_char
from netbench.pattern_match import sym_char_class
from netbench.pattern_match import b_symbol
from netbench.pattern_match.algorithms.clark_nfa import clark_nfa
from netbench.pattern_match.pattern_exceptions import COMPUTE_ERROR, empty_automaton_exception
from netbench.pattern_match import aux_func
from nfa_split_det_part import nfa_det_part


# Import general modules
import random
import gc
import re
import cPickle
from collections import deque
import os
import sys
import math
import copy

class nfa_nondet_part(clark_nfa.clark_nfa):
    """
        NonDeterministic part of NFA. Based on Clark's NFA implementation.     \
        Supports also sharing of char classes if b_Sym_char_class symbols are  \
        used.
        
        Generated VHDL code depends on those modules: value_decoder.vhd,       \
        state.vhd and final_bitmap.vhd. Those modules are located in directory \
        templates/vhdl.
        
        Supported symbols for this algorithm are sym_char and sym_char_class.
    """
    def __init__(self):
        """
            Class constructor.
        """
        # Call clark's constructor - Noncanonical variant, no stride
        clark_nfa.clark_nfa.__init__(self, False, 1)
        # Set data width - fixed on 8 bits in current implementation
        self.width = 8
        # Set path to VHDL template
        self.template = aux_func.getPatternMatchDir() + "/algorithms/experimental/nfa_split/vhdl/nondet_part.vhd"
        # Init statistic data dictionary
        self._statistic = dict()
        # Set number of inputs per lut - 6 for Virtex 5/6
        self._LUTInputs = 6
        # Init set of input states
        self._inset = set()
        # Init set of output states
        self._outset = set()
        
    def  get_HDL(self, useBram = False):
        """
            Return VHDL description of NFA unit implemented in Clark's approach.
            
            :param useBram: Defines if BRAM implementation of character        \
                            decoder will be used. (True = BRAM, False = LUT) 
            
                NOTE: Only False is currently supported.
            :type useBram: boolean
            
            :returns: VHDL description of NFA unit implemented in Clark's      \
                      approach.
            :rtype: string
        """
        # Open VHDL template
        f = open(self.template, "rb");
        # Read the whole template
        blob = f.read()
        # Split the template acording to the insert here marks (%$%)
        tmp = re.split("%\$%", blob)
        
        # Set the use BRAM parameter
        self._useBram = useBram
        # Init number of used luts to 0
        self._luts = 0
        
        # Get VHDL description of shared character decoder
        chdec = self._get_char_dec_HDL()
        # Get VHDL description of mapped automaton's logic
        logic = self._get_logic_HDL()        
        # Get VHDL description of final states
        final = self._get_final_HDL()
        # Get VHDL description of input/output transitions logic
        ioset = self._get_io_HDL()
        
        # Concatenate signal description
        dataSignal = chdec[0] + logic[0] + final[0] + ioset[0]
        # Concatenate logic description - architecture
        dataImplementation = chdec[1] + logic[1] + final[1] + ioset[1]
        
        # Set generated code markers - begin
        textSignal = "-- --------------- GENERATED BY NFA_SPLIT.PY ----------------\n"
        textImplementation = "-- --------------- GENERATED BY NFA_SPLIT.PY ----------------\n"
        
        # Add signal description to output
        for element in dataSignal:
            textSignal += element
        # Add logic description to output
        for element in dataImplementation:
            textImplementation += element
        
        # Set generated code markers - end
        textSignal += "-- --------------- END ----------------\n"
        textImplementation += "-- --------------- END ----------------\n"
        # Add result into loaded VHDL template
        result = tmp[0] + str(self.width) + tmp[1] + str(self._fStateNum) + tmp[2] + str(len(self._inset)) + tmp[3] + str(len(self._outset)) + tmp[4] + textSignal + tmp[5] + textImplementation + tmp[6]
        
        # Return complete VHDL code.
        return result
        
    def _get_io_HDL(self):
        """
            Return HDL description of input/output transitions (input/output   \
            transitions are transitions between NFA part and DFA parts) 
            (signals, description). 
            
            :returns: HDL description of input/output transitions as           \
                     (signals, description).
            :rtype: tuple(list(string),list(string))
        """
        # List of signal definitions
        signalList = list()
        # list of logic description - architecture
        descriptionList = list()
        
        # Dictionary mapping index of input transition to its target state
        self._insetMap = dict()
        # Dictionary mapping index of output transition to its source state
        self._outsetMap = dict()
        
        # init the index of input transitions
        index = 0
        # Create logic for all states in inset
        for state in self._inset:
            # Add signal definition into signalList
            signalList.append("    signal inset_" + str(state) + " : std_logic;\n")
            # Add logic description into descriptionList
            descriptionList.append("    inset_" + str(state) + " <= INPUT(" + str(index) + ");\n")
            # Update the _insetMap dictionary with mapping between the state 
            # and its index
            self._insetMap[index] = state
            # Update the index
            index += 1
        
        # init the index of output transitions    
        index = 0
        # Create logic for all states in outset
        for state in self._outset:
            # Add signal definition into signalList
            signalList.append("    signal outset_" + str(state) + " : std_logic;\n")
            # Add logic descriptions into descriptionList
            descriptionList.append("    OUTPUT(" + str(index) + ") <= outset_" + str(state) + ";\n")
            descriptionList.append("    outset_" + str(state) + " <= state_in_" + str(state) + ";\n")
            # Update the _outsetMap dictionary with mapping between the state 
            # and its index
            self._outsetMap[index] = state
            # Update the index
            index += 1
        
        # Return generated signal and description list   
        return (signalList, descriptionList)
        
    def _get_logic_HDL(self):
        """ 
            Return HDL description of states and transitions as                \
            (signals, description). 
            
            :returns: HDL description of states and transitions as             \
                     (signals, description).
            :rtype: tuple(list(string),list(string)) 
        """
        # List of signal definitions
        signalList = list()
        # list of logic description - architecture
        descriptionList = list()
        
        #print "NFA part states: " + str(self._automaton.states.keys())
        #print "NFA transitions: " + str(self._automaton.transitions)
        #print "NFA inset:" + str(self._inset)
        #print "NFA outset:" + str(self._outset)
        
        # For all states generate states modules
        for state in self._automaton.states.keys():
            # Define signals for current state
            signalList.append("    signal state_in_" + str(state) + " : std_logic;\n")
            signalList.append("    signal state_out_" + str(state) + " : std_logic;\n")
            # If state is start state of automaton, set generic DEFAULT to '1'
            if state == self._automaton.start:
                text  =  "    STATE_" + str(state) + ": entity work.STATE\n"
                text +=  "    generic map(\n"
                text +=  "        DEFAULT     => '1'\n"
                text +=  "    )\n"
                text +=  "    port map(\n"
                text +=  "        CLK    => CLK,\n"
                text +=  "        RESET  => local_reset,\n"
                text +=  "        INPUT  => '0',\n"
                text +=  "        WE     => we,\n"
                text +=  "        OUTPUT => state_out_" + str(state) + "\n"
                text +=  "    );\n\n"
            # Otherwise set generic DEFAULT to '0' for any other state
            else:
                text  =  "    STATE_" + str(state) + ": entity work.STATE\n"
                text +=  "    generic map(\n"
                text +=  "        DEFAULT     => '0'\n" 
                text +=  "    )\n"
                text +=  "    port map(\n"
                text +=  "        CLK    => CLK,\n"
                text +=  "        RESET  => local_reset,\n"
                text +=  "        INPUT  => state_in_" + str(state) + ",\n"
                text +=  "        WE     => we,\n"
                text +=  "        OUTPUT => state_out_" + str(state) + "\n"
                text +=  "    );\n\n"
            # Add generated state module instantiation to corespondig
            # description list
            descriptionList.append(text)
        
        # Handle orphan outset states
        for state in self._outset - set(self._automaton.states.keys()):
            # Define signals for current state
            signalList.append("    signal state_in_" + str(state) + " : std_logic;\n")
        
        # Compute for all states their input transitions
        inputTransitions = dict()
        for transition in self._automaton.transitions:
            if inputTransitions.has_key(transition[2]) == True:
                inputTransitions[transition[2]].add((transition[0], transition[1]))
            else:
                inputTransitions[transition[2]] = set()
                inputTransitions[transition[2]].add((transition[0], transition[1]))
        
        used_inset = set()
        # Generate interconection between states and between character decoders
        # and states
        for state in inputTransitions.keys():
            # If only 1 input transition is present and state is not in inset,
            # just "and" previous state output and character decoder output
            if len(inputTransitions[state]) == 1 and state not in self._inset:
                # Get transition data
                data = inputTransitions[state].pop()
                # Add code to decription list
                descriptionList.append("    state_in_" + str(state) + " <= state_out_" + str(data[0]) + " and symbol_" + str(data[1]) + ";\n")
                # Update number of used LUTs
                self._luts += self._countLUTsForInpunts(2)
            # For more input transitions (or if state is in inset) create 
            # coresponding conection structure
            # (OUTPUT AND CHAR) or (OUTPUT AND CHAR) or . or (OUTPUT AND INSET) 
            else:
                text = str()
                first = True
                # Generate VHDL code
                for transition in inputTransitions[state]:
                    if first == True:
                        text += "    state_in_" + str(state) + " <= (state_out_" + str(transition[0]) + " and symbol_" + str(transition[1]) + ")"
                        first = False
                    else:
                        text += " or (state_out_" + str(transition[0]) + " and symbol_" + str(transition[1]) + ")"
                # if state is in inset create connection to this state
                if state in self._inset:
                    # Generate VHDL code
                    text += " or inset_" + str(state)
                    used_inset.add(state)
                    # Update number of used LUTs
                    self._luts += self._countLUTsForInpunts(2*len(inputTransitions[state]) + 1)
                # Otherwise just update number of LUTs
                else:
                    self._luts += self._countLUTsForInpunts(2*len(inputTransitions[state]))
                # Finisgh VHDL code line
                text += ";\n"
                # Add code to decription list
                descriptionList.append(text)
        
        # Handle orphan states - usualy final
        for state in self._inset - used_inset:
            # Generate VHDL code
            text = str()
            text += "    state_in_" + str(state) + " <= inset_" + str(state) + ";\n"
            # Add code to decription list
            descriptionList.append(text)
            
        # Return generated signal and description list 
        return (signalList, descriptionList)
