-- ----------------------------------------------------------------------------
-- Entity for pattern match unit wraping FL->UNIT->FL
-- ----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;

-- library containing log2 function
use work.math_pack.all;

-- ----------------------------------------------------------------------------
--                        Entity declaration
-- ----------------------------------------------------------------------------
entity WRAPPER is
   generic(
      RX_DATA_WIDTH : integer;
      TX_DATA_WIDTH : integer;
      RULES      : integer
   );
   port(
      CLK            : in std_logic;
      RESET          : in std_logic;

      -- input FL interface
      RX_DATA             :  in std_logic_vector(RX_DATA_WIDTH - 1 downto 0);
      RX_REM              :  in std_logic_vector(log2(RX_DATA_WIDTH/8)-1 downto 0);
      RX_SOF_N            :  in std_logic;
      RX_SOP_N            :  in std_logic;
      RX_EOP_N            :  in std_logic;
      RX_EOF_N            :  in std_logic;
      RX_SRC_RDY_N        :  in std_logic;
      RX_DST_RDY_N        : out std_logic;
      -- output data interface
      TX_DATA             : out std_logic_vector(TX_DATA_WIDTH - 1 downto 0);
      TX_REM              : out std_logic_vector(log2(TX_DATA_WIDTH/8)-1 downto 0);
      TX_SOF_N            : out std_logic;
      TX_SOP_N            : out std_logic;
      TX_EOP_N            : out std_logic;
      TX_EOF_N            : out std_logic;
      TX_SRC_RDY_N        : out std_logic;
      TX_DST_RDY_N        :  in std_logic
   );

end entity WRAPPER;

-- ----------------------------------------------------------------------------
--                     Architecture: full
-- ----------------------------------------------------------------------------
architecture full of WRAPPER is
    signal sof     : std_logic;
    signal eof     : std_logic;
    signal src_rdy : std_logic;
    signal dst_rdy : std_logic;
    signal vld     : std_logic;
    signal ack     : std_logic;
    signal data    : std_logic_vector(RX_DATA_WIDTH - 1 downto 0);
    signal bitmap  : std_logic_vector(RULES - 1 downto 0);
    signal zeros   : std_logic_vector(TX_DATA_WIDTH - 1 downto 0);
begin
    zeros <= (others => '0');
    
    -- Map input ports to inner signals
    data <= RX_DATA;
    sof  <= not RX_SOF_N;
    eof  <= not RX_EOF_N;
    src_rdy <= not RX_SRC_RDY_N;
    RX_DST_RDY_N <= not dst_rdy;

    -- Map inner signals to output ports 
    TX_REM <= (others => '1');
    TX_DATA <= zeros(TX_DATA_WIDTH - 1 downto RULES) & bitmap;
    TX_SOF_N <= not vld;
    TX_SOP_N <= not vld;
    TX_EOP_N <= not vld;
    TX_EOF_N <= not vld;
    TX_SRC_RDY_N <= not (vld and not ack);
    
    -- ACK register
    ack_reg: process(CLK)
    begin
        if (CLK'event and CLK = '1') then
            if (RESET = '1') then
                ack <= '0';
            else
                if vld = '1' then
                    ack <= not TX_DST_RDY_N;
                else
                    ack <= '0';
                end if;
            end if;
        end if;
    end process ack_reg;
    
    -- NOTE: set corect unit name
    unit: entity work.%$%
    generic map(
        DATA_WIDTH => RX_DATA_WIDTH,
        RULES      => RULES
    )
    port map(
        CLK            => CLK,
        RESET          => RESET,

        -- input data interface
        DATA           => data,
        SOF            => sof,
        EOF            => eof,
        SRC_RDY        => src_rdy,
        DST_RDY        => dst_rdy,

        -- output data interface
        BITMAP         => bitmap,
        VLD            => vld,
        ACK            => ack
    );

end architecture full;
